# (c) Copyright 2005-2009. CodeWeavers, Inc.
package CXMenuXDG;
use strict;

use CXLog;
use CXUtils;
use CXXMLDOM;
use CXMenu;
use base "CXMenu";



#####
#
# Folder helper functions
#
#####

sub get_root_menu($$$)
{
    my ($self, $component)=@_;

    if (!$self->{root_subname} or
        $self->{root_allowed}->{$component->{_xname}})
    {
        $component->{_xpath}="/$component->{_xname}";
        return $self->{root_menu};
    }

    if (!$self->{root_submenu})
    {
        $self->{root_submenu}=get_child_by_name($self->{root_menu}, "Menu", $self->{root_subname});
        if (!$self->{root_submenu})
        {
            $self->{root_submenu}=add_new_element($self->{root_menu}, "Menu");
            add_new_element($self->{root_submenu}, "Name", $self->{root_subname});
        }
    }
    $component->{_xpath}="/$self->{root_subname}/$component->{_xname}";
    return $self->{root_submenu};
}

sub get_xpath($$$)
{
    my ($self, $path, $create)=@_;

    my $xpath=mangle_path($path);
    # Demangle spaces. See create_folder() for details.
    $xpath =~ s/\+/ /g;

    if (defined $self->{root_subname})
    {
        my $topmenu=$xpath;
        $topmenu=~ s!^/([^/]+)(?:/.*)?$!$1!;
        if (!$self->{root_allowed}->{$topmenu})
        {
            $xpath="/$self->{root_subname}$xpath";
        }
    }
    cxlog("xpath='$xpath'\n");

    return $xpath;
}

sub get_folder_by_xpath($$)
{
    my ($self, $xpath)=@_;

    my $folder=$self->{folders}->{$xpath};
    # Check the parent node to verify that the folder has not been deleted
    return $folder if ($folder and $folder->getParentNode());

    $folder=get_tag_by_path($self->{root_menu}, "Menu", $xpath);
    # $self->{folders} also serves as a negative cache
    $self->{folders}->{$xpath}=$folder;
    return $folder;
}

sub xpath_to_folder_dir($$)
{
    my ($self, $xpath)=@_;
    # Remangle spaces. See create_folder() for details.
    $xpath =~ s/ /+/g;
    return "$self->{desktopdata}/cxmenu/$self->{data_dir}$xpath";
}

sub get_folder_dir($$$)
{
    my ($self, $folder, $xpath)=@_;

    my $folder_dir;
    my $appdir=get_child($folder, "AppDir");
    $folder_dir=get_cdata($appdir) if ($appdir);
    $folder_dir=$self->xpath_to_folder_dir($xpath) if (!defined $folder_dir);
    return $folder_dir;
}

sub create_folder($$$)
{
    my ($self, $parent, $component)=@_;
    cxlog("create_folder($component->{_xpath})\n");

    my $sname;
    my $folder=$self->{folders}->{$component->{_xpath}};
    if (!defined $folder)
    {
        $folder=get_child_by_name($parent, "Menu", $component->{_xname});
    }

    # Update the XML tree
    my ($folder_dir, $new_folder);
    if (!defined $folder)
    {
        # Add a new <Menu> tag
        my $pos=get_child($parent, "Menu");
        $folder=add_new_element($parent, "Menu", undef, $pos);
        # gnome-panel < 2.10 (e.g. Fedora Core 2 and 3) is confused if it
        # finds a '%' in the XML file. Using the XML character entity
        # instead does not help. Furthermore the Name field in the
        # .directory file will override the XML <Name> tag anyway, so let's
        # just use the mangled folder name.
        add_new_element($folder, "Name", $component->{_xname});
        $folder_dir=$self->xpath_to_folder_dir($component->{_xpath});
        add_new_element($folder, "Directory", ".directory");
        add_new_element($folder, "DirectoryDir", $folder_dir);
        add_new_element($folder, "AppDir", $folder_dir);

        if ($self->{hidden_hack})
        {
            # KDE 3.5.1 is buggy and won't show folders that have no menu entry
            # and a single sub-folder. So we always add a an empty '#Hidden'
            # folder:
            # - The '#' garantees there won't be a name collision with a real
            #   mangled folder name.
            # - Because that folder is empty it will not be shown.
            # - Folders will always have at leat two entries and thus will
            #   always be shown.
            # - However this also disables KDE's 'collapsing' feature for
            #   single-entry folders and thus we only enable this hack when
            #   needed.
            my $hidden=add_new_element($folder, "Menu");
            add_new_element($hidden, "Name", "#Hidden");
        }

        $self->{folders}->{$component->{_xpath}}=$folder;
        $self->{modified}=1;
        $new_folder=1;
    }
    else
    {
        $folder_dir=$self->get_folder_dir($folder, $component->{_xpath});
    }

    # Update or create the .directory file
    if ($new_folder or !$component->{intermediate} or
        !-f "$folder_dir/.directory")
    {
        if (!-d $folder_dir and !cxmkpath($folder_dir))
        {
            cxerr("unable to create the '$folder_dir' directory: $@\n");
        }

        # (Re)Create the corresponding directory file (from scratch)
        # kg_create_directory() takes care of everything, we just need
        # to figure out the filename
        CXMenu::kg_create_directory("$folder_dir/.directory", $component, 1);
    }

    return $folder;
}

sub delete_folder($$$)
{
    my ($self, $folder, $xpath)=@_;
    cxlog("delete_folder '$xpath'\n");

    # Delete the desktop files
    my $folder_dir=$self->get_folder_dir($folder, $xpath);
    if (-d $folder_dir)
    {
        cxlog("Deleting the '$folder_dir' directory\n");
        require File::Path;
        if (!File::Path::rmtree($folder_dir))
        {
            cxwarn("unable to delete the '$folder_dir' directory: $!\n");
        }
    }

    # Delete the <Menu> tag
    remove_element($folder->getParentNode(), $folder);
    delete $self->{folders}->{$xpath};
    $self->{modified}=1;
}


#####
#
# XDG XML file creation and parsing
#
#####

sub setup_default_merge_dir($)
{
    my ($self)=@_;
    my $merge_dir=$self->{menu};
    $merge_dir =~ s/\.menu$/-merged/;
    # Override the existing merge_dir if any
    $self->{merge_dir}=$merge_dir;
    push @{$self->{merge_dirs}}, $merge_dir;
}

sub get_xdg_config($$)
{
    my ($self)=@_;
    return if (defined $self->{root_name});

    if (-f "/etc/linspire-version" and
        -f "/etc/xdg/menus/applications-merged/los-programs.menu")
    {
        $self->{root_subname}="Programs";
        $self->{root_allowed}={};
    }

    # Read the master XML file to extract configuration information
    my $filename=$self->{menu};
    $filename=$self->{master_menu} if (!-f $filename);
    if (-f $filename)
    {
        my $xml=CXXMLDOM::parse_xml_file($filename);
        if ($xml)
        {
            my $root_menu=find_tag($xml, "Menu");
            my $child=$root_menu->getFirstChild();
            while (defined $child)
            {
                my $tag=$child->getNodeName();
                if ($tag eq "Name")
                {
                    $self->{root_name}=get_cdata($child);
                }
                elsif ($tag eq "DefaultMergeDirs")
                {
                    cxlog("found <DefaultMergeDirs>\n");
                    $self->setup_default_merge_dir();
                }
                elsif ($tag eq "MergeDir")
                {
                    my $merge_dir=get_cdata($child);
                    if ($merge_dir)
                    {
                        if ($merge_dir !~ m%^/%)
                        {
                            my $dir=cxdirname($self->{menu});
                            $merge_dir="$dir/$merge_dir";
                        }
                        if (!-d $merge_dir)
                        {
                            cxlog("the '$merge_dir' MergeDir does not exist. Ignoring it.\n");
                        }
                        elsif (!-w $merge_dir)
                        {
                            cxlog("the '$merge_dir' MergeDir is not writable. Ignoring it.\n");
                        }
                        else
                        {
                            if (!defined $self->{merge_dir})
                            {
                                $self->{merge_dir}=1;
                            }
                            push @{$self->{merge_dirs}}, $merge_dir;
                        }
                    }
                }
                elsif ($tag eq "Layout")
                {
                    cxlog("found <Layout>\n");
                    my $lspec=$child->getFirstChild();
                    while (defined $lspec)
                    {
                        $tag=$lspec->getNodeName();
                        if ($tag eq "Merge")
                        {
                            my $type=$lspec->getAttribute("type") || "";
                            cxlog("type=$type\n");
                            if ($type =~ /^(menus|all)$/i)
                            {
                                # We won't have trouble adding our menus
                                delete $self->{root_subname};
                                delete $self->{root_allowed};
                                last;
                            }
                        }
                        elsif ($tag eq "Menuname")
                        {
                            my $name=get_cdata($lspec);
                            if ($name)
                            {
                                if (!defined $self->{root_subname})
                                {
                                    $self->{root_subname}=$name;
                                }
                                $self->{root_allowed}->{$name}=1;
                            }
                        }
                        $lspec=$lspec->getNextSibling();
                    }
                    if ($self->{root_subname})
                    {
                        cxlog("sub root='$self->{root_subname}'\n");
                        cxlog("allowed root directories:\n  ",
                              join("\n  ", keys %{$self->{root_allowed}}), "\n");
                    }
                }
                $child=$child->getNextSibling();
            }
        }
    }
    $self->{root_name} ||= "Applications";
    $self->setup_default_merge_dir() if (!defined $self->{merge_dir});
}


sub init_xml($$)
{
    my ($self)=@_;

    return $self->{xml} if (defined $self->{xml_file});
    $self->get_xdg_config();
    $self->{xml_file}="$self->{merge_dir}/cxmenu-$self->{tag}.menu";

    my $xml;
    if (-f $self->{xml_file})
    {
        $xml=CXXMLDOM::parse_xml_file($self->{xml_file});
    }
    else
    {
        cxlog("Creating XDG menu file\n");
        my $xdg="<Menu>\n" .
                "  <Name>$self->{root_name}</Name>\n" .
                "</Menu>\n";
        $xml=CXXMLDOM::parse_xml_string($xdg);
    }
    return 0 if (!$xml);
    $self->{xml}=$xml;

    # Make sure we have a root menu
    $self->{root_menu}=find_tag($xml, "Menu");
    if (!$self->{root_menu})
    {
        $self->{root_menu}=add_new_element($xml, "Menu");
        add_new_element($self->{root_menu}, "Name", $self->{root_name});
    }

    return 1;
}


#####
#
# Main
#
#####

sub new($$$$$)
{
    my ($class, $cxoptions, $gui_info, $menu, $master_menu)=@_;

    my $data_dir=cxbasename($menu);
    $data_dir =~ s/\.menu$//;

    my $self={
        tag             => $cxoptions->{tag},
        desktopdata     => $cxoptions->{desktopdata},
        encoding        => "UTF-8",
        menu            => $menu,
        master_menu     => $master_menu,
        data_dir        => "xdg-$data_dir"
    };

    if (($gui_info->{kde_version} || "") =~ /^3\.5\./)
    {
        $self->{hidden_hack}=1;
    }

    bless $self, $class;
    return $self;
}

sub detect($$$$)
{
    my ($class, $cxoptions, $cxconfig, $gui_info)=@_;
    return () if (!$gui_info->{xdg_menu_on});

    # Check for XML::DOM support
    my $err=CXXMLDOM::get_xml_load_error();
    if ($err)
    {
        cxerr("unable to use XML::DOM:\n$err\n");
        return  ();
    }

    # Notes:
    # - CXMenuXDG does not handle the XDG desktop icons as these
    #   have nothing in common with the regular XDG menus.
    # - Also, each XDG menu instance could in theory share its files
    #   including the XML file with a simple symbolic link in the right place.
    #   However, each instance must also be independent. That is, it should be
    #   possible to disable an instance while still creating the others,
    #   and uninstalling a given menu in an instance should not delete the
    #   corresponding menu in the others. This means either introducing
    #   complex usage tracking (slowing things down too), or duplicating the
    #   menus and keeping the current relatively simple code at the cost of
    #   doing the same work twice when multiple instances are present.
    #   The latter solution was chosen.
    my @selves;
    for (my $i=1; exists $gui_info->{"xdg_preferred_menu$i"}; $i++)
    {
        push @selves, new($class, $cxoptions, $gui_info,
                          $gui_info->{"xdg_preferred_menu$i"},
                          $gui_info->{"xdg_global_menu$i"});
    }
    return @selves;
}

sub id($)
{
    my ($self)=@_;
    my $id="CXMenuXDG/$self->{menu}";
    $id =~ s%/+%/%g;
    return $id;
}

sub install($$)
{
    my ($self, $components)=@_;

    my $menu=@$components[-1];
    return 1 if ($menu->{is_desktop});
    return 0 if (!$self->init_xml());

    my ($folder, $xpath);
    foreach my $component (@$components)
    {
        # Xandros does not allow any extra top-level folders but makes an
        # exception for 'Windows Applications'. However if we mangle spaces,
        # then we get 'Windows+Applications' which does not match the Xandros
        # exception. The XML folder names don't have problems with spaces so
        # we simply demangle all spaces, just for the XML folder names.
        $component->{_xname}=mangle_string($component->{name});
        $component->{_xname}=~s/\+/ /g;
        if (!$folder)
        {
            $folder=$self->get_root_menu($component);
            $xpath=$component->{_xpath};
        }
        else
        {
            $xpath="$xpath/$component->{_xname}";
            $component->{_xpath}=$xpath;
        }
        $component->{_encoding}="UTF-8";
        $component->{_creator}="";
        last if (!$component->{is_dir});
        $folder=$self->create_folder($folder, $component);
        return 0 if (!defined $folder);
    }

    if ($menu->{is_dir})
    {
        # Make sure our brand new (maybe empty) folder will not be
        # 'garbage collected' by finalize()
        $self->{status}->{$menu->{_xpath}}="in-use";
    }
    else
    {
        my $xdir=cxdirname($menu->{_xpath});
        $self->{status}->{$xdir}="in-use";

        my $basename="$menu->{_xname}.desktop";
        # Remangle spaces. See create_folder() for details.
        $basename=~s/ /+/g;
        my $filetag;
        my $inctag=get_child($folder, "Include");
        if ($inctag)
        {
            $filetag=$inctag->getFirstChild();
            while (defined $filetag)
            {
                last if ($filetag->getNodeName() eq "Filename" and
                         get_cdata($filetag) eq $basename);
                $filetag=$filetag->getNextSibling();
            }
        }
        else
        {
            my $pos=get_child($folder, "Menu");
            $inctag=add_new_element($folder, "Include", undef, $pos);
            $self->{modified}=1;
        }
        if (!defined $filetag)
        {
            add_new_element($inctag, "Filename", $basename);
            $self->{modified}=1;
        }

        # Create the desktop file
        my $filename=$self->get_folder_dir($folder, $xdir);
        if (!-d $filename and !cxmkpath($filename))
        {
            cxerr("unable to create the '$filename' directory: $@\n");
        }
        $filename.="/$basename";
        # We don't need _escaping (neither for Gnome nor KDE)
        $menu->{_section_name}="Desktop Entry";
        $menu->{_creator}=" (" . $self->id() . ")" if ($ENV{CX_TAGALL});
        $menu->{_shebang}="/usr/bin/env xdg-open";
        CXMenu::kg_create_desktop_file($filename, $menu);
    }

    return 1;
}

sub query($$)
{
    my ($self, $components)=@_;
    return ($self->id(), "") if (!defined $components);

    my $menu=@$components[-1];
    return "" if ($menu->{is_desktop});
    return "0" if (!$self->init_xml());

    # Check that everything is in place in the XML file
    my $xpath=$menu->{is_dir} ? $menu->{path} : $menu->{dir};
    $xpath=$self->get_xpath($xpath);
    my $folder=$self->get_folder_by_xpath($xpath);
    return "" if (!$folder);

    my $folder_dir=$self->get_folder_dir($folder, $xpath);
    return "" if (!-d $folder_dir);

    # Then check the presence of the .desktop or .directory file
    if ($menu->{is_dir})
    {
        return $self->id() if (-f "$folder_dir/.directory");
    }
    else
    {
        my $basename=mangle_string($menu->{name}) . ".desktop";
        return "" if (!-f "$folder_dir/$basename");

        my $include=get_child($folder, "Include");
        if ($include)
        {
            my $filename=$include->getFirstChild();
            while (defined $filename)
            {
                if ($filename->getNodeName() eq "Filename" and
                    get_cdata($filename) eq $basename)
                {
                    return $self->id();
                }
                $filename=$filename->getNextSibling();
            }
        }
    }

    return "";
}

sub uninstall($$)
{
    my ($self, $components)=@_;

    my $menu=@$components[-1];
    return 1 if ($menu->{is_desktop});
    return 0 if (!$self->init_xml());

    my $xpath=$menu->{is_dir} ? $menu->{path} : $menu->{dir};
    $xpath=$self->get_xpath($xpath);
    my $folder=$self->get_folder_by_xpath($xpath);
    return 1 if (!$folder);

    # Always mark the parent folder for garbage collection
    my $xdir=$xpath;
    if ($menu->{is_dir})
    {
        $xdir=cxdirname($xdir);
    }
    $self->{garbage_collect}->{$xdir}=1;

    if ($menu->{is_dir})
    {
        $self->delete_folder($folder, $xpath);
    }
    else
    {
        my $basename=mangle_string($menu->{name}) . ".desktop";
        my $inctag=get_child($folder, "Include");
        if ($inctag)
        {
            my $filetag=$inctag->getFirstChild();
            while (defined $filetag)
            {
                if ($filetag->getNodeName() eq "Filename" and
                    get_cdata($filetag) eq $basename)
                {
                    remove_element($inctag, $filetag);
                    $self->{modified}=1;
                    last;
                }
                $filetag=$filetag->getNextSibling();
            }
        }

        my $folder_dir=$self->get_folder_dir($folder, $xpath);
        return 1 if (!-d $folder_dir);

        my $filename="$folder_dir/$basename";
        if (-f $filename)
        {
            cxlog("Deleting '$filename'\n");
            if (!unlink $filename)
            {
                cxerr("unable to delete '$filename': $!\n");
                return 0;
            }
        }
    }

    return 1;
}

sub removeall($$)
{
    my ($self, $pattern)=@_;
    $self->get_xdg_config();

    my $xml_pattern;
    if ($pattern eq "legacy")
    {
        my $legacy_dir="$ENV{CX_ROOT}/support/xdg-legacy-menus";
        if (-d $legacy_dir)
        {
            cxlog("Deleting the '$legacy_dir' directory\n");
            require File::Path;
            if (!File::Path::rmtree($legacy_dir))
            {
                cxerr("unable to delete the '$legacy_dir' directory: $!\n");
            }
        }
        $xml_pattern="^cxlegacy.menu\$";
    }
    else
    {
        $xml_pattern=$pattern;
        $xml_pattern.=".*" if ($xml_pattern !~ s/\$$//);
        $xml_pattern="^cxmenu-$xml_pattern\\.menu\$";
    }

    foreach my $dir (@{$self->{merge_dirs}})
    {
        next if (!-d $dir);
        if (!-w $dir)
        {
            cxlog("skipping read-only '$dir' directory\n");
            next;
        }
        CXUtils::delete_files($dir, $xml_pattern);
    }

    if ($self->{tag} and "cxmenu-$self->{tag}.menu" =~ /$xml_pattern/)
    {
        my $dir="$self->{desktopdata}/cxmenu/$self->{data_dir}";
        if (-d $dir)
        {
            cxlog("Deleting the '$dir' directory\n");
            require File::Path;
            if (!File::Path::rmtree($dir))
            {
                cxerr("unable to delete the '$dir' directory: $!\n");
            }
        }
        CXUtils::garbage_collect_subdirs($self->{desktopdata}, "/cxmenu", 1);
    }
    else
    {
        require CXBottle;
        CXBottle::removeall_desktopdata_dirs($pattern, "/cxmenu/$self->{data_dir}");
    }

    return 1;
}

sub gc_get_folder_status($$)
{
    my ($self, $xpath)=@_;
    my $folder=$self->get_folder_by_xpath($xpath);
    if (!$folder)
    {
        cxlog("folder '$xpath' has already been deleted\n");
        return undef;
    }

    # Check for sub-folders
    my $child=$folder->getFirstChild();
    while (defined $child)
    {
        if ($child->getNodeName() eq "Menu")
        {
            my $text=get_child($child, "Name");
            if (get_cdata($text) ne "#Hidden")
            {
                cxlog(" -> found a sub-folder\n");
                return ("in-use", $folder);
            }
        }
        $child=$child->getNextSibling();
    }
    cxlog(" -> no folders\n");

    # Check the folder's .desktop files
    my $folder_dir=$self->get_folder_dir($folder, $xpath);
    if (opendir(my $dh, $folder_dir))
    {
        foreach my $dentry (readdir $dh)
        {
            if ($dentry =~ /\.desktop$/)
            {
                closedir($dh);
                cxlog(" -> found a menu item\n");
                return ("in-use", $folder);
            }
        }
        closedir($dh);
    }

    cxlog(" -> empty\n");
    return ("empty", $folder);
}

sub gc_delete_folder($$$)
{
    my ($self, $xpath, $folder)=@_;
    $self->delete_folder($folder, $xpath);
}

sub finalize($)
{
    my ($self)=@_;
    return 1 if (!$self->{xml});

    $self->SUPER::finalize();
    CXUtils::garbage_collect_subdirs($self->{desktopdata},
                                     "/cxmenu/$self->{data_dir}", 1);

    if ($self->{modified})
    {
        if (find_tag($self->{xml}, "Menu"))
        {
            my $dir=cxdirname($self->{xml_file});
            if (!-d $dir and $dir !~ /kde/)
            {
                # Gnome is buggy and won't detect our menu file if it is
                # created at the same time as the directory. So rather than
                # introducing delays, we schedule a 'delete and recreate' of
                # the file for a bit later (a simple touch is not enough).
                $dir=shquote_string($dir);
                cxsystem("sleep 5 && touch $dir/cxmenu-gnome-bug-482845.menu && rm $dir/cxmenu-gnome-bug-482845.menu 2>/dev/null &");
            }
            return CXXMLDOM::save_xml_file($self->{xml_file}, $self->{xml},
                "<!DOCTYPE Menu PUBLIC \"-//freedesktop//DTD Menu 1.0//EN\" \"http://www.freedesktop.org/standards/menu-spec/1.0/menu.dtd\">\n");
        }
        else
        {
            # The file is empty so let's just delete it
            if (-f $self->{xml_file})
            {
                cxlog("Deleting '$self->{xml_file}'\n");
                if (!unlink $self->{xml_file})
                {
                    cxerr("unable to delete '$self->{xml_file}': $!\n");
                    return 0;
                }
            }
        }
    }
    else
    {
        cxlog("XML file '$self->{xml_file}' not modified\n");
    }
    return 1;
}

return 1;
