# (c) Copyright 2009. CodeWeavers, Inc.

import gobject

import bottlequery
import bottlewrapper
import bottlemanagement

import pyop

# for localization
from cxutils import cxgettext as _

ALL_INFO = ("basic", "controlpanel", "applications")

def sharedCollection(infoToTrack=()):
    global sharedBottleCollection
    try:
        sharedBottleCollection.track_info(infoToTrack)
    except NameError:
        sharedBottleCollection = BottleCollection(infoToTrack)

    return sharedBottleCollection
    

def waitForSharedCollectionToBeReady(inFunction, infoToTrack):
    collection = sharedCollection(infoToTrack)
    for bottleName in collection.bottleList():
        bottleObj = collection.bottleObject(bottleName)
        if ('basic' in infoToTrack and not bottleObj.bottle_info_ready or
            'controlpanel' in infoToTrack and not bottleObj.control_panel_ready or
            'applications' in infoToTrack and not bottleObj.installed_packages_ready):
            gobject.timeout_add(10, waitForSharedCollectionToBeReady, inFunction, infoToTrack)
            return()

    inFunction()


class BottleCollection(object):

    bottleCollectionDict = None
    changeDelegate = None
    bottleChangeDelegate = None

    def __init__(self, infoToTrack=ALL_INFO):
        self.bottleCollectionDict = {}
        self.ignored_bottles = []
        self._infoToTrack = set(infoToTrack)
        self.refresh()
        # Set up a timer so we can poll for newly-created bottles.
        gobject.timeout_add(2000, self.refresh)

    def queue_bottle_updates(self, bottles, info):
        if 'basic' in info:
            for bottle in bottles:
                updateOp = RefreshBasicInfoOperation(bottle)
                pyop.sharedOperationQueue.enqueue(updateOp)
        if 'applications' in info:
            for bottle in bottles:
                updateOp = RefreshApplicationsOperation(bottle)
                pyop.sharedOperationQueue.enqueue(updateOp)
        if 'controlpanel' in info:
            for bottle in bottles:
                updateOp = RefreshControlPanelInfoOperation(bottle)
                pyop.sharedOperationQueue.enqueue(updateOp)

    def refresh(self):
        bottleList = bottlequery.get_bottle_list()
        changes = False
        removeList = []

        # Remove deleted bottles
        for bottleName in self.bottleCollectionDict:
            if bottleName not in bottleList:
                changes = True
                removeList.append(bottleName)

        for bottleName in removeList:
            self._removeBottle(bottleName)

        # Check for updates
        bottles_to_update = []
        for bottleName in self.bottleCollectionDict:
            if self.bottleCollectionDict[bottleName].get_needs_update(updating = True):
                bottles_to_update.append(self.bottleCollectionDict[bottleName])

        # Add new bottles
        for bottleName in bottleList:
            if bottleName not in self.bottleCollectionDict and \
               bottleName not in self.ignored_bottles:
                self._addBottle(bottleName)
                self.bottleCollectionDict[bottleName].get_needs_update(updating = True)
                bottles_to_update.append(self.bottleCollectionDict[bottleName])
                changes = True

        if bottles_to_update:
            # FIXME:  We don't really need to reload 'basic' info
            #         if this bottle was just added.
            self.queue_bottle_updates(bottles_to_update, self._infoToTrack)

        if self.changeDelegate and changes:
            self.changeDelegate.bottleCollectionChanged()

        return True

    def track_info(self, new_info):
        if not self._infoToTrack.issuperset(new_info):
            new_info = set(new_info)
            new_info.difference_update(self._infoToTrack)
            self._infoToTrack.update(new_info)
            self.queue_bottle_updates(self.bottleCollectionDict.values(), self._infoToTrack)

    def add_ignored_bottle(self, newname):
        self.ignored_bottles.append(newname)

    def remove_ignored_bottle(self, newname):
        self.ignored_bottles.remove(newname)

    def refreshDefaultBottle(self):
        self.defaultBottleName = bottlequery.get_default_bottle()

    def getDefaultBottle(self):
        return self.defaultBottleName

    def bottleObject(self, inBottleName):
        return self.bottleCollectionDict[inBottleName]

    def bottleList(self):
        return self.bottleCollectionDict.keys()

    def bottles(self):
        return self.bottleCollectionDict.values()

    def bottleCount(self):
        return len(self.bottleCollectionDict)

    def _addBottle(self, inBottleName):
        bottleobj = bottlewrapper.BottleWrapper(inBottleName)
        self.bottleCollectionDict[inBottleName] = bottleobj
        if self.bottleChangeDelegate:
            bottleobj.change_delegate = self.bottleChangeDelegate

        self.refreshDefaultBottle()

        return bottleobj

    def _removeBottle(self, inBottleName):
        self.bottleCollectionDict.pop(inBottleName)
        self.refreshDefaultBottle()

    def setChangeDelegate(self, inDelegate):
        self.changeDelegate = inDelegate

    def setBottleChangeDelegate(self, inDelegate):
        self.bottleChangeDelegate = inDelegate
        for bottleKey in self.bottleCollectionDict.keys():
            self.bottleCollectionDict[bottleKey].change_delegate = inDelegate
        
    def archiveBottle(self, inBottleName, inArchivePath):
        bottle_obj = self.bottleObject(inBottleName)
        bottle_obj.add_status_override(_("Archiving"))
        try:
            rVal = bottlemanagement.archive_bottle(inBottleName, inArchivePath)
        finally:
            bottle_obj.remove_status_override(_("Archiving"))
        return rVal

class ArchiveBottleOperation(pyop.PythonOperation):

    def __init__(self, inBottleName, inArchivePath):
        pyop.PythonOperation.__init__(self)
        self.bottleName = inBottleName
        self.archivePath = inArchivePath

    def __unicode__(self):
        return "ArchiveBottleOperation for " + self.bottleName


    def main(self):
        sharedCollection().archiveBottle(self.bottleName, self.archivePath)
        

class RefreshBasicInfoOperation(pyop.PythonOperation):
    def __init__(self, inBottle):
        pyop.PythonOperation.__init__(self)
        self.bottleObject = inBottle

    def __unicode__(self):
        return " ".join("RefreshBasicInfoOperation for" + repr(self.bottleObject))

    def enqueued(self):
        pyop.PythonOperation.enqueued(self)
        self.bottleObject.pre_load_basic_info()

    def main(self):
        self.bottleObject.load_basic_info()

    def finish(self):
        self.bottleObject.post_load_basic_info()
        pyop.PythonOperation.finish(self)

class RefreshControlPanelInfoOperation(pyop.PythonOperation):
    def __init__(self, inBottle):
        pyop.PythonOperation.__init__(self)
        self.bottleObject = inBottle

    def __unicode__(self):
        return " ".join("RefreshControlPanelInfoOperation for" + repr(self.bottleObject))

    def enqueued(self):
        pyop.PythonOperation.enqueued(self)
        self.bottleObject.pre_load_control_panel_info()

    def main(self):
        self.bottleObject.load_control_panel_info()
    
    def finish(self):
        self.bottleObject.post_load_control_panel_info()
        pyop.PythonOperation.finish(self)

class RefreshApplicationsOperation(pyop.PythonOperation):
    def __init__(self, inBottle):
        pyop.PythonOperation.__init__(self)
        self.bottleObject = inBottle

    def __unicode__(self):
        return " ".join("RefreshApplicationsOperation for" + repr(self.bottleObject))

    def enqueued(self):
        pyop.PythonOperation.enqueued(self)
        self.bottleObject.pre_load_installed_applications()

    def main(self):
        self.bottleObject.load_installed_applications()
    
    def finish(self):
        self.bottleObject.post_load_installed_applications()
        pyop.PythonOperation.finish(self)


class QuitBottleOperation(pyop.PythonOperation):

    def __init__(self, inBottleList, inForce=False, inFinishDelegate=None):
        pyop.PythonOperation.__init__(self)
        self.bottleList = inBottleList
        self.force = inForce
        self.finishDelegate = inFinishDelegate

    def __unicode__(self):
        return " ".join(["QuitBottleOperation for"] + self.bottleList)

    def main(self):
        for bottleName in self.bottleList:
            bottleObj = sharedCollection().bottleObject(bottleName)
            if self.force:
                bottleObj.force_quit()
            else:
                bottleObj.quit()
            bottleObj.cancelShutdown()


    def finish(self):
        if self.finishDelegate:
            self.finishDelegate.opFinished(self)
        pyop.PythonOperation.finish(self)


