# (c) Copyright 2009. CodeWeavers, Inc.

"""Handles refreshing and reading the C4 profile information."""

import os
import threading
import time
import UserDict

import cxobjc
import cxlog
import cxutils

import c4parser


UNKNOWN = "com.codeweavers.unknown"
# This is the application id of the unknown application profile


class _C4PSource(object):
    """This is a helper class for C4ProfilesSet.

    It keeps track of a couple of extra data about a C4Pfile object: its old
    list of profiles and its last modification time.

    The 'trusted' flag marks profiles that are official
    CodeWeavers profiles.  Untrusted profiles will produce a 
    user warning upon selection.  Ideally this will be handled via 
    signing but at the moment it is based on where the files came from.
    """

    def __init__(self, c4pfile, trusted = True):
        self.c4pfile = c4pfile
        self.profiles = {}
        self.mtime = 0
        self.trusted = trusted


class C4Profiles(cxobjc.Proxy):
    pass

@cxobjc.method(C4Profiles, 'profilesSet')
def profilesSet():
    return C4ProfilesSet.all_profiles()

@cxobjc.method(C4Profiles, 'addC4PFile_trusted_')
def addC4PFile_trusted_(c4pfilename, trusted=True):
    return C4ProfilesSet.add_c4p_file(c4pfilename.encode('utf8'), trusted)

class C4ProfilesSet(UserDict.IterableUserDict):
    """A mapping of application ids to the corresponding C4Profile objects."""

    #####
    #
    # Instance methods
    #
    #####

    def __init__(self):
        """Adds a data mapping attribute that third parties can use to store
        or cache data related to this profiles set.
        """
        UserDict.IterableUserDict.__init__(self)
        self.userdata = {}


    def installer_profiles(self, appid):
        """Returns the application's installer profiles if any, and the
        unknown application installer profiles otherwise.
        """
        installers = None
        if appid in self:
            profile = self[appid]
            if profile.app_profile and profile.app_profile.installer_id:
                try:
                    profile = self[profile.app_profile.installer_id]
                except KeyError:
                    profile = None
            if profile:
                installers = profile.installer_profiles
        if not installers:
            installers = self[UNKNOWN].installer_profiles
        return installers

    installers_ = installer_profiles

    def unknown_installer(self):
        return self[UNKNOWN].installer_profiles[0]

    def contentsDict(self):
        return self.data


    #####
    #
    # Class methods and data
    #
    #####

    _lock = threading.Lock()
    # To avoid races while updating _sources.

    _sources = {}
    # Keep track of the files the profiles come from. This is so we know when
    # to refresh the profiles set, and also so we know to delete a given
    # profile if the file goes away. This is a mapping so each file is present
    # only once (and for now order does not matter).

    _last_checked = 0
    # The timestamp of the last time the freshness of the _profiles data
    # was checked

    _profiles = None
    # The cached profiles set. This variable is updated atomically, so this
    # set can be returned at any time without locking.


    @classmethod
    def add_c4p_file(cls, c4pfilename, trusted=True):
        """Reads the specified c4p file and add its profiles to the global
        database.

        The c4p file is then watched for updates and re-read if it is modified.
        This is mostly meant for telling the current process to use the
        contents of a standalone c4p file and thus is not persistent.
        """
        c4pfile = c4parser.C4PFile(c4pfilename, trusted)
        cls._sources[c4pfile.filename] = _C4PSource(c4pfile, trusted)
        cls._last_checked = 0
        return c4pfile


    @classmethod
    def all_profiles(cls):
        """Returns a mapping of application ids to the corresponding C4Profile
        objects.

        This automatically reads or re-reads the relevant c4p files if needed
        but will always return a consistent view of the underlying c4p profile
        database.
        """
        # The consistency guarantee means that even if we split the local C4
        # database across many different c4p files, and these are in various
        # stages of being updated, the returned profiles will not mix old data
        # with new data.
        if cls._profiles is None or cls._last_checked + 2 < time.time():
            cls._lock.acquire()
            try:
                # Check if one of the source files changed
                changes = False
                for source in cls._sources.itervalues():
                    if source.c4pfile.update():
                        changes = True
                        break

                if changes:
                    # If a profile has changed for a given appid in one of the
                    # source files, then we need to reevaluate whether it's the
                    # profile that best matches the current product for that
                    # appid, or whether a profile from another source file
                    # should be used instead.
                    # The simplest way to do so, and possibly fastest too, is
                    # to rebuild the list from scratch.
                    profiles = C4ProfilesSet()
                    for source in cls._sources.itervalues():
                        try:
                            # Also refreshes that source's profile list
                            profile_list = source.c4pfile.profiles
                            if profile_list is None:
                                # If the file does not exist, then we get
                                # None with no exception
                                continue
                        # pylint: disable-msg=W0703
                        except Exception, exception:
                            cxlog.err("an error occurred while reading %s:\n%s" % (cxlog.debug_str(filename), cxlog.debug_str(exception)))
                            continue

                        for profile in profile_list:
                            appid = profile.appid
                            if appid not in profiles:
                                profiles[appid] = profile
                            elif profile.score < profiles[appid].score:
                                profiles[appid] = profile

                    # Atomically update cls._profiles
                    cls._profiles = profiles
                    cls._last_checked = time.time()
            finally:
                cls._lock.release()
        return cls._profiles


for basename in ("crossover.c4p", "unknown.c4p"):
    filename = os.path.join(cxutils.CX_ROOT, "share/crossover/data", basename)
    C4ProfilesSet.add_c4p_file(filename)
