# (c) Copyright 2010. CodeWeavers, Inc.

import os.path

import cxconfig
import cxutils
import distversion


#####
#
# The current product id
#
#####

_PRODUCT_ID = None
def get_product_id():
    """Gets and checks the effective product id.

    An invalid product id is a fatal error and will raise an exception.
    """
    # pylint: disable-msg=W0603
    global _PRODUCT_ID
    if _PRODUCT_ID is None:
        filename = os.path.join(cxutils.CX_ROOT, ".productid")
        if os.path.exists(filename):
            try:
                thisfile = open(filename, 'r')
                _PRODUCT_ID = thisfile.readline().rstrip()
                thisfile.close()
            except OSError, ose:
                raise RuntimeError("unable to open, read or close '%s': %s" % (filename, unicode(ose)))
        else:
            _PRODUCT_ID = distversion.BUILTIN_PRODUCT_ID
        if len(_PRODUCT_ID) < 4:
            raise RuntimeError("product id '%s' is too short" % _PRODUCT_ID)
        if not _PRODUCT_ID.isalnum():
            raise RuntimeError("product id '%s' contains bad characters" % _PRODUCT_ID)
    return _PRODUCT_ID


#####
#
# CrossOver's important directories
#
#####

def get_user_dir():
    if distversion.IS_MACOSX:
        return os.path.join(os.environ["HOME"], "Library/Application Support/", distversion.PRODUCT_NAME)
    else:
        return os.path.join(os.environ["HOME"], "." + get_product_id())


def get_managed_dir():
    if distversion.IS_MACOSX:
        return os.path.join("/Library/Application Support", distversion.PRODUCT_NAME)
    else:
        return os.path.join(cxutils.CX_ROOT, "support")

def get_bottle_path():
    # FIXME: This differs from the perl code.
    config = get_ro_config()
    path = config["EnvironmentVariables"].get("CX_BOTTLE_PATH")
    if path:
        return path
    try:
        return os.environ["CX_BOTTLE_PATH"]
    except KeyError:
        if distversion.IS_MACOSX:
            return os.path.join(get_user_dir(), "Bottles")
        else:
            return get_user_dir()

def get_managed_bottle_path():
    # FIXME: This differs from the perl code.
    config = get_ro_config()
    path = config["EnvironmentVariables"].get("CX_MANAGED_BOTTLE_PATH")
    if path:
        return path
    try:
        return os.environ["CX_MANAGED_BOTTLE_PATH"]
    except KeyError:
        if distversion.IS_MACOSX:
            return os.path.join(get_managed_dir(), "Bottles")
        else:
            return get_managed_dir()


#####
#
# CrossOver's configuration
#
#####

def get_ro_config():
    """Returns a read-only object containing the CrossOver configuration,
    merged from the relevant sources as needed.
    """
    basename = get_product_id() + ".conf"
    config = cxconfig.CXMergeConfig()

    global_config_file = os.path.join(cxutils.CX_ROOT, 'etc', basename)
    config.read(global_config_file)

    user_config_file = os.path.join(get_user_dir(), basename)
    config.read(user_config_file)

    return config

def get_rw_config():
    """Opens the CrossOver configuration file for modification.

    To avoid races when changing the configuration file it is necessary to
    follow these rules:
    - The data returned by a get_ro_config() call made prior to
      get_rw_config() must not be trusted. This is because the configuration
      file was not yet locked and thus may have changed between the two calls.
    - The data returned by get_rw_config() must not be trusted. This is
      because the CXRWConfig object does not aggregate the data from all the
      configuration files. So it is missing any setting not present in the file
      where we save modifications.
    - So the proper way to test the value of a setting and then modify it
      race-free is:
          cxrwconfig = crossover.get_rw_config()
          cxconfig = crosover.get_ro_config()
          if cxconfig['section']['field'] ...:
              cxrwconfig['section']['field'] = ...
          cxrwconfig.save()
          cxrwconfig.unlock()
          cxrwconfig = None       # Should not be reused once unlocked
          cxconfig = None         # Does not contain the modified setting
    """
    basename = get_product_id() + ".conf"
    config_file = os.path.join(get_user_dir(), basename)
    if not os.access(config_file, os.F_OK):
        global_config_file = os.path.join(cxutils.CX_ROOT, 'etc', basename)
        if os.path.isfile(global_config_file) and \
                os.access(global_config_file, os.R_OK | os.W_OK):
            config_file = global_config_file

    config = cxconfig.CXRWConfig()
    config.lock_and_read(config_file)
    return config

def save_setting(section, field, value):
    """Saves the specified setting in the CrossOver configuration file.

    This function must not be used if value is based on  another configuration
    setting as this would have races.
    """
    config = get_rw_config()
    config[section][field] = value
    config.save()
    config.unlock()
