# (c) Copyright 2009-2010. CodeWeavers, Inc.

import errno
import os
import os.path
import shutil
import zipfile
import tarfile
import tempfile

import cxlog
import cxutils
import globtree

import bottlequery
import cxaiebase
import cxaiemedia
import appdetector
import distversion


# for localization
from cxutils import cxgettext as _

#####
#
# The core application installation task
#
#####

class AIECore(cxaiebase.AIETask):
    """Installs a specific application."""

    #####
    #
    # Limit concurrency
    #
    #####

    def _getniceness(self):
        """Start downloads before all else."""
        return self.depth

    niceness = property(_getniceness)


    def can_run(self):
        """Returns True if this task is runnable and no other instance of it
        is running.
        """
        if 'AIECore' in self.scheduler.userdata:
            return False
        return cxaiebase.AIETask.can_run(self)


    def schedule(self):
        cxaiebase.AIETask.schedule(self)
        self.scheduler.userdata['AIECore'] = self


    def done(self):
        cxaiebase.AIETask.done(self)
        try:
            del self.scheduler.userdata['AIECore']
        except KeyError:
            cxlog.err(cxlog.to_str(self) + ": AIECore should have been set")


    #####
    #
    # Initialization
    #
    #####

    def __init__(self, engine, installer):
        cxaiebase.AIETask.__init__(self, engine, _("Installing %s") % installer.parent.name)
        self.installer = installer

        # Related objects to download the application or handle CDs
        self.download = None
        self.insert_media = None
        self.scan_media = None
        self.release_media = None

        # Nearest AIECore objects with an associated media
        self.parent_media = set()

        # depth is 0 for the top profile, -1 for those it depends on, etc.
        # Note that depth always reflects the depth of the longuest path from
        # the top-level profile.
        self.depth = 0

        # Stores the state information used at run time
        self.state = {}


    def _getname(self):
        """A helper function that returns the associated profile name."""
        return self.installer.parent.name

    name = property(_getname)


    def __unicode__(self):
        return "%s(%s)" % (self.__class__.__name__, self.name)


    def create_download_task(self):
        self.download = cxaiemedia.AIEDownload(self)


    def create_media_tasks(self):
        self.insert_media = cxaiemedia.AIEInsertMedia(self)
        self.scan_media = cxaiemedia.AIEScanMedia(self)
        self.scan_media.add_dependency(self.insert_media)
        self.release_media = cxaiemedia.AIEReleaseMedia(self)
        self.release_media.add_dependency(self.scan_media)


    def run_routines(self, prefix, routines):
        if 'iter' in self.state:
            iterator = self.state['routine_iterator']
        else:
            iterator = iter(routines)
            self.state['routine_iterator'] = iterator
        for routine in iterator:
            skip_routine = 'no_' + routine
            if skip_routine in self.state:
                continue
            try:
                func = AIECore.__dict__['_%s_%s' % (prefix, routine)]
            except KeyError:
                cxlog.err('AIECore._%s_%s does not exist!' % (cxlog.to_str(prefix), cxlog.to_str(routine)))
                continue
            cxlog.log_("aie", "  running " + cxlog.to_str(routine))
            if not func(self):
                cxlog.log_("aie", "  -> returned False")
                if not self.error:
                    cxlog.err(cxlog.to_str(routine) + ' did not set self.error')
                    self.error = _("An unknown error occurred in %s") % routine
                return False

        del self.state['routine_iterator']
        return True



    #####
    #
    # Prepare
    #
    #####

    def _prep_all_fonts(self):
        cxlog.log_('aie', 'appid=' + cxlog.to_str(self.installer.parent.appid))
        if self.installer.parent.appid == 'com.codeweavers.c4.6959':
            for font in self.dependencies:
                if isinstance(font, AIECore):
                    font.state['silentfontinstall'] = True
                    cxlog.log_('aie', 'set silentfontinstall on ' + cxlog.to_str(font.name))
        return True


    def _prep_collate_skips(self):
        engine = self.scheduler
        if 'skipassoc' not in engine.state:
            engine.state['skipassoc'] = True
            engine.state['skipmenu'] = True
            # These are only set when the corresponding task succeeds though
            engine.state['defaulteassocs'] = set()
            engine.state['alteassocs'] = set()
        engine.state['skipassoc'] &= self.installer.skip_assoc_creation
        engine.state['skipmenu'] &= self.installer.skip_menu_creation
        return True


    def _prep_collate_nsplugins(self):
        engine = self.scheduler
        if 'pre_nsplugins' not in engine.state:
            engine.state['pre_nsplugins'] = set()
            # This one is only set when the corresponding task is run though
            engine.state['ignore_nsplugins'] = set()
        engine.state['pre_nsplugins'].update(self.installer.pre_install_nsplugin_dlls)
        cxlog.log_("aie", "collate_nsplugins %s" % cxlog.to_str(engine.state['pre_nsplugins']))
        return True


    def prepare(self):
        """This method is called by the install engine after all the
        AIECore tasks have been created. The calls are made on children before
        the tasks that depend on them. They are not allowed to fail.
        """
        routines = ('all_fonts',
                    'collate_skips',
                    'collate_nsplugins',
                    )
        return self.run_routines('prep', routines)



    #####
    #
    # Main
    #
    #####

    def _main_set_installer_file(self):
        """Check that we have an installation source and initialize
        installer_file as appropriate.
        """
        install_source = self.state.get('install_source')
        if 'virtual' in self.installer.parent.app_profile.flags:
            # Virtual profiles are not supposed to have an installation source
            if install_source:
                cxlog.warn("install_source = %s for the %s virtual profile!" % (cxlog.debug_str(install_source), cxlog.debug_str(self.name)))
                del self.state['install_source']
            return True
        if not install_source:
            self.error = _("No installation source was provided for %s") % self.name
            return False

        if os.path.isfile(install_source):
            self.state['installer_file'] = install_source
            self.state['install_source'] = os.path.dirname(install_source)
        elif not os.path.isdir(install_source):
            self.error = _("The '%(source)s' installation source of %(name)s is neither a file nor a directory!") % {'source':install_source, 'name':self.name}
            return False

        return True


    def _main_setup_environment(self):
        """Set up the Unix environment to be used by this task's child
        processes.
        """
        env = self.state['environ'] = self.scheduler.state['environ'].copy()
        for envvar in self.installer.installer_environment:
            env[envvar.name] = self.scheduler.expand_string(envvar.value)

        # Defer the menu creation until we run cxmenu --sync
        env['CX_NO_WINESHELLLINK'] = '1'

        # Prevent the application installers from scanning all of the user's
        # hard drive.
        env['CX_HACK_REMOTE_DRIVES'] = 'yz'
        return True


    def _main_setup_registry_pre(self):
        """Tweak the windows registry for child processes.

        Do it early enough so that it's ready self-extracting installers.
        """
        for regentry in self.installer.pre_install_registry:
            for regvalue in regentry.values:
                if not bottlequery.set_registry_key(self.scheduler.installtask.bottlename, regentry.key, regvalue.name, regvalue.data):
                    return False
        return True


    def _main_uncompress(self):
        """Check if the installer file is an archive.

        If so uncompress it into a temporary directory and arrange for it to
        be used as the install_source in the following steps. This installers
        can be compressed and it's transparent to the later steps. Also arrange
        for the temporary directory to be deleted at the end.
        """
        if 'install_source' not in self.state:
            # Nothing to do for virtual packages
            return True

        installer_file = self.state.get('installer_file')
        if installer_file is None:
            # Nothing to uncompress. We are done here.
            return True

        archive_type = None
        if self.installer.installer_treatas:
            lower = self.installer.installer_treatas.lower()
        else:
            lower = installer_file.lower()
        if lower.endswith('.zip'):
            archive_type = 'zip'
        elif lower.endswith('.cab'):
            archive_type = 'cab'
        elif lower.endswith('.tgz'):
            archive_type = 'tar'
        elif lower.endswith('.tar.gz'):
            archive_type = 'tar'
        elif lower.endswith('.tar.bz2'):
            archive_type = 'tar'
        elif lower.endswith('.tbz'):
            archive_type = 'tar'
        elif lower.endswith('.tb2'):
            archive_type = 'tar'
        elif self.installer.selfextract_threshold is not None:
            try:
                if os.path.getsize(installer_file) >= self.installer.selfextract_threshold * 1024:
                    archive_type = 'selfextract'
            except OSError, ose:
                cxlog.log_("aie", cxlog.to_str(ose))
        if archive_type is None:
            # Not an archive. We are done here.
            return True

        # Get a temporary directory to extract into
        wintmpdir = self.scheduler.expand_string("%temp%")
        tmpdir = bottlequery.get_native_path(self.scheduler.installtask.bottlename, wintmpdir)
        if not tmpdir:
            self.error = _("Unable to get a temporary directory for %s") % self.name
            return False
        tmpdir = tempfile.mkdtemp(dir=tmpdir)
        self.state['temporary_source'] = tmpdir

        # Zip file support
        if archive_type == 'zip':
            zipf = zipfile.ZipFile(installer_file, 'r')
            for source in zipf.namelist():
                target = os.path.join(tmpdir, source)
                if target.endswith('\\') or target.endswith('/'):
                    os.makedirs(target)
                else:
                    data = zipf.read(source)
                    dfile = open(target, 'wb')
                    dfile.write(data)
                    dfile.close()
            zipf.close()

        # Microsoft Cabinet file support
        if archive_type == 'cab':
            engine = self.scheduler
            cmd = [os.path.join(cxutils.CX_ROOT, "bin", "wine"),
                   "--bottle", engine.installtask.bottlename,
                   "--wait-children", "--wl-app", "extract.exe", "--",
                   "/E", "/L", tmpdir, "/R", installer_file]

            # FIXME: Make interruptible.
            retcode, _out, err = cxutils.run(cmd, env=engine.state['environ'],
                                             stderr=cxutils.GRAB)
            if retcode:
                self.error = err
                return False

        # Tar file support
        if archive_type == 'tar':
            tarf= tarfile.open(installer_file, 'r')
            tarf.extractall(tmpdir)
            tarf.close()

        # Self-extracting executable
        if archive_type == 'selfextract':
            engine = self.scheduler

            # Set %ExtractDir% in the 'Windows' environment variable block so
            # it is available when building the self-extracting installer
            # command. This way it will know where to extract the files.
            # Note that we need to put a Windows path there. We cannot rely on
            # the wine script doing the conversion for us because the path is
            # likely to be embedded in an option like so '/T:%ExtractDir%'.
            # Also note that we modify a variable which may be used by the
            # other tasks. That's ok because no other task is going to use
            # 'winenv' while AIECore is running.
            if not wintmpdir.endswith('\\'):
                wintmpdir += '\\'
            winenv = engine.get_win_environ()
            # The key must be lower case
            winenv['extractdir'] = wintmpdir + os.path.basename(tmpdir)

            cmd = [os.path.join(cxutils.CX_ROOT, "bin", "wine"),
                   "--bottle", engine.installtask.bottlename,
                   "--wait-children", installer_file]
            for option in self.installer.selfextract_options:
                cmd.append(engine.expand_string(option))
            if self.state.get('silentinstall'):
                for option in self.installer.selfextract_silent_options:
                    cmd.append(engine.expand_string(option))
            del winenv['extractdir']

            # FIXME: Make interruptible.
            retcode, _out, err = cxutils.run(cmd, env=engine.state['environ'],
                                             stderr=cxutils.GRAB)
            if retcode:
                self.error = err
                return False

        self.state['install_source'] = tmpdir
        del self.state['installer_file']
        return True


    def _main_locate_installer(self):
        if not cxaiemedia.needs_installer_file(self.installer):
            # Virtual and 'copy-only' profiles don't need an installer file
            # and should skip the corresponding steps.
            self.state['no_installer'] = True
            return True
        if 'installer_file' in self.state:
            # Nothing to do if we already have an installer file
            return True

        install_source = self.state['install_source']
        if os.path.isdir(install_source):
            installer_file = cxaiemedia.locate_installer(self.installer, install_source)
            if installer_file:
                self.state['installer_file'] = installer_file
                cxlog.log_("aie", "installer_file = " + cxlog.debug_str(installer_file))
                return True
        self.error = _("Unable to find an installer for %(name)s in '%(source)s'") % {'name':self.name, 'source':install_source}
        return False


    def _main_build_install_command(self):
        if 'no_installer' in self.state:
            return True
        if 'installer_command' in self.state:
            # Nothing to do if we already have a command
            return True

        installer_file = self.state['installer_file']
        cmd = [os.path.join(cxutils.CX_ROOT, "bin", "wine"),
               "--bottle", self.scheduler.installtask.bottlename,
               "--untrusted", "--wait-children"]
        if self.installer.installer_dlloverrides:
            cmd.extend(['--dll', self.installer.installer_dlloverrides])
        if self.installer.installer_winver:
            cmd.extend(['--winver', self.installer.installer_winver])
        cmd.append('--')

        if self.installer.installer_treatas:
            lower = self.installer.installer_treatas.lower()
        else:
            lower = installer_file.lower()
        if lower.endswith('/autorun.inf'):
            cmd.extend(['autorun.exe', os.path.dirname(installer_file)])

        elif lower.endswith('.exe') or lower.endswith('.com') or lower.endswith('.bat') or lower.endswith('.cmd'):
            cmd.append(installer_file)
            for option in self.installer.installer_options:
                cmd.append(self.scheduler.expand_string(option))
            if self.state.get('silentinstall'):
                for option in self.installer.installer_silent_options:
                    cmd.append(self.scheduler.expand_string(option))

        elif lower.endswith('.msi'):
            cmd.extend(['msiexec.exe', '/i', installer_file])

        elif lower.endswith('.msp'):
            cmd.extend(['msiexec.exe', '/p', installer_file])

        elif lower.endswith('.otf') or \
             lower.endswith('.ttc') or \
             lower.endswith('.ttf'):
            cmd.extend(['cxinstallfonts.exe', installer_file])

        else:
            self.error = _("'%(file)s' is of an unknown installer type for %(name)s") % {'file':installer_file, 'name':self.name}
            return False

        self.state['installer_command'] = cmd
        cxlog.log_("aie", "installer_command=" + ' '.join(cxlog.debug_str(x) for x in cmd))
        return True


    def _main_collate_eassocs(self):
        engine = self.scheduler
        engine.state['defaulteassocs'].update(self.installer.default_eassocs)
        engine.state['alteassocs'].update(self.installer.alt_eassocs)
        return True


    def _main_collate_nsplugins(self):
        engine = self.scheduler
        engine.state['ignore_nsplugins'].update(self.installer.ignore_nsplugin_dlls)
        return True


    def _main_run_installer_and_check(self):
        if 'no_installer' in self.state:
            return True
        # We cannot trust the installer's return code so we ignore it entirely.
        # FIXME: We should try to be interruptible
        cxutils.system(self.state['installer_command'],
                       env=self.state['environ'])

        profile = self.installer.parent
        if not profile.app_profile.installed_key_pattern and \
           not profile.app_profile.installed_display_pattern and \
           not profile.app_profile.installed_registry_globs and \
           not profile.app_profile.installed_file_globs:
            # Bad news... we have no idea if this installed or not.
            cxlog.log_("aie", "We have no way to detect success or failure for this package. Returning True, with optimism.")
            return True
        if appdetector.is_profile_installed(self.scheduler.installtask.bottlename, profile):
            return True
        self.error = _("The installer has exited but %s does not seem to be installed") % self.name
        return False


    def _main_remove_fake_dlls(self):
        engine = self.scheduler
        for dllfile in self.installer.pre_rm_fake_dlls:
            source = engine.expand_string("%WinSysDir%")
            target = source
            source = os.path.join(source, dllfile)
            source = bottlequery.get_native_path(engine.installtask.bottlename, source)
            if os.path.exists(source):
                sfile = open(source,"rb")
                sfile.seek(0x40)
                tag = sfile.read(20)
                if tag == "Wine placeholder DLL":
                    target = os.path.join(target, dllfile + ".bak")
                    target = bottlequery.get_native_path(engine.installtask.bottlename, target)
                    if os.path.exists(target):
                        os.remove(target)
                    shutil.move(source, target)
                else:
                    cxlog.log(" %s is not a fake dll " % (cxlog.debug_str(dllfile)))
        return True


    def _main_copy_files(self):
        if 'install_source' not in self.state:
            # Nothing to do for virtual packages
            return True

        engine = self.scheduler
        globber = globtree.FileGlobTree()
        for copyentry in self.installer.files_to_copy:
            globber.add_glob(engine.expand_string(copyentry.glob), copyentry)

        matches = {}
        install_source = self.state['install_source']
        for match, copyentry in globber.matches(install_source):
            if copyentry in matches:
                matches[copyentry].append(match)
            else:
                matches[copyentry] = [match]

        destinations = {}
        # Preserve the order of the copy entries so the later ones overwrite
        # the former ones
        for copyentry in self.installer.files_to_copy:
            if copyentry not in matches:
                continue
            sources = matches[copyentry]
            cxlog.log_("aie", "* %s -> %s" % (cxlog.debug_str(copyentry.glob), cxlog.debug_str(copyentry.destination)))

            # First figure out what the destination is, trying to minimise the
            # number of patch conversions
            if copyentry.destination in destinations:
                destination = destinations[copyentry.destination]
            else:
                destination = bottlequery.get_native_path(engine.installtask.bottlename, engine.expand_string(copyentry.destination))
                destinations[copyentry.destination] = destination

            # Note that if destination exists, it may have lost its trailing
            # '/'. So check the profile intent using copyentry.destination.
            # Also the copy operation should be idempotent. This is what
            # requires the isdir(sources[0]) check.
            if len(sources) == 1 and \
                    not copyentry.destination.endswith("/") and \
                    not os.path.isdir(destination) and \
                    not os.path.isdir(sources[0]):
                dst_dir = os.path.dirname(destination)
            else:
                dst_dir = destination
                destination = None

            # Then create the appropriate destination directory
            try:
                os.makedirs(dst_dir)
            except OSError, ose:
                if ose.errno != errno.EEXIST:
                    self.error = _("Unable to create the desination directory '%(dir)s' for '%(profile)s':\n%(error)s") % {
                        'dir': dst_dir,
                        'profile': self.name,
                        'error': unicode(ose)}
                    return False
            if not os.path.isdir(dst_dir):
                self.error = _("Cannot copy the %(profile)s file(s) into '%(dir)s' because it is not a directory.") % {
                    'profile': self.name,
                    'dir': dst_dir}
                return False

            # Finally, do the actual copying
            for source in sources:
                dst = destination
                if not dst:
                    dst = os.path.join(dst_dir, os.path.basename(source))
                try:
                    cxlog.log_("aie", "copying %s to '%s'" % (cxlog.debug_str(source), cxlog.debug_str(dst)))
                    if os.path.isfile(source):
                        shutil.copy2(source, dst)
                    else:
                        # FIXME: Should also work if the destination exists
                        shutil.copytree(source, dst)
                except OSError, ose:
                    self.error = _("An error occurred while copying '%(source)s' to '%(destination)s':\n%(error)s") % {
                        'source': source,
                        'destination': dst,
                        'error': unicode(ose)}
                    return False
        return True


    def _main_reboot(self):
        if not self.installer.post_install_reboot:
            return True

        engine = self.scheduler
        cmd = [os.path.join(cxutils.CX_ROOT, "bin", "wine"),
               "--bottle", engine.installtask.bottlename,
               "--wait-children", "--wl-app", "reboot.exe"]
        # FIXME: Make interruptible.
        retcode, _out, err = cxutils.run(cmd, env=engine.state['environ'],
                                         stderr=cxutils.GRAB)
        if retcode:
            self.error = err
            return False
        return True


    def _main_create_lnk_files(self):
        engine = self.scheduler
        for lnk in self.installer.lnk_files:
            cmd = [os.path.join(cxutils.CX_ROOT, "bin", "wine"),
                   "--bottle", engine.installtask.bottlename,
                   "--wait-children", "--wl-app", "cxmklnk.exe", "--",
                   "--lnkfile", engine.expand_string(lnk.shortcut) + ".lnk",
                   "--target", engine.expand_string(lnk.target)]
            if lnk.workdir:
                cmd.extend(("--workdir", engine.expand_string(lnk.workdir)))

            # FIXME: Make interruptible.
            retcode, _out, err = cxutils.run(cmd, env=engine.state['environ'],
                                             stderr=cxutils.GRAB)
            if retcode:
                self.error = err
                return False
        return True


    def _main_register_dlls(self):
        engine = self.scheduler
        for dll in self.installer.post_registerdll:
            cmd = [os.path.join(cxutils.CX_ROOT, "bin", "wine"),
                   "--bottle", engine.installtask.bottlename,
                   "--wait-children", "--wl-app", "regsvr32.exe", "--", "/s",
                   engine.expand_string(dll)]

            # FIXME: Make interruptible.
            retcode, _out, err = cxutils.run(cmd, env=engine.state['environ'],
                                             stderr=cxutils.GRAB)
            if retcode:
                self.error = err
                return False
        return True


    def _main_delete_uncompressed_files(self):
        if 'temporary_source' in self.state:
            shutil.rmtree(self.state['temporary_source'])
        return True


    # The routines implementing the 'All Fonts' hack

    def _main_check_silent_font_install(self):
        cxlog.log_('aie', 'appid=%s silentfontinstall=%s global=%s' % (cxlog.to_str(self.installer.parent.appid), cxlog.to_str(self.state.get('silentfontinstall')), cxlog.to_str(self.scheduler.state.get('silentfontinstall'))))
        if self.state.get('silentfontinstall') and self.scheduler.state.get('silentfontinstall'):
            self.state['silentinstall'] = True
            cxlog.log_('aie', ' -> silentinstall = True')
        return True


    def _main_set_silent_font_install(self):
        cxlog.log_('aie', 'appid=%s' % cxlog.to_str(self.installer.parent.appid))
        if self.state.get('silentfontinstall'):
            cxlog.log_('aie', ' -> global silentfontinstall = True')
            self.scheduler.state['silentfontinstall'] = True
        return True


    def _main_cxhack_steam(self):
        # custom hacks for com.codeweavers.c4.206 aka Steam
        # occurs after install finishes on non-macs
        if not distversion.IS_MACOSX:
            if self.installer.parent.appid == 'com.codeweavers.c4.206' or self.installer.parent.appid == 'com.codeweavers.c4.steamcd' :
                engine = self.scheduler
                target = os.path.join(engine.expand_string("%ProgramFiles%"), "Steam")
                target = bottlequery.get_native_path(engine.installtask.bottlename, target)
                if target:
                    cmd = ["chattr", "-R", "-D", target]
                    cxutils.run(cmd, env=engine.state['environ'], stderr=cxutils.GRAB)
        return True

    def _main_cxhack_ie6_pre(self):
        # custom hacks for com.codeweavers.c4.15 aka IE6
        # occurs before install
        if self.installer.parent.appid == 'com.codeweavers.c4.15':
            engine = self.scheduler
            cmd = [os.path.join(cxutils.CX_ROOT, "bin", "wine"),
                   "--bottle", engine.installtask.bottlename,
                   "--wait-children", "--wl-app", "iexplore.exe", "-unregserver"]
            retcode, _out, err = cxutils.run(cmd, env=engine.state['environ'],
                                             stderr=cxutils.GRAB)
            if retcode:
                self.error = err
                return False
            # move the builtin iexplore.exe out of the way
            source = os.path.join(engine.expand_string("%ProgramFiles%"), "Internet Explorer")
            target = source
            source = os.path.join(source, "iexplore.exe")
            source = bottlequery.get_native_path(engine.installtask.bottlename, source)
            target = os.path.join(target, "iexplore.exe.bak")
            target = bottlequery.get_native_path(engine.installtask.bottlename, target)
            if os.path.exists(source):
                if os.path.exists(target):
                    os.remove(target)
                shutil.move(source, target)
        return True

    def _main_cxhack_ie6_post(self):
        # custom hacks for com.codeweavers.c4.15 aka IE6
        # occurs after install
        if self.installer.parent.appid == 'com.codeweavers.c4.15':
            engine = self.scheduler
            # if iexplore.exe does not exist restore our backup
            source = os.path.join(engine.expand_string("%ProgramFiles%"), "Internet Explorer")
            target = source
            target = os.path.join(source, "iexplore.exe")
            target = bottlequery.get_native_path(engine.installtask.bottlename, target)
            if not os.path.exists(target):
                source = os.path.join(source, "iexplore.exe.bak")
                source = bottlequery.get_native_path(engine.installtask.bottlename, source)
                if os.path.exists(source):
                    shutil.move(source, target)

        return True

    def _main_cxhack_mplayer2(self):
        # custom hacks for com.codeweavers.c4.96 aka Windows Media Player 6.4
        # occurs before install
        if self.installer.parent.appid == 'com.codeweavers.c4.96':
            engine = self.scheduler
            # remove wininit.ini
            target = os.path.join(engine.expand_string("%WinDir%"), "wininit.ini")
            target = bottlequery.get_native_path(engine.installtask.bottlename, target)
            if os.path.exists(target):
                os.remove(target)
        return True

    def _main_cxhack_wow(self):
        # custom hacks for com.codeweavers.c4.1185 aka World of Warcraft
        # occurs before install
        if self.installer.parent.appid == 'com.codeweavers.c4.1185':
            engine = self.scheduler
            wtfdir = os.path.join(engine.expand_string("%ProgramFiles%"), "World of Warcraft")
            wtfdir = os.path.join(wtfdir, "WTF")
            wtfdir = bottlequery.get_native_path(engine.installtask.bottlename, wtfdir)
            wtffile = os.path.join(wtfdir, "Config.wtf")
            os.makedirs(wtfdir)
            if os.path.exists(wtffile):
                dfile = open(wtffile, 'a')
            else:
                dfile = open(wtffile, 'w')
            dfile.writelines(['\r\n',
                '"#  CrossOver settings:" "CrossOver settings"\r\n',
                '{SET SoundOutputSystem "1"} {SET SoundOutputSystem}\r\n',
                '{SET SoundBufferSize "232"} {SET SoundBufferSize}\r\n',
                '{SET gxApi "opengl"} {SET gxApi}\r\n',
                '{SET ffxDeath "0"} {SET ffxDeath}\r\n',
                '{SET ffx "0"} {SET ffx}\r\n',
                '{SET ffxGlow "0"} {SET ffxGlow}\r\n',
                '{SET M2UseShaders "0"} {SET M2UseShaders}\r\n',
                '{SET pixelShaders "1"} {SET pixelShaders}\r\n',
                '{SET SmallCull "0.07"} {SET SmallCull}\r\n',
                '{SET specular "0"} {SET specular}\r\n',
                '{SET trilinear "0"} {SET trilinear}\r\n',
                '{SET useWeatherShaders "0"} {SET useWeatherShaders}\r\n',
                '{SET weatherDensity "1"} {SET weatherDensity}\r\n',
                '"#  (end CrossOver settings)" "end CrossOver settings"\r\n',
                '\r\n'])
            dfile.close()
        return True

    def _main_cxhack_cxhtml(self):
        # custom hacks for com.codeweavers.c4.cxhtml
        # occurs before install
        if self.installer.parent.appid == 'com.codeweavers.c4.cxhtml':
            engine = self.scheduler

            cmd = [os.path.join(cxutils.CX_ROOT, "bin", "wine"),
                   "--bottle", engine.installtask.bottlename,
                   "--wait-children", "--wl-app", "iexplore.exe", "-regserver"]
            retcode, _out, err = cxutils.run(cmd, env=engine.state['environ'],
                                             stderr=cxutils.GRAB)
            if retcode:
                self.error = err
                return False
        return True

    def main(self):
        routines = ('set_installer_file',
                    'check_silent_font_install',
                    'setup_environment',
                    'setup_registry_pre',
                    'uncompress',
                    'locate_installer',
                    'build_install_command',
                    'collate_eassocs',
                    'collate_nsplugins',
                    'cxhack_ie6_pre',
                    'cxhack_mplayer2',
                    'cxhack_wow',
                    'cxhack_cxhtml',
                    'remove_fake_dlls',
                    'run_installer_and_check',
                    'copy_files',
                    'reboot',
                    'create_lnk_files',
                    'register_dlls',
                    'set_silent_font_install',
                    'delete_uncompressed_files',
                    'cxhack_steam',
                    'cxhack_ie6_post',
                    )
        return self.run_routines('main', routines)
