# (c) Copyright 2009-2010. CodeWeavers, Inc.

import os
import UserDict

import cxutils
import cxlog


def _unescape_literal(string):
    # 0 = an invalid literal
    # 1 = a quoted string that is not closed
    # string value = a valid literal
    string = string.strip()
    if '"' not in string:
        return string
    elif not string.startswith('"'):
        return 0 #invalid
    elif not string.endswith('"'):
        return 1 #not closed
    else:
        contents = string[1:-1]
        unescaped_quotes = 0
        for substring in contents.split('"')[:-1]:
            if (len(substring) - len(substring.rstrip('\\'))) % 2 != 1:
                unescaped_quotes += 1
        if unescaped_quotes % 2 == 1:
            return 1 # a quoted substring is open
        elif unescaped_quotes:
            return string # string consists of multiple quoted substrings
        else:
            return cxutils.unescape_string(contents)

class CXSection(object, UserDict.DictMixin):

    def __init__(self, name):
        self.name = name
        self.keypairs = {}
        self.original_lines = []

    def __getitem__(self, key):
        return self.keypairs[key.lower()][1]

    def __setitem__(self, key, value):
        self.keypairs[key.lower()] = (key, value)

    def __delitem__(self, key):
        del self.keypairs[key.lower()]

    def keys(self):
        result = []
        for key, _unused in self.keypairs.itervalues():
            result.append(key)
        return result

    def __contains__(self, item):
        return item.lower() in self.keypairs

    def read(self, lineno, lines, config_file):
        self.original_lines = []
        for lineno in range(lineno, len(lines)):
            line = lines[lineno].strip()

            if line.startswith(';') or not line:
                # comment
                self.original_lines.append((0, line))
            elif line.startswith('[') and line.endswith(']'):
                # start of a new section
                break
            else:
                # key/value pair?
                tokens = line.split('=')
                if len(tokens) < 2:
                    # line does not contain =
                    cxlog.warn("invalid line %s in section %s of file %s" % (cxlog.to_str(line), cxlog.to_str(self.name), cxlog.to_str(config_file)))
                    continue
                for i in range(1, len(tokens)):
                    key = _unescape_literal('='.join(tokens[0:i]))
                    if key == 2:
                        # quoted literal not closed
                        continue
                    else:
                        # literal is known to be valid or invalid
                        break
                if not isinstance(key, basestring):
                    # invalid literal or quoted literal not closed at end of line
                    cxlog.warn("invalid line %s in section %s of file %s" % (cxlog.to_str(line), cxlog.to_str(self.name), cxlog.to_str(config_file)))
                    continue
                # pylint 0.18.1 on the Mac seems to be buggy and does not want
                # us to remove the I0011!
                # pylint: disable-msg=I0011,W0631
                value = _unescape_literal('='.join(tokens[i:]))
                self.original_lines.append((1, key))
                self[key] = os.path.expanduser(os.path.expandvars(value))
        else:
            # no more sections
            lineno = len(lines)
        # pylint: disable-msg=W0631
        return lineno

    def _write_keypair(self, file_obj, keyname, values_written):
        lowerkey = keyname.lower()
        if lowerkey not in values_written and lowerkey in self.keypairs:
            key, value = self.keypairs[lowerkey]
            file_obj.write('"%s" = "%s"\n' % (cxutils.escape_string(key), cxutils.escape_string(value)))
            values_written.add(lowerkey)

    def write(self, file_obj):
        if not self.keypairs and not self.original_lines:
            # this section is empty so don't write it
            return

        file_obj.write('[%s]\n' % self.name)

        values_written = set()

        # first write the keys and comments from the original file
        for iskey, line in self.original_lines:
            if iskey:
                self._write_keypair(file_obj, line, values_written)
            else:
                file_obj.write('%s\n' % line)

        # now write any remaining keys
        for key in self.keypairs:
            self._write_keypair(file_obj, key, values_written)


class CXConfig(object, UserDict.DictMixin):
    """Base class for reading/writing/editing configurations.
    
    Users should use CXMergeConfig or CXRWConfig, depending on whether they want
    to merge or write."""

    canmerge = True
    canwrite = True

    def __init__(self):
        if type(self) is CXConfig:
            raise TypeError("CXConfig is an abstract base class, use CXMergeConfig or CXRWConfig")
        self.files = []
        self.sections = {}
        self.original_comments = []
        self.original_sections = []

    def _read_lines(self, lines, config_file):
        # reset the stored comments and section order
        self.original_comments = []
        self.original_sections = []

        # go through the lines until we hit a section
        lineno = 0
        while lineno < len(lines):
            line = lines[lineno].strip()

            if line.startswith('[') and line.endswith(']'):
                break
            elif line.startswith(';') or not line:
                self.original_comments.append(line)
            else:
                cxlog.warn("invalid line %s before start of section in file %s" % (cxlog.to_str(line), cxlog.to_str(config_file)))
            lineno += 1

        # now go through the sections
        while lineno < len(lines):
            # line.startswith('[') and line.endswith(']')
            line = lines[lineno].strip()

            sectionname = line[1:-1]
            self.original_sections.append(sectionname)
            section = self[sectionname]
            lineno = section.read(lineno+1, lines, config_file)

    def read(self, config_file):
        """Reads the specified configuration file and merges it with the
        settings that have already been read.

        It's ok to try to read a nonexistent file.
        """
        if self.files and not self.canmerge:
            raise TypeError("This class cannot merge config files")

        try:
            # Grab the mtime so that later we can check whether our
            # in-memory data is still up to date
            config_mtime = os.path.getmtime(config_file)
        except OSError:
            config_mtime = None
        self.files.append((config_file, config_mtime))
        try:
            thisfile = open(config_file, 'r')
        except IOError:
            # The file does not exist so there is nothing more to do
            return
        lines = list(thisfile.readlines())
        thisfile.close()

        if len(lines) > 1 and lines[0].startswith('\xef\xbb\xbf'): #utf8 bom(b)
            lines[0] = lines[0][3:]

        self._read_lines(lines, config_file)

    def read_string(self, config_string):
        """Reads configuratin data from the given string and merges it with the
        settings that have already been read."""
        self._read_lines(config_string.splitlines(), '<string>')

    def is_up_to_date(self):
        """Returns True if the configuration files have not changed since
        they were read, and False otherwise.
        """
        for entry in self.files:
            try:
                mtime = os.path.getmtime(entry[0])
            except OSError:
                mtime = None
            if mtime != entry[1]:
                return False
        return True

    def _write_section(self, file_obj, sectionname, sections_written):
        lowername = sectionname.lower()
        if lowername not in sections_written:
            self.sections[lowername].write(file_obj)
            sections_written.add(lowername)

    def write(self, config_file):
        """Writes the in-memory settings to the specified file.

        The configuration file is replaced atomically so that another process
        or thread should never see an incomplete configuration file.
        """
        if not self.canwrite:
            raise TypeError("This class cannot write config files")

        if isinstance(config_file, file):
            file_obj = config_file
        else:
            # make sure we have a directory to write to.
            dirname = os.path.dirname(config_file)
            if not os.path.exists(dirname):
                os.makedirs(dirname)

            # don't clobber the existing config, just in case there's an
            # error in writing
            temp_config_file = config_file + '.tmp'
            file_obj = open(temp_config_file, 'w')

        try:
            # write the comments from before the first section
            for line in self.original_comments:
                file_obj.write('%s\n' % line)

            sections_written = set()

            # write the sections from the original file
            for section in self.original_sections:
                self._write_section(file_obj, section, sections_written)

            # write any new sections
            for section in self.sections.keys():
                self._write_section(file_obj, section, sections_written)
        finally:
            if not isinstance(config_file, file):
                file_obj.close()
                # now overwrite the config with the amazing atomic rename
                # command
                os.rename(temp_config_file, config_file)

    def __getitem__(self, key):
        try:
            return self.sections[key.lower()]
        except KeyError:
            # unlike an explicit assignment, setdefault is threadsafe
            return self.sections.setdefault(key.lower(), CXSection(key))

    def __setitem__(self, key, value):
        # FIXME: create a new section?
        raise NotImplemented

    def __delitem__(self, key):
        # FIXME: delete a section?
        raise NotImplemented

    def keys(self):
        result = []
        for item in self.sections.itervalues():
            result.append(item.name)
        return result

    def __contains__(self, item):
        return item.lower() in self.sections

    def _getfilenames(self):
        """Returns the filenames of the configuration files that were read."""
        filenames = []
        for entry in self.files:
            filenames.append(entry[0])
        return filenames

    filenames = property(_getfilenames)


class CXMergeConfig(CXConfig):
    """CXMergeConfig reads one or more configuration files and returns
    a read-only view of the settings.
    """

    canmerge = True
    canwrite = False


class CXRWConfig(CXConfig):
    """CXRWConfig reads a single configuration file that can be modified
    in memory. The settings can then be written back to the file while
    minimizing the differences with the previous version.
    """

    canmerge = False
    canwrite = True

    def __init__(self):
        CXConfig.__init__(self)
        self._lock = None

    def _getfilename(self):
        """Returns the filename of the configuration file that was read."""
        if self.files:
            return self.files[0][0]
        return None

    filename = property(_getfilename)


    def lock_and_read(self, config_file, lock_name=None):
        """Creates and takes a write lock for the specified configuration file
        and then reads it.

        Note that the OS-level lock is not taken on the configuration file
        itself so we can lock nonexistent files and also so we allow
        unfettered read access. The latter means using atomic renames to put
        the new file in place when saving which would interfer locks taken on
        that file.
        """
        if self.files and not self.canmerge:
            raise TypeError("This class cannot merge config files")

        try:
            if not lock_name:
                lock_name = os.path.basename(config_file)
            self._lock = cxutils.FLock(lock_name)
        except:
            self._lock = None

        self.read(config_file)


    def unlock(self):
        """Unlocks the configuration file."""
        if self._lock:
            try:
                self._lock.release()
            finally:
                self._lock = None


    def save(self):
        """Saves the configuration file."""
        if self.files:
            self.write(self.filename)
