# (c) Copyright 2009. CodeWeavers, Inc.

import UserDict
import os

import cxobjc
import cxconfig
import cxutils

import bottlequery

from cxutils import cxgettext as _

class Error(Exception):
    pass

class CXMenu(cxobjc.Proxy):
    pass

def cxmenu_path():
    return os.path.join(cxutils.CX_ROOT, "bin", "cxmenu")

def run_cxmenu(args, grab_stdout = False, background = False):
    if grab_stdout:
        stdout = cxutils.GRAB
    else:
        stdout = None

    if background:
        stderr = None
    else:
        stderr = cxutils.GRAB

    retcode, out, err = cxutils.run((cxmenu_path(),)+tuple(args), stderr=stderr, stdout=stdout, background=background)

    if retcode:
        raise Error(err)
    else:
        return out


def get_menuroot(bottlename):
    config = cxconfig.CXMergeConfig()

    config_filename = os.path.join(bottlequery.get_prefix_for_bottle(bottlename), 'cxbottle.conf')

    config.read(config_filename)

    return config['Bottle'].get('MenuRoot', '')


def set_menuroot(bottlename, new_root):
    config = cxconfig.CXRWConfig()

    config_filename = os.path.join(bottlequery.get_prefix_for_bottle(bottlename), 'cxbottle.conf')

    config.read(config_filename)

    config['Bottle']['MenuRoot'] = new_root

    config.write(config_filename)

    run_cxmenu(('--removeall', '--install', '--bottle', bottlename))


@cxobjc.method(CXMenu, 'localStartMenuPath')
def local_start_menu_path():
    return _("Start Menu")


@cxobjc.method(CXMenu, 'localDesktopPath')
def local_desktop_path():
    return _("Desktop")


@cxobjc.method(CXMenu, 'createShortcutForBottle_command_name_folder_localDesktopPath_andStartMenuPath_')
def create_custom_shortcut(bottle, command, name, folder, local_desktop, local_startmenu):

    if folder.startswith(local_desktop):
        desktop_folder = folder[len(local_desktop):]
        if desktop_folder[0:1] in ('', '/', '\\'):
            link = 'Desktop%s/%s' % (desktop_folder.replace('\\','/'), name)
    elif folder.startswith(local_startmenu):
        startmenu_folder = folder[len(local_startmenu):]
        if startmenu_folder[0:1] in ('', '/', '\\'):
            link = 'StartMenu%s/%s' % (startmenu_folder.replace('\\','/'), name)

    if not link:
        folder = folder.replace('\\','/')
        if not folder.startswith('/'):
            folder = '/'+folder
        if not folder.endswith('/'):
            folder = folder+'/'
        link = 'StartMenu%s%s' % (folder, name)

    argv = cxutils.cmdlinetoargv(command)
    prefix = bottlequery.get_prefix_for_bottle(bottle)+'/'
    for n, arg in enumerate(argv):
        if arg.startswith(prefix):
            argv[n] = '${WINEPREFIX}/' + arg[len(prefix):]
    unix_cmd = '${CX_ROOT}/bin/cxstart --bottle \'${CX_BOTTLE}\' -- %s' % cxutils.shell_escape(argv)

    cxutils.run((os.path.join(cxutils.CX_ROOT, "bin", "cxmenu"),
        '--bottle', bottle,
        '--create', link,
        '--type', 'raw',
        '--command', unix_cmd,
        '--install'), background=True)


class MenuItem(object):
    """A single menu shortcut."""

    def __init__(self, parent, section):
        self.parent = parent

        # encoded path
        self.path = section.name

        # 'install' or 'ignore'
        self.mode = section['Mode']

        # this mode will be set when MenuPrefs.commit() is called
        self.new_mode = self.mode

        # 'windows' or 'raw'
        self.type = section['Type'].lower()

        # optional name of command-line shortcut
        self.shortcut = section.get('Shortcut')

        # optional description
        self.description = section.get('Description')

        # command, if type is 'raw'
        self.command = section.get('Command')

        # optional icon filename; use iconfile attribute instead
        self._config_icon = section.get('Icon')

    def _get_iconfile(self):
        if self._config_icon:
            return self._config_icon
        else:
            return os.path.join(
                bottlequery.get_prefix_for_bottle(self.parent.bottlename),
                'windata',
                cxutils.mangle_path(self.path)+'.xpm')

    iconfile = property(_get_iconfile)

    def start(self):
        run_cxmenu(('--start', '--bottle', self.parent.bottlename, '--filter', self.path), background=True)

class MenuPrefs(object, UserDict.IterableUserDict):
    """An editable menu configuration for a bottle"""

    def __init__(self, bottlename):
        UserDict.IterableUserDict.__init__(self)
        self.bottlename = bottlename

    def read_config(self):
        """Reads the menu information from cxmenu.conf."""
        config = cxconfig.CXMergeConfig()

        config_filename = os.path.join(bottlequery.get_prefix_for_bottle(self.bottlename), 'cxmenu.conf')

        config.read(config_filename)

        self.data.clear()
        for section in config:
            self.data[section] = MenuItem(self, config[section])

    def query_config(self):
        """Update menu settings from cxmenu --query."""
        data = run_cxmenu(('--query', '--bottle', self.bottlename), grab_stdout=True)

        config = cxconfig.CXMergeConfig()

        config.read_string(data)

        for section in config:
            if section in self.data:
                if config[section].get('IDs'):
                    self.data[section].mode = 'install'
                else:
                    self.data[section].mode = 'ignore'

    def refresh(self):
        """Read the menu configuration for this bottle"""
        self.read_config()
        self.query_config()

        self.new_menuroot = self.menuroot = get_menuroot(self.bottlename)

    def commit(self):
        """Save and apply the changes using cxmenu."""

        # Find all changed associations, grouped by the new setting
        changes = {'ignore': [],
                   'install': [],
                   }
        for menu in self.data.values():
            if menu.new_mode != menu.mode:
                changes[menu.new_mode].append(menu.path)

        if changes['install']:
            run_cxmenu(('--bottle', self.bottlename,
                        '--mode', 'install',
                        '--install',
                        '--filter', ':'.join(changes['install'])))
        if changes['ignore']:
            run_cxmenu(('--bottle', self.bottlename,
                        '--mode', 'ignore',
                        '--uninstall',
                        '--filter', ':'.join(changes['ignore'])))
        if self.new_menuroot != self.menuroot:
            #FIXME: Don't bother with --install and --uninstall in previous
            # calls, if we're going to do this.
            set_menuroot(self.bottlename, self.new_menuroot)
            self.menuroot = self.new_menuroot

        for path in self.data:
            self.data[path].mode = self.data[path].new_mode


