# (c) Copyright 2009. CodeWeavers, Inc.

import UserDict
import os

import cxconfig
import cxutils

import bottlequery

class Error(Exception):
    pass

def cxnsplugin_path():
    return os.path.join(cxutils.CX_ROOT, "bin", "cxnsplugin")

def run_cxnsplugin(args, grab_stdout = False, background = False):
    if grab_stdout:
        stdout = cxutils.GRAB
    else:
        stdout = None

    if background:
        stderr = None
    else:
        stderr = cxutils.GRAB

    retcode, out, err = cxutils.run((cxnsplugin_path(),)+tuple(args), stderr=stderr, stdout=stdout, background=background)

    if retcode:
        raise Error(err)
    else:
        return out

class PluginItem(object):
    """A single plugin dll."""

    def __init__(self, parent, section):
        self.parent = parent

        # dll identifier
        self.dllkey = section.name

        # dll path
        self.path = cxutils.unmangle(self.dllkey)

        # dll basename
        self.basename = self.path.rsplit('/',2)[-1]

        # 'install' or 'ignore'
        self.mode = section['Mode']

        # this mode will be set when PluginPrefs.commit() is called
        self.new_mode = self.mode

        # human-readable name
        self.name = section.get('Name')

        # human-readable description
        self.description = section.get('Description')

        # pipe-delimited list of mime types
        self.mime_types = section.get('MimeTypes')

        # pipe-delimited list of filename globs
        self.extensions = section.get('Extensions')

        # pipe-delimited list of file filters
        self.open_masks = section.get('OpenMasks')

class PluginPrefs(object, UserDict.IterableUserDict):
    """An editable plugin configuration for a bottle"""

    def __init__(self, bottlename):
        UserDict.IterableUserDict.__init__(self)
        self.bottlename = bottlename

    def config_filename(self):
        return os.path.join(bottlequery.get_prefix_for_bottle(self.bottlename), 'cxnsplugin.conf')

    def sync_config(self):
        run_cxnsplugin(('--sync', '--bottle', self.bottlename))

    def read_config(self):
        """Reads the plugin information from cxplugin.conf."""
        config = cxconfig.CXMergeConfig()

        config.read(self.config_filename())

        self.data.clear()
        for section in config:
            self.data[section] = PluginItem(self, config[section])

    def query_config(self):
        """Update settings from cxnsplugin --query."""
        data = run_cxnsplugin(('--query', '--bottle', self.bottlename), grab_stdout=True)

        config = cxconfig.CXMergeConfig()

        config.read_string(data)

        all_plugindirs = config['.all']
        for section in config:
            if section in self.data:
                if config[section] == all_plugindirs:
                    self.data[section].mode = 'install'
                else:
                    self.data[section].mode = 'ignore'

    def refresh(self):
        """Read the plugin configuration for this bottle"""
        if not os.path.exists(self.config_filename()):
            self.sync_config()
        self.read_config()
        self.query_config()

    def commit(self):
        """Save and apply the changes using cxnsplugin."""

        # Find all changed associations, grouped by the new setting
        changes = {'ignore': [],
                   'install': [],
                   }
        for plugin in self.data.values():
            if plugin.new_mode != plugin.mode:
                changes[plugin.new_mode].append(plugin.dllkey)

        if changes['install']:
            run_cxnsplugin(('--bottle', self.bottlename,
                        '--mode', 'install',
                        '--install',
                        '--filter', ':'.join(changes['install'])))
        if changes['ignore']:
            run_cxnsplugin(('--bottle', self.bottlename,
                        '--mode', 'ignore',
                        '--uninstall',
                        '--filter', ':'.join(changes['ignore'])))

        for plugin in self.data.values():
            plugin.mode = plugin.new_mode

    def recreate_plugins(self):
        run_cxnsplugin(('--bottle', self.bottlename,
                    '--install'))

