# (c) Copyright 2009-2010. CodeWeavers, Inc.

import errno
import fcntl
import locale
import os
import stat
import subprocess
import time

import cxlog
import distversion
import cxobjc

import gettext

class CXUtils(cxobjc.Proxy):
    pass

#####
#
# CX_ROOT
#
#####

# Points to CrossOver's root directory.
# This should be set at startup and is then considered to be constant for the
# duration of the process.
CX_ROOT = None



#####
#
# File and directory helpers
#
#####

def mkdirs(path, *args, **kwargs):
    """Similar to os.makedirs() but returns False to signal errors and logs
    errors.
    """
    try:
        os.makedirs(path, *args, **kwargs)
    except OSError, ose:
        if ose.errno != errno.EEXIST:
            cxlog.err("unable to create the '%s' directory: %s" % (cxlog.debug_str(path), ose.strerror))
            return False
    return True


#####
#
# Cross-process lock
#
#####

class FLock(object):

    def __init__(self, name):
        """Creates and takes a lock called name using flock().

        This is used to avoid races between processes.
        """
        self._name = name

        # Create the lock directory and perform some sanity checks
        lockdir = os.path.join(os.environ.get('TMPDIR', '/tmp'),
                               ".wine-%d" % os.geteuid())
        mkdirs(lockdir, 0700)
        st = os.lstat(lockdir)
        if not stat.S_ISDIR(st.st_mode):
            raise ValueError("'%s' is not owned by you" % lockdir)
        if st.st_uid != os.geteuid():
            raise ValueError("'%s' is not owned by you" % lockdir)
        if st.st_mode & 077:
            raise ValueError("'%s' must not be accessible by other users" % lockdir)

        # Perform some more sanity checks and create the lock file
        self._path = os.path.join(lockdir, name + ".lock")
        try:
            st = os.lstat(self._path)
            if not stat.S_ISREG(st.st_mode):
                raise ValueError("'%s' is not a regular file" % self._path)
        except OSError, ose:
            if ose.errno != errno.ENOENT:
                raise ose
        self._lock = open(self._path, 'w')

        # Take the lock
        cxlog.log("%d: Grabbing the '%s' lock" % (os.getpid(), self._path))
        try:
            fcntl.flock(self._lock, fcntl.LOCK_EX)
        except Exception, e:
            # The filesystem probably does not support locking
            cxlog.warn("unable to take the '%s' lock: %s" % (name, e))
            self._lock.close()
            self._lock = None
            raise # re-raise the exception

    def _getname(self):
        """Returns the lock name."""
        return self._name

    name = property(_getname)

    def locked(self):
        """Returns True if the lock has not been released yet."""
        return self._lock is not None

    def release(self):
        """Releases the lock.

        Note that deleting a lock carries a very high risk of causing races
        outside the scope of this class and has essentially no benefit. So the
        file created for the lock is not deleted and no support is provided
        for deleting it.
        """
        if self._lock:
            cxlog.log("%d: Releasing the '%s' lock" % (os.getpid(), self._name))
            try:
                try:
                    fcntl.flock(self._lock, fcntl.LOCK_UN)
                except Exception, e:
                    cxlog.warn("unable to release the '%s' lock: %s" % (self._name, e))
            finally:
                self._lock.close()
                self._lock = None


#####
#
# String handling
#
#####

def unescape_string(string):
    """Unescapes a string that was places in double-quotes."""
    string = string.replace('\\\"', '\"')
    string = string.replace("\\\\", "\\")
    return string


def escape_string(string):
    """Escapes the string so it can be placed in double-quotes."""
    string = string.replace("\\", "\\\\")
    string = string.replace('"', '\\"')
    return string


def shquote_string(string):
    """Quote a string so it can be used in shell commands.

    Note that this implies escaping '$'s and '`'s which may not be appropriate
    in another context.
    """
    string = string.replace('\\', '\\\\')
    string = string.replace('\$', '\\\$')
    string = string.replace('\"', '\\\"')
    string = string.replace("\'", "\\\'")
    return '"' + string + '"'


def mangle(string, characters):
    result = []
    for char in string:
        if char in characters or not (0x1f < ord(char) < 0x80):
            if char == ' ':
                result.append('+')
            elif char == '/':
                result.append('_')
            elif char == '\\':
                result.append('~')
            else:
                result.append('^%02X' % ord(char))
        else:
            result.append(char)
    return ''.join(result)

def mangle_path(string):
    return mangle(string, '\\ $"\'`<>#&:;%^_~+')

def unmangle(string):
    result = []
    x = 0
    while x < len(string):
        char = string[x]
        x += 1
        if char == '+':
            result.append(' ')
        elif char == '_':
            result.append('/')
        elif char == '~':
            result.append('\\')
        elif char == '^':
            try:
                result.append(chr(int(string[x:x+2], 16)))
            except ValueError:
                # not followed by valid hex digits
                pass
            else:
                x += 2
        else:
            result.append(char)
    return ''.join(result)


def md5_hasher():
    """Returns an object for computing MD5 hashes."""
    try:
        # pylint: disable-msg=F0401
        import hashlib
        return hashlib.md5()
    except ImportError:
        import md5
        return md5.new()


def bottlename_character_filter(character):
    """Returns False for characters that are best avoided
    in bottle names.
    """
    # \/*?:|<>" are not valid characters in Windows paths
    # {} are not valid in paths stored in Debian and Solaris packages
    return character not in "\\/*?:|<>\"'{}"


@cxobjc.method(CXUtils, 'sanitizedBottleName_')
def bottlename_sanitize(bottlename):
    """Removes characters that are best avoided
      in bottle names. """

    bottlename = ''.join(x for x in bottlename if bottlename_character_filter(x))
    # . causes the bottle to be hidden if it's the first character
    return bottlename.lstrip(".")

def html_escape(string):
    """Escape characters that have special meaning in html."""
    replacements = {
        '<': '&lt;',
        '>': '&gt;',
        '&': '&amp;'}
    result = []
    for char in string:
        result.append(replacements.get(char, char))
    return ''.join(result)

@cxobjc.method(CXUtils, 'cmdLineToArgv_')
def cmdlinetoargv(string):
    """Convert a windows-style quoted command line to an argument list. This is
    NOT equivalent to CommandLineToArgvW."""
    in_quotes = False
    pos = 0
    result = []
    arg = []
    while pos < len(string):
        char = string[pos]
        if char == ' ' and not in_quotes:
            if arg:
                result.append(''.join(arg))
                del arg[:]
        elif char == '\\' and pos+1 < len(string) and string[pos+1] in '\\"':
            pos += 1
            arg.append(string[pos])
        elif char == '"':
            in_quotes = not in_quotes
        else:
            arg.append(string[pos])
        pos += 1
    if arg:
        result.append(''.join(arg))
    return result

def shell_escape(argv):
    """Convert an argument list to a string that can safely be passed to a shell."""
    escaped_args = []
    for arg in argv:
        escaped_args.append("'%s'" % arg.replace("'", "'\\''"))
    return ' '.join(escaped_args)


#####
#
# Locale functions
#
#####

def get_preferred_languages(language=None):
    """Returns a sequence of the user's preferred languages if language is
    None; and a sequence of variants otherwise.

    In the None case the result corresponds to the user's preferences on Mac
    OS X, while on Unix it is derived from the user's locale.
    When starting from a single language, the variants added allow broader
    matches to be established. For instance 'pt_br' gives ('pt-br', 'pt', '').
    """

    if not language:
        languages = distversion.get_user_languages()

        if not languages:
            # Theoretically, we should arrange to call
            # locale.setlocale(locale.LC_ALL, '') at program start and use
            # getlocale here. Other functions, e.g. localization, might need
            # the current locale to be set. If we do that, we should use
            # getlocale() here, but for now getdefaultlocale() works fine.
            language, _encoding = locale.getdefaultlocale()
            if language is not None:
                languages = (language,)
            else:
                languages = ()
    else:
        languages = (language,)

    # Add the user's preferred languages
    seen = set(('C', ''))  # Ignore these two 'languages'
    preferred = []
    for language in languages:
        language = language.lower().replace('_', '-')
        if language in seen:
            continue
        preferred.append(language)
        seen.add(language)

        # Try the fallback
        # We intermingle them with the user's preferences because when the user
        # says he wants ('en_US, 'fr') he really expects us to give him English
        # strings which means trying ('en_US', 'en', 'fr').
        language = language.split('-', 1)[0]
        if language in seen:
            continue
        preferred.append(language)
        seen.add(language)

    # Add the ultimate fallback
    preferred.append('')
    return preferred


@cxobjc.method(CXUtils, 'getLanguageValue_forLanguage_')
def get_language_value(dictionary, language=None):
    """Returns the appropriate value for the specified language from a
    dictionary mapping language names to values.

    The given dictionary must contain a '' key for the default value and may
    contain other keys named like 'pt' or 'pt-br' for language-specific
    values.
    """
    languages = get_preferred_languages(language)
    for lang in languages:
        if lang in dictionary:
            return (lang, dictionary[lang])

    if dictionary:
        # The dictionary is not empty so it should really have had a default
        # value.
        cxlog.warn("get_language_value() could not find a string, names=%s, mapping=%s" % (cxlog.debug_str(languages), cxlog.debug_str(dictionary)))
    return ('', '')


gettext_initalized = False

def init_gettext( ):
    """Initializes the gettext library """
    # pylint: disable-msg=W0603
    global gettext_initalized
    if not gettext_initalized:
        lang = get_preferred_languages()
        locale_path = os.path.join(CX_ROOT, "share", "locale")
        cxlog.log("locale path %s" % cxlog.debug_str(locale_path))
        gettext.bindtextdomain("crossover", locale_path)
        gettext.bind_textdomain_codeset("crossover","UTF-8")
        gettext.textdomain("crossover")
        gettext.translation("crossover",locale_path,languages=lang,fallback=True)
        gettext_initalized = True
    return()

def cxgettext( message ):
    # pylint: disable-msg=W0603
    global gettext_initalized
    if not gettext_initalized:
        init_gettext( )
    return gettext.gettext( message )

#####
#
# Dealing with subprocesses
#
#####

GRAB = 1
NULL = 2
STDOUT = 3
NOLOG = 4

def run(args, bufsize=0, executable=None, stdout=None, stderr=None,
        preexec_fn=None, shell=False, background=False, cwd=None, env=None,
        universal_newlines=True, logprefix=""):
    """Works the same way as subprocess.Popen, but simplifies retrieving the
    output and exit code. Returns (retcode, out, err) triplets.

    If stdout or stderr are set to GRAB, then they are captured and returned.
    If they are set to NULL, then they are redirected to /dev/null instead.
    If stderr is set to STDOUT, then it is redirected to stdout.
    If the command cannot be run, then returns -1 as the exit code instead of
    raising an exception.

    Finally, the execution is timed and traced with cxlog. If stdout or stderr
    are being captured, then they are logged too, but that can be disabled by
    or-ing NOLOG. If logprefix is set, then it is prepended to the trace.
    """
    if cxlog.is_on():
        start = time.time()
        if logprefix != "":
            logprefix = logprefix + ": "
        bg = ""
        if background:
            bg = " &"
        cxlog.log("%sRunning: %s%s" % (logprefix, " ".join(cxlog.debug_str(arg) for arg in args), bg))

    out_log = False
    if stdout:
        if (stdout & ~NOLOG) == GRAB and not background:
            subout = subprocess.PIPE
            if stdout == GRAB:
                out_log = True
        elif stdout == NULL:
            subout = open("/dev/null", "w")
        else:
            raise TypeError("invalid value for stdout")
    else:
        subout = None

    err_log = False
    if stderr:
        if (stderr & ~NOLOG) == GRAB and not background:
            suberr = subprocess.PIPE
            if stderr == GRAB:
                err_log = True
        elif stderr == STDOUT:
            suberr = subprocess.STDOUT
        elif stderr == NULL:
            if cxlog.is_on():
                suberr = None
            else:
                suberr = open("/dev/null", "w")
        else:
            raise TypeError("invalid value for stderr")
    else:
        suberr = None

    try:
        subp = subprocess.Popen(args, bufsize, executable, None, subout,
                                suberr, preexec_fn, close_fds=True,
                                shell=shell, cwd=cwd, env=env,
                                universal_newlines=universal_newlines)
        if not background:
            out, err = subp.communicate()
            retcode = subp.wait()
        else:
            out = err = ""
            retcode = 0
    except OSError, exception:
        retcode = -1
        out = ""
        err = cxlog.to_str(exception)

    if isinstance(subout, file):
        # pylint: disable-msg=E0012,E1103
        subout.close()
    if isinstance(suberr, file):
        # pylint: disable-msg=E0012,E1103
        suberr.close()
    if not background and cxlog.is_on():
        cxlog.log("%s%s -> rc=%d  (took %0.3f seconds)" % (logprefix, cxlog.debug_str(os.path.basename(args[0])), retcode, time.time() - start))
        if out_log:
            count = out.count("\n")
            if count > 20:
                cxlog.log("out=<%d lines>" % count)
            else:
                cxlog.log("out=[%s]" % out)
        if err_log:
            count = err.count("\n")
            if count > 20:
                cxlog.log("err=<%d lines>" % count)
            else:
                cxlog.log("err=[%s]" % err)

    return (retcode, out, err)

def system(args, executable=None, preexec_fn=None, shell=False,
           background=False, cwd=None, env=None, logprefix=""):
    """Works the same way as run(), except that neither stdout nor stderr are
    captured, and only the retcode is returned.
    """
    retcode, _out, _err = run(args, executable=executable,
                              preexec_fn=preexec_fn, shell=shell,
                              background=background, cwd=cwd, env=env,
                              logprefix=logprefix)
    return retcode


def launch_url(url):
    """Uses launchurl to open the specified URL in a web browser."""
    return system([os.path.join(CX_ROOT, "bin", "launchurl"), url],
                  background=True)



#####
#
# Select wrapper
#
#####


# Functions and module imports are stored in the same namespace. This means a
# simple 'import select' would get overriden by our select() function, thus
# preventing it from accessing the content of the module, and necessitating
# annoying workarounds (and rightful Pylint warnings).
#
# So we just import the parts we need, renaming them in the process to avoid
# any conflict.
from select import select as select_select
from select import error as select_error

def select(iwtd, owtd, ewtd, timeout=None):
    """Same as select.select() but retries on EINTR (with the same timeout)."""
    while True:
        try:
            return select_select(iwtd, owtd, ewtd, timeout)
        except select_error, err:
            if err.args[0] == errno.EINTR:
                if timeout is not None:
                    timeout = 0
                continue
            else:
                raise



#####
#
# umask cache
#
#####

# Retrieve the umask on startup, that is before we have multiple threads all
# trying to get it at the same time. os.umask() MUST NOT be used in anywhere
# else. Should we need to modify the umask, then we will need to write a
# proper umask() wrapper.
UMASK = os.umask(0)
os.umask(UMASK)
