# (c) Copyright 2008, 2010. CodeWeavers, Inc.
package CXBT_winvista;
use strict;

use CXLog;
use CXUtils;

my $productid=CXUtils::get_product_id();
my $template="winvista";
my $datadir="$ENV{CX_ROOT}/share/crossover/bottle_data";
my $templatedir="$ENV{CX_ROOT}/share/crossover/bottle_templates/$template";
my $windows_version="vista";


my $default_eassocids=".asf:.cfc:.disco:.doc:.dochtml:.dot:.dwr:.dwt:.efx:.em:.emp:.fla:.mdb:.moh:.mov:.mpw:.mpx:.net:.pdf:.pic:.pict:.ppt:.prj:.psd:.psf:.psp:.qdb:.qdf:.qdt:.qfx:.qt:.sch:.swf:.vrml:.vsd:.vst:.vss:.wma:.wmv:.xls";
my $alternative_eassocids=".csv:.rtf";

my $ignore_plugins="_NPDOC\\.DLL\$:_NPDocBox\\.dll\$:_npmozax\\.dll\$:_nppdf32\\.dll\$";


#
# Bottle template query
#

sub new($$)
{
    my ($class)=@_;

    my $self={ };
    bless $self, $class;
    return $self;
}

sub query($$$)
{
    my ($self, $type, $params)=@_;

    if ($type eq "introspect")
    {
        # Return a very general description of the template.
        # This query is supported by all templates.
        return { properties => { category => "crossover" } };
    }

    $@="Unknown query type '$type'";
    return undef;
}


#
# Bottle creation / upgrade helpers
#

sub create_windows_label()
{
    my $filename="$ENV{WINEPREFIX}/dosdevices/c:/.windows-label";
    if (!-f $filename and open(my $fh, ">", $filename))
    {
        print $fh "drive_c";
        close($fh);
    }
}

sub run_inf($$$)
{
    my ($winver, $scope, $upgrade)=@_;

    my @cxoptions=();
    if (defined $winver)
    {
        # We must both use --winver, for when called from cxbottle, and set
        # CX_WINDOWS_VERSION, for when called by wine, i.e. when the Wine
        # environment is already set.
        $ENV{CX_WINDOWS_VERSION}=$winver;
        push @cxoptions, "--winver", $winver;
    }

    # Set up some dll overrides to run wine.inf
    $ENV{WINEDLLOVERRIDES}=join(";",
        "oleaut32=b",     # Workaround for a native oleaut32 bug
        "rpcrt4=b",       # Incompatibility with native rpcrt4
        "shdocvw=b",      # wine.inf invokes 'iexplore -regserver' which
        "*iexplore.exe=b" # is specific to the builtin version
    );
    if (cxsystem("$ENV{CX_ROOT}/bin/wine", "--wl-app", "rundll32.exe",
                 "--no-lock", "--no-quotes", "--scope", $scope, @cxoptions,
                 "--dll", $ENV{WINEDLLOVERRIDES},
                 "setupapi.dll,InstallHinfSection", "DefaultInstall", "128",
                 "$ENV{CX_ROOT}/share/wine/wine.inf"))
    {
        $@="'rundll32 DefaultInstall wine.inf' failed\n";
        return undef;
    }
    delete $ENV{CX_WINDOWS_VERSION};

    my $section;
    if ($upgrade)
    {
        # Upgrades shouldn't change the bottle version
        $section="DefaultInstall";
    }
    elsif (($winver || "") eq "win98")
    {
        # No need to change the version set by wine.inf
        $section="DefaultInstall";
    }
    elsif (($winver || "") eq "win2000")
    {
        $section="Win2000Install";
    }
    elsif (($winver || "") eq "winxp")
    {
        # No need to change the version set by wine.inf
        $section="DefaultInstall";
    }
    elsif (($winver || "") eq "vista")
    {
        $section="WinVistaInstall";
    }
    else
    {
        cxerr("unknown Windows version '$winver'\n");
    }
    if ($section and
        cxsystem("$ENV{CX_ROOT}/bin/wine", "--wl-app", "rundll32.exe",
                 "--no-lock", "--no-quotes", "--scope", $scope,
                 "setupapi.dll,InstallHinfSection", $section, "128",
                 "$datadir/crossover.inf"))
    {
        $@="'rundll32 $section crossover.inf' failed\n";
        return undef;
    }
    return 1;
}


#
# Bottle creation
#

sub create($$$)
{
    my ($self, $cxconfig, $params)=@_;
    $@="";

    # Create the core bottle directories
    foreach my $dir ("$ENV{WINEPREFIX}/drive_c/Program Files/Netscape/Communicator/Program/Plugins",
                     "$ENV{WINEPREFIX}/drive_c/windows/command",
                     "$ENV{WINEPREFIX}/drive_c/windows/fonts",
                     "$ENV{WINEPREFIX}/drive_c/windows/inf",
                     "$ENV{WINEPREFIX}/drive_c/windows/system",
                     "$ENV{WINEPREFIX}/drive_c/windows/system32/drivers",
                     "$ENV{WINEPREFIX}/drive_c/windows/temp")
    {
        if (!-d $dir)
        {
            cxlog("creating directory '$dir'\n");
            if (!cxmkpath($dir))
            {
                $@="unable to create the '$dir' directory: $@";
                return undef;
            }
        }
    }

    # This is not yet considered to be a bottle because there is no system.reg
    # file. But now is the time to lock the bottle to avoid races.
    my $lock=CXBottle::lock_bottle($ENV{WINEPREFIX});

    # Set up the drives
    CXBottle::setup_drives($ENV{WINEPREFIX});
    create_windows_label();

    # Set up the cxbottle.conf configuration file
    require CXRWConfig;
    my $cxbottle=CXRWConfig->new("$datadir/cxbottle.conf");
    my $s=$cxbottle->append_section("Bottle");
    $s->set("BottleID", CXUtils::get_unique_id($ENV{WINEPREFIX}));
    $s->set("Timestamp", $cxconfig->get("CrossOver", "BuildTimestamp"));
    $s->set("Encoding", CXUtils::get_system_encoding(1));
    $s->set("Template", $template);
    $s->set("Description", $params->{description}) if (defined $params->{description});
    $s->set("Updater", $params->{updater}) if (defined $params->{updater});
    my $menu_root=$cxconfig->get("BottleDefaults", "MenuRoot");
    $s->set("MenuRoot", $menu_root) if (defined $menu_root);
    my $menu_strip=$cxconfig->get("BottleDefaults", "MenuStrip");
    $s->set("MenuStrip", $menu_strip) if (defined $menu_strip);
    $s->set("MenuMode", "ignore");
    $s->set("AssocMode", "ignore");
    $s->set("NSPluginMode", "ignore");
    $cxbottle->write("$ENV{WINEPREFIX}/cxbottle.conf");

    # CXBottle::find_bottle() checks for the presence of system.reg to detect
    # invalid bottle directories. This creates a race condition due to
    # wineserver's delayed save. So touch system.reg now.
    if (open(my $fh, ">", "$ENV{WINEPREFIX}/system.reg"))
    {
        print $fh "WINE REGISTRY Version 2\n";
        close($fh);
    }

    if (!run_inf($windows_version, $params->{scope}, 0))
    {
        CXUtils::cxunlock($lock);
        return undef;
    }

    # Update the registry
    my $start=CXLog::cxtime();
    my $cmd=shquote_string("$ENV{CX_ROOT}/bin/wine") .
            " --no-lock --scope $params->{scope} --wl-app regedit.exe -";
    cxlog("Piping into $cmd\n");
    if (open(my $regedit, "| $cmd"))
    {
        my $proxy_host=$cxconfig->get("BottleDefaults", "HttpProxyHost");
        if (defined $proxy_host)
        {
            my $proxy_port=$cxconfig->get("BottleDefaults", "HttpProxyPort", "8080");
            my $values="\"ProxyEnable\"=dword:1\n" .
                       "\"ProxyServer\"=\"http=$proxy_host:$proxy_port\"\n";
            print $regedit "[HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Internet Settings]\n";
            print $regedit $values;
            print $regedit "[HKEY_USERS\\.Default\\Software\\Microsoft\\Windows\\CurrentVersion\\Internet Settings]\n";
            print $regedit $values;
        }

        my $outlook=$cxconfig->get("BottleDefaults", "OutlookSecurity");
        if ($outlook)
        {
            $outlook=$cxconfig->get("BottleDefaults", "BlockedExtensions");
            print $regedit "[HKEY_CURRENT_USER\\Software\\Wine\\AppDefaults\\Outlook.exe]\n";
            print $regedit "\"DenyShellExecute\"=\"$outlook\"\n";
        }

        close($regedit);
        cxlog("-> rc=$?  (took ", CXLog::cxtime()-$start, " seconds)\n");
        if ($?)
        {
            $@="'$cmd' failed: $?\n";
            CXUtils::cxunlock($lock);
            return undef;
        }
    }
    CXUtils::cxunlock($lock);

    # Notify the bottle hooks
    CXBottle::run_bottle_hooks(["create", $templatedir]);
    return 1;
}


#
# Bottle upgrade
#

sub rename_fake_windows()
{
    my $fake_windows="$ENV{WINEPREFIX}/fake_windows";
    if (-d $fake_windows and !-l $fake_windows)
    {
        # Rename fake_windows to drive_c, but leave a symlinked
        # fake_windows behind in order not to break existing scripts.
        my $drive_c="$ENV{WINEPREFIX}/drive_c";
        if (!rename $fake_windows, $drive_c)
        {
            $@="unable to move '$fake_windows' to '$drive_c': $!\n";
            return undef;
        }
        if (!symlink "drive_c", $fake_windows)
        {
            cxwarn("unable to create '$fake_windows' link: $!\n");
        }

        my $drive_link="$ENV{WINEPREFIX}/dosdevices/c:";
        if (-l $drive_link and readlink($drive_link) eq "../fake_windows")
        {
            unlink $drive_link;
            if (!symlink "../drive_c", $drive_link)
            {
                $@="unable to create '$drive_link' link\n";
                return undef;
            }
        }
    }
    return 1;
}

sub upgrade_cxbottle($$)
{
    my ($cxconfig, $params)=@_;
    my $updater;

    require CXUpgrade;
    my $filename="$ENV{WINEPREFIX}/cxbottle.conf";
    cxlog("\n** Upgrading $filename\n\n");
    my $src=CXUpgrade->new($filename);
    $src=CXUpgrade->new(undef) if (!defined $src);
    my $s=$src->get_section("Bottle");
    my $old_version;
    if (defined $s)
    {
        $old_version=$s->get("Version", "2.0.0");
        unlink "$ENV{WINEPREFIX}/.eval" if ($old_version ne CXUtils::get_product_version());
        $old_version=~s/[^0-9.].*$//;
    }
    else
    {
        $old_version=$params->{old_version} || "1.3.1";
    }
    cxlog("Old Version=$old_version\n");

    my $backup="$filename.$old_version";
    if (-f $filename and !-f $backup)
    {
        require File::Copy;
        if (!File::Copy::copy($filename, $backup))
        {
            cxwarn("unable to write backup to '$backup': $!\n");
        }
    }

    my $template="$datadir/cxbottle.conf";
    my $dst=CXUpgrade->new($template);
    if (!defined $dst)
    {
        cxerr("unable to read '$template': $!\n");
        return ($old_version, $updater);
    }

    # Tweak the configuration values
    my ($encoding, $bottleid);
    $s=$src->get_section("Bottle");
    if (defined $s)
    {
        # Reset the Version field
        $s->remove_field("Version");
        $s->set("Timestamp", $cxconfig->get("CrossOver", "BuildTimestamp", ""));
        $encoding=$s->get("Encoding");

        # Get the updater field so we can return it
        $updater=$s->get("Updater");

        $bottleid=$s->get("BottleID");
    }
    else
    {
        $s=$src->add_section("Bottle");
    }

    # Before these existed, bottles were always considered to be installed
    $s->set("MenuMode", "install") if (!defined $s->get("MenuMode"));
    $s->set("AssocMode", "install") if (!defined $s->get("AssocMode"));
    $s->set("NSPluginMode", "install") if (!defined $s->get("NSPluginMode"));

    # In version < 2.0 the Updater field was called WinePrefixCreate
    # and was stored in cxoffice.conf! So for these antique versions,
    # simply assume that if this looks like a managed bottle,
    # it probably is.
    if (!defined $updater and $old_version lt "2.0" and $> == 0 and
        $ENV{WINEPREFIX} eq "$ENV{CX_ROOT}/support/dotwine")
    {
        $s->set("Updater", "wineprefixcreate");
        $updater="wineprefixcreate";
    }
    if (!defined $bottleid)
    {
        # Make sure the bottle gets an id
        # (in case of an upgrade from a pre-2.0 tree)
        $s->set("BottleID", CXUtils::get_unique_id($ENV{WINEPREFIX}));
    }
    if (($encoding || "ANSI_X3.4-1968") eq "ANSI_X3.4-1968")
    {
        # ANSI_X3.4-1968 is ASCII's little name so we assume that it's
        # going to be compatible with whatever encoding is used by
        # this system.
        $s->set("Encoding", CXUtils::get_system_encoding(1));
    }
    if (!$src->get_filename())
    {
        # Brand new cxbottle.conf, get MenuRoot from $productid.conf
        require CXConfig;
        my $cxconfig=CXConfig->new("$ENV{CX_ROOT}/etc/$productid.conf");
        my $menu_root=$cxconfig->get("BottleDefaults", "MenuRoot");
        $s->set("MenuRoot", $menu_root) if (defined $menu_root);
        my $menu_strip=$cxconfig->get("BottleDefaults", "MenuStrip");
        $s->set("MenuStrip", $menu_strip) if (defined $menu_strip);
    }

    # Do the final merge and write the new configuration file
    $dst->merge($src);
    $dst->write($filename);

    return ($old_version, $updater);
}

sub upgrade_wine_config($)
{
    my ($old_version)=@_;
    my $reg_lines=[];

    my $filename="$ENV{WINEPREFIX}/config";
    return (undef, $reg_lines) if (!-f $filename or $old_version gt "5.0");

    require CXConfig;
    cxlog("\n** Upgrading $filename\n\n");
    my $src=CXConfig->new($filename);

    # Convert the drive settings floppy and cdrom drives
    cxmkpath("$ENV{WINEPREFIX}/dosdevices") if (!-d "$ENV{WINEPREFIX}/dosdevices");
    for (my $i=0; $i<26; $i++)
    {
        my $drive=chr(ord('a')+$i);
        my $path=$src->get("Drive $drive", "Path");
        next if (!defined $path);
        $path=~ s/\${(\w+)}/\%$1\%/g;
        $path=~ s/%(\w+)%/$ENV{$1} || ""/ge;
        $path="../$path" if ($path !~ m%^/%);
        if ($drive ne "c" and ($drive ne "x" or $path ne "/tmp"))
        {
            cxlog("Upgrading $drive: -> $path\n");
            symlink($path, "$ENV{WINEPREFIX}/dosdevices/$drive:");
        }
    }

    # Transfer the Windows version
    my $winver=$src->get("Version", "Windows");

    # Transfer the audio driver settings
    my $value=$src->get("WinMM", "Drivers", "oss");
    $value =~ s/wine([^.]*)\.drv/$1/g;
    push @{$reg_lines}, "[HKEY_CURRENT_USER\\Software\\Wine\\Drivers]\n",
                        "\"Audio\"=\"$value\"\n";

    # Transfer the x11drv section settings
    my $s=$src->get_section("x11drv");
    if (defined $s)
    {
        foreach my $field_name ("ClientSideWithCore",
                                "ClientSideWithRender",
                                "ClientSideAntiAliasWithCore",
                                "ClientSideAntiAliasWithRender")
        {
            $value=$s->get("Dont$field_name");
            $value=(($value || "N") =~ /^N/i ? "Y" : "N");
            if ($value eq "N")
            {
                push @{$reg_lines}, "[HKEY_CURRENT_USER\\Software\\Wine\\X11 Driver]\n",
                                   "\"$field_name\"=\"N\"\n";
            }
        }
    }

    # Transfer the DenyShellExecute setting
    my $deny=$src->get("AppDefaults\\OUTLOOK.EXE", "DenyShellExecute");
    if (!defined $deny or
        # versions <= 3.0.0
        $deny eq "bat;com;exe;hta;lnk;pif;scr;vbe;vbs;wsf;wsh" or
        # versions < 2.0.0
        $deny eq "vbs;wsf;vbe;wsh;hta;bat;pif;exe;scr;lnk")
    {
        my $cxconfig=CXConfig->new("$ENV{CX_ROOT}/etc/$productid.conf");
        my $outlook;
        my $defaults=$cxconfig->get_section("BottleDefaults");
        if (!$defaults)
        {
            # FIXME: We probably have not upgraded the CrossOver
            # configuration file yet so this field is not set... yet.
            # FIXME: We should really upgrade cxoffice.conf first
            $deny="bat;com;cpl;exe;hta;lnk;pif;scr;vbe;vbs;wsf;wsh";
        }
        elsif ($defaults->get("OutlookSecurity"))
        {
            $deny=$defaults->get("BlockedExtensions");
        }
        else
        {
            $deny=undef;
        }
    }
    if ($deny)
    {
        push @{$reg_lines}, "[HKEY_CURRENT_USER\\Software\\Wine\\AppDefaults\\Outlook.exe]\n",
        "\"DenyShellExecute\"=\"$deny\"\n";
    }

    if (!rename $filename, "$filename.$old_version")
    {
        cxwarn("unable to rename '$filename' to '$filename.$old_version'\n");
    }
    return ($winver, $reg_lines);
}

sub upgrade($$$)
{
    my ($self, $cxconfig, $params)=@_;
    $@="";

    return undef if (!rename_fake_windows());

    # Upgrade the configuration files
    my ($old_version, $updater)=upgrade_cxbottle($cxconfig, $params);
    my ($winver, $reg_lines)=upgrade_wine_config($old_version);

    # Setup the drives
    require CXBottle;
    CXBottle::setup_drives($ENV{WINEPREFIX});
    create_windows_label();

    # Grab all the Wine paths we'll need in one go
    my $username;
    my @wine_paths=("c:/Windows/Icons");
    if ($old_version lt "6.0.0")
    {
        push @wine_paths, "c:/Windows/Fonts", "c:/Windows/profiles/crossover";
        $username=(getpwuid($>))[0];
        push @wine_paths, "c:/Windows/profiles/$username" if ($username);
    }
    my @old_binaries=("c:/windows/notepad.exe",
                      "c:/windows/regedit.exe",
                      "c:/windows/system/rundll32.exe",
                      "c:/windows/system/ws2_32.dll"
                     );
    push @wine_paths, @old_binaries if ($old_version lt "6.0.1");

    my $winepath=shquote_string("$ENV{CX_ROOT}/bin/wine") .
                 " --no-convert --scope $params->{scope} --wl-app winepath.exe -- " .
                 join(" ", map { shquote_string($_) } @wine_paths);
    my %unix_paths;
    my $i=0;
    foreach my $path (cxbackquote("$winepath 2>/dev/null", 1))
    {
        chomp $path;
        $unix_paths{$wine_paths[$i]}=$path;
        cxlog("[$wine_paths[$i]] -> [$path]\n");
        $i++;
    }

    my $dir=$unix_paths{"c:/Windows/Fonts"};
    if ($old_version lt "6.0.0" and defined $dir and opendir(my $dh, $dir))
    {
        # We're not doing this in crossover.inf because this gets Wine
        # confused, causing it to build the old-style font metrics
        foreach my $dentry (readdir $dh)
        {
            if ($dentry =~ /\.(ttf|fon)$/i and -f "$dir/$dentry")
            {
                my $ref="$ENV{CX_ROOT}/share/wine/fonts/" . lc($dentry);
                if (-f $ref)
                {
                    cxlog("Deleting '$dir/$dentry'\n");
                    unlink "$dir/$dentry";
                }
            }
        }
        closedir($dh);
    }

    $dir=$unix_paths{"c:/Windows/profiles/crossover"};
    if ($old_version lt "6.0.0" and $username and defined $dir and !-e $dir)
    {
        my $profile=$unix_paths{"c:/Windows/profiles/$username"};
        if (defined $profile and -d $profile)
        {
            cxlog("Renaming '$profile' to 'crossover'\n");
            if (!rename $profile, $dir or !symlink "crossover", $profile)
            {
                cxlog("unable to move and symlink '$profile' to '$dir'\n");
            }
        }
    }

    if ($old_version lt "6.0.1")
    {
        foreach my $wine_path (@old_binaries)
        {
            my $unix_path=$unix_paths{$wine_path};
            next if (!defined $unix_path or !-f $unix_path);
            if (!cxsystem("fgrep -l libwine.so.1 " . shquote_string($unix_path) . " >/dev/null 2>&1"))
            {
                cxlog("Deleting '$unix_path'\n");
                unlink $unix_path;
            }
        }
    }

    if ($old_version lt "6.0.0")
    {
        if (cxsystem("$ENV{CX_ROOT}/bin/wine", "--wl-app", "rundll32.exe",
                     "--scope", $params->{scope},
                     "setupapi.dll,InstallHinfSection", "PreUpgrade", "128",
                     "$datadir/crossover.inf"))
        {
            $@="'rundll32 PreUpgrade crossover.inf' failed\n";
            return undef;
        }
    }
    if ($old_version lt "7.0")
    {
        # FIXME: It would be nicer to do it using crossover.inf but Wine's
        #        setupapi does not know how to recursively delete registry
        #        keys (see bug 13548).
        my $cmd=shquote_string("$ENV{CX_ROOT}/bin/wine") .
                " --scope $params->{scope} --wl-app regedit.exe -";
        cxlog("Piping into $cmd\n");
        if (open(my $regedit, "| $cmd"))
        {
            print $regedit "[-HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Print\\Environments\\Windows 4.0\\Drivers]\n";
            close($regedit);
            if ($?)
            {
                $@="'$cmd' failed: $?\n";
                return undef;
            }
        }
    }
    if ($old_version lt "9.0")
    {
        # remove the Locale key inorder to recreate the proper locale strings
        # dont do this with reg_lines as I want it to happen before the INF
        my $cmd=shquote_string("$ENV{CX_ROOT}/bin/wine") .
                " --scope $params->{scope} --wl-app regedit.exe -";
        cxlog("Piping into $cmd\n");
        if (open(my $regedit, "| $cmd"))
        {
            print $regedit "[HKEY_CURRENT_USER\\Control Panel\\International\n";
            print $regedit "\"Locale\"=\"\"\n";
            close($regedit);
            if ($?)
            {
                $@="'$cmd' failed: $?\n";
                return undef;
            }
        }
    }
    return undef if (!run_inf($winver || $windows_version, $params->{scope}, 1));

    $dir=$unix_paths{"c:/Windows/Icons"};
    if (defined $dir and opendir(my $dh, $dir))
    {
        foreach my $dentry (readdir $dh)
        {
            if ($dentry !~ /^\.\.?$/ and -d "$dir/$dentry")
            {
                cxlog("Deleting the '$dir/$dentry' directory\n");
                require File::Path;
                if (!File::Path::rmtree("$dir/$dentry"))
                {
                    cxwarn("unable to delete the '$dir/$dentry' directory\n");
                }
            }
        }
        closedir($dh);
    }

    # Wine now has native support for $http_proxy
    if ($old_version lt "7.0" and defined $ENV{http_proxy})
    {
        push @{$reg_lines},
            "[HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Internet Settings]\n",
            "\"ProxyServer\"=-\n",
            "\"ProxyEnable\"=-\n",
            "[HKEY_USERS\\.Default\\Software\\Microsoft\\Windows\\CurrentVersion\\Internet Settings]\n",
            "\"ProxyServer\"=-\n",
            "\"ProxyEnable\"=-\n";
    }

    # Upgrade the registry
    if (@$reg_lines)
    {
        my $cmd=shquote_string("$ENV{CX_ROOT}/bin/wine") .
                " --scope $params->{scope} --wl-app regedit.exe -";
        cxlog("Piping into $cmd\n");
        cxlog(@$reg_lines);
        if (open(my $regedit, "| $cmd"))
        {
            print $regedit @$reg_lines;
            close($regedit);
            if ($?)
            {
                $@="'$cmd' failed: $?\n";
                return undef;
            }
        }
    }

    # Old versions used to hardcode the external browser / email tools names
    if ($old_version lt "7.0")
    {
        if (cxsystem("$ENV{CX_ROOT}/bin/wine", "--wl-app", "rundll32.exe",
                     "--scope", $params->{scope},
                     "setupapi.dll,InstallHinfSection", "WineBrowserUpgrade",
                     "128", "$datadir/crossover.inf"))
        {
            $@="'rundll32 WineBrowserUpgrade crossover.inf' failed\n";
            return undef;
        }
    }

    # This should only be needed for 1.2 and lower bottles but they did not
    # have reliable version numbers then so we do it for 1.3.1 or older.
    if ($old_version le "1.3.1")
    {
        cxsystem("$ENV{CX_ROOT}/bin/wine", "--scope", $params->{scope}, "hackmsi.exe");
    }

    if ($old_version lt "5.0")
    {
        # If the same user who owns the bottle installed CrossOver,
        # then the legacy menus, associations and plugins have already been
        # removed. Otherwise we must do it.
        if (!-o $ENV{CX_ROOT})
        {
            cxsystem("$ENV{CX_ROOT}/bin/cxmenu", "--scope", $params->{scope},
                     "--removeall", "--pattern", "legacy", "--ignorelist", "");
            cxsystem("$ENV{CX_ROOT}/bin/cxassoc", "--scope", $params->{scope},
                     "--removeall", "--pattern", "legacy", "--ignorelist", "");
            cxsystem("$ENV{CX_ROOT}/bin/cxnsplugin", "--scope", $params->{scope},
                     "--removeall", "--pattern", "legacy");
        }

        # Recreate the menus, associations and plugins from scratch
        cxsystem("$ENV{CX_ROOT}/bin/cxmenu", "--scope", $params->{scope},
                 "--sync", "--mode", "install");
        my $mode="mime;default=$default_eassocids;alternative=$alternative_eassocids";
        $mode =~ s/\./\\./g;
        cxsystem("$ENV{CX_ROOT}/bin/cxassoc", "--scope", $params->{scope},
                 "--sync", "--mode", $mode);
        cxsystem("$ENV{CX_ROOT}/bin/cxnsplugin", "--scope", $params->{scope},
                 "--sync", "--mode", "install;ignore=$ignore_plugins");
    }
    elsif ($old_version lt "6.0.1")
    {
        # Reinstall the menus and associations because the icons moved around
        # For menus --removeall is needed to fix the implicit menus icons
        cxsystem("$ENV{CX_ROOT}/bin/cxmenu", "--scope", $params->{scope},
                 "--removeall", "--install");
        cxsystem("$ENV{CX_ROOT}/bin/cxassoc", "--scope", $params->{scope}, "--install");
    }
    elsif ($old_version lt "6.2")
    {
        # Remove, redetect and reinstall the plugins because the naming
        # scheme has changed
        cxsystem("$ENV{CX_ROOT}/bin/cxnsplugin", "--scope", $params->{scope},
                 "--removeall", "--sync",
                 "--mode", "install;ignore=$ignore_plugins");
    }

    # Notify the bottle hooks
    CXBottle::run_bottle_hooks(["upgrade-from", $old_version]);

    return 1;
}


#
# Bottle resync
#

sub needs_resync($$)
{
    # This type of bottle never needs to be resync-ed with an external source
    return 0;
}

return 1;
