# (c) Copyright 2009. CodeWeavers, Inc.

import os
import subprocess

import cxobjc
import distversion

import bottlequery
import bottlemanagement
import appdetector
import c4profilesmanager
import cxutils

# for localization
from cxutils import cxgettext as _

class BottleWrapper(cxobjc.Proxy):

    def __init__(self, inName):
        cxobjc.Proxy.__init__(self)
        self._change_delegate = None
        self._bottle_info_ready = False
        self._property_dict = {}
        self.current_description = None
        self.bottletemplate = ""

        self._control_panel_ready = False
        self._control_panel_table = []
        self._control_panel_off_thread_table = []

        self.name = inName
        self.changeablename = inName

        self._installed_packages = {}
        self.installedPackageValues = []
        self._installed_packages_off_thread = {}
        self._installed_packages_ready = False

        self.status = _("Initializing")
        self.status_overrides = []

        # mtime values:
        #  0 = not yet updated
        #  -1 = no config file
        #  None = deleted
        #  anything else = mtime of config file
        self.mtime = 0

        # This is something we need to know immediately for 
        #  Mac initialization purposes.  Hopefully it's not too
        #  expensive to do here.
        self.is_managed = bottlequery.is_managed(self.name)
        self.refresh_up_to_date()


        self.marked_for_death = False

        # Let's learn a bit about ourselves.  Everything here
        #  should be very quick.
        self.wine_prefix = bottlequery.get_prefix_for_bottle(self.name)
        self.system_drive = os.path.join(self.wine_prefix, 'drive_c')
        if self.name == bottlequery.get_default_bottle():
            self.is_default = True
        else:
            self.is_default = False

        self.pre_load_basic_info()
        self.load_basic_info()
        self.post_load_basic_info()


    # This is a special initializer for objc.  It must always be called
    #  explicitly on the mac.
    def initWithName_(self, inName):
        self = cxobjc.Proxy.nsobject_init(self)
        if self is not None:
            self.__init__(inName)
        return self

    # This will be called, below, only on ObjC.  It sets up
    #  some miscellaneous notifications that the GUI needs.
    @classmethod
    def prepare_kvo_notification(cls):
        # The setKeys_... functions are implemented in pyobjc and thus
        # invisible to pylint on non-Mac platforms.
        # pylint: disable-msg=E1101
        cls.setKeys_triggerChangeNotificationsForDependentKey_(["status","up_to_date","marked_for_death"],"can_run_commands")
        cls.setKeys_triggerChangeNotificationsForDependentKey_(["name","is_default"],"display_name")
        cls.setKeys_triggerChangeNotificationsForDependentKey_(["name"],"changeablename")
        cls.setKeys_triggerChangeNotificationsForDependentKey_(["status","up_to_date"],"is_active")
        cls.setKeys_triggerChangeNotificationsForDependentKey_(["is_managed","marked_for_death"],"can_edit")



    #####
    #
    # load basic info
    #
    #####
        
    def pre_load_basic_info(self):
        """This function is pretty fast and MUST be called in the main thread,
        BEFORE calling load_basic_info().
        """
        self.add_status_override(_("Initializing"))
        self.bottle_info_ready = False


    def load_basic_info(self):
        """This function loads the bottle's basic properties.

        It takes quite a bit of time to run and thus must be run in the
        background. Thus, for purposes of thread management, it is divided into
        three parts, with pre_ and post_ functions that are cheap and need to
        be called on the main thread, in order to properly arrange the state of
        our properties.
        """
        self.set_property_dict(bottlequery.get_bottle_properties(self.name))


    def post_load_basic_info(self):
        """This function is pretty fast and MUST be called in the main thread,
        AFTER load_basic_info() has returned.
        """
        if (self._property_dict):
            # copy everything we've learned out of the 
            #  property dict and into actual properties.
            self.bottletemplate = self._property_dict["template"]

            if "managed" in self._property_dict:
                self.is_managed = self._property_dict["managed"]
            else:
                self.is_managed = False

            self.current_description = self._property_dict["description"]

            # Notify everyone that we're ready to go.
            self.bottle_info_ready = True
        else:
            self.bottle_info_ready = False

        self.remove_status_override(_("Initializing"))

    # Aliases for objective C:
    preLoadBasicInfo = pre_load_basic_info
    loadBasicInfo = load_basic_info
    postLoadBasicInfo = post_load_basic_info

    def _get_scope(self):
        if self.is_managed:
            return _("Managed")
        else:
            return _("Private")

    
    scope = property(_get_scope)


    def _maybe_upgrade(self):
        # If we get the control panel or installed app info for a bottle that
        # needs upgraded, this upgrades the bottle, which updates the mtime of
        # cxbottle.conf. This would cause us to incorrectly assume our
        # information is out of date, so if we need an upgrade we trigger it
        # explicitly and ignore the mtime change.
        stub_needs_upgrade = False
        if self.is_managed:
            stub_needs_upgrade = not bottlemanagement.get_up_to_date(self.name, "private")

        if not self.up_to_date or stub_needs_upgrade:
            args = [os.path.join(cxutils.CX_ROOT, "bin", "wine"),
                    "--bottle", self.name, '--no-gui',
                    "--ux-app", "true"]
            cxutils.system(args)
            self.mtime = self._real_mtime()
            self.refresh_up_to_date()


    #####
    #
    # load control panel info
    #
    # Three sections, same as load_basic_info, above.
    #
    # Loads the list of control panels available for this bottle,
    # and sets up icons and other user-readable-data for each.
    #
    #####

    def pre_load_control_panel_info(self):
        """This function is pretty fast and MUST be called in the main thread,
        BEFORE calling load_control_panel_info().
        """
        self.control_panel_ready = False
        self.add_status_override(_("Initializing"))

    def load_control_panel_info(self):
        """This function initializes the bottle's list of control panel applets.

        It takes quite a bit of time to run and thus must be run in the
        background. Thus, for purposes of thread management, it is divided into
        three parts, with pre_ and post_ functions that are cheap and need to
        be called on the main thread, in order to properly arrange the state of
        our properties.
        """
        self._maybe_upgrade()

        self.control_panel_off_thread_table = bottlequery.get_control_panel_info(self.name)

        if not distversion.IS_MACOSX:
            self.control_panel_off_thread_table.append(["cxassocedit", _("Edit Associations"),
                _("Manage the Windows programs used to open files in the native environment"), "", bottlequery.resolve_icon_path("cx_edit_assoc.png", self.name)])

            self.control_panel_off_thread_table.append(["cxmenuedit", _("Edit Menus"),
                _("Manage the menus and desktop icons in this bottle"), "", bottlequery.resolve_icon_path("cx_edit_menu.png", self.name)])

            if distversion.HAS_PLUGIN:
                self.control_panel_off_thread_table.append(["cxpluginedit", _("Edit Plugins"),
                    _("Manage the browser plugins in this bottle"), "", bottlequery.resolve_icon_path("cx_edit_plugin.png", self.name)])


    def post_load_control_panel_info(self):
        """This function is pretty fast and MUST be called in the main thread,
        AFTER load_control_panel_info() has returned.
        """
        cptable = []
        for panel in self.control_panel_off_thread_table:
            cpdict = {}
            cpdict["exe"] = panel[0]
            cpdict["name"] = panel[1]
            cpdict["description"] = panel[2]
            if distversion.IS_MACOSX:
                # conveniently, cxcplinfo provides us with .ico files in addition
                #  to the .xpms which the Mac can't handle.
                base, ext = os.path.splitext(panel[4])
                if ext.lower() == ".xpm":
                    ext = ".ico"
                cpdict["icon"] = base + ext
            else:
                cpdict["icon"] = panel[4]
            cptable.append(cpdict)
        self.control_panel_table = cptable
        self.control_panel_ready = True
        self.remove_status_override(_("Initializing"))


    def launch_control_panel_applet(self, applet):
        wine = os.path.join(cxutils.CX_ROOT, "bin", "wine")
        if applet.lower().endswith(".exe.so"):
            args = [wine, "--bottle", self.name, "--wl-app", applet]
        elif applet.lower().endswith(".cpl"):
            args = [wine, "--bottle", self.name, "--wl-app", "rundll32", "shell32.dll,Control_RunDLL", applet]
        else:
            args = [os.path.join(cxutils.CX_ROOT, "bin", applet), "--bottle", self.name]

        subprocess.Popen(args, close_fds=True)


    # Aliases for objective C:
    preLoadControlPanelInfo = pre_load_control_panel_info
    loadControlPanelInfo = load_control_panel_info
    postLoadControlPanelInfo = post_load_control_panel_info
    launchControlPanelApplet_ = launch_control_panel_applet


    #####
    # load installed applications
    #
    # Three sections, same as load_basic_info, above.
    #
    #
    # Loads the list of applications installed in the bottle.
    # This can take quite a while.
    #
    #####
    def pre_load_installed_applications(self):
        """This function is pretty fast and MUST be called in the main thread,
        BEFORE calling load_installed_applications().
        """
        self.installed_packages_ready = False
        self.installed_packages_off_thread = {}
        self.add_status_override(_("Initializing"))

    def load_installed_applications(self):
        """This function determines the list of applications that are installed
        in the bottle.

        It takes quite a bit of time to run and thus must be run in the
        background. Thus, for purposes of thread management, it is divided into
        three parts, with pre_ and post_ functions that are cheap and need to
        be called on the main thread, in order to properly arrange the state of
        our properties.
        """
        self._maybe_upgrade()

        profiles = c4profilesmanager.C4ProfilesSet.all_profiles()
        self.installed_packages_off_thread = appdetector.fast_get_installed_applications(self.name, profiles)

    def post_load_installed_applications(self):
        """This function is pretty fast and MUST be called in the main thread,
        AFTER load_installed_applications() has returned.
        """
        self.installed_packages = self.installed_packages_off_thread
        self.installed_packages_ready = True
        self.remove_status_override(_("Initializing"))


    # Aliases for objective C:
    preLoadInstalledApplications = pre_load_installed_applications
    loadInstalledApplications = load_installed_applications
    postLoadInstalledApplications = post_load_installed_applications

    def change_description(self, inDescription):
        self.current_description = inDescription

    def write_description(self):
        #  Note:  Expensive!  Should only be done within an operation.
        if self._property_dict["description"]  != self.current_description:
            bottlemanagement.set_bottle_description(self.name, self.current_description)
            self._property_dict["description"] = self.current_description


    def pre_rename(self):
        if self.changeablename == self.name:
            return 

        self.add_status_override(_("Renaming"))
        self.marked_for_death = True

    def rename(self):
        if self.changeablename == self.name:
            return

        (success, err) = bottlemanagement.rename_bottle(self.name, self.changeablename)
        if not success:
            print "Rename from %s to %s failed:  %s" % (self.name, self.changeablename, err)
            self.changeablename = self.name
        return success, err

    def post_rename(self):
        if self.changeablename == self.name:
            return

        self.name = self.changeablename
        self.wine_prefix = bottlequery.get_prefix_for_bottle(self.name)
        self.system_drive = os.path.join(self.wine_prefix, 'drive_c')
        self.remove_status_override(_("Renaming"))
        self.marked_for_death = False

    preRename = pre_rename
    postRename = post_rename

    def get_control_panel_table(self):
        return self._control_panel_table

    def set_control_panel_table(self, in_table):
        self._control_panel_table = in_table
        if self.change_delegate:
            self.change_delegate.bottleChanged(self)

    control_panel_table = property(get_control_panel_table, set_control_panel_table)

    def get_control_panel_ready(self):
        return self._control_panel_ready

    def set_control_panel_ready(self, inBool):
        self._control_panel_ready = inBool
        if self.change_delegate:
            self.change_delegate.bottleChanged(self)

    control_panel_ready = property(get_control_panel_ready, set_control_panel_ready)


    def get_installed_packages(self):
        return self._installed_packages

    def set_installed_packages(self, installed_apps):
        self._installed_packages = installed_apps
        # installedPackageValues stores the same data but as an array.
        # We need this for bindings, and having an actual member
        # allows us to use KVO properly.
        self.installedPackageValues = installed_apps.values()
        if self.change_delegate:
            self.change_delegate.bottleChanged(self)

    installed_packages = property(get_installed_packages, set_installed_packages)

    def get_installed_packages_ready(self):
        return self._installed_packages_ready

    def set_installed_packages_ready(self, inBool):
        self._installed_packages_ready = inBool
        if self.change_delegate:
            self.change_delegate.bottleChanged(self)

    installed_packages_ready = property(get_installed_packages_ready, set_installed_packages_ready)

    def set_property_dict(self, inDict):
        self._property_dict = inDict

        # if current_description varies from the
        #  dict value, we know it's dirty and needs
        #  to be written.
        if "description" in inDict:
            self.current_description = inDict["description"]
        else:
            self.current_description = None

        if self.change_delegate:
            self.change_delegate.bottleChanged(self)

    def _real_mtime(self):
        config_file_path = bottlequery.config_file_path(self.name)
        try:
            mtime = os.path.getmtime(config_file_path)
        except OSError:
            # distinguish between a bottle with no config file and a deleted bottle
            if os.path.exists(self.wine_prefix):
                mtime = -1
            else:
                mtime = None
        return mtime

    def get_needs_update(self, updating=False):
        if self.status_overrides:
            return False

        mtime = self._real_mtime()

        if self.mtime != mtime:
            if updating:
                self.mtime = mtime
            return True
        else:
            return False

    needs_update = property(get_needs_update)

    def get_bottle_info_ready(self):
        return self._bottle_info_ready

    def set_bottle_info_ready(self, inBool):
        self._bottle_info_ready = inBool
        if self.change_delegate:
            self.change_delegate.bottleChanged(self)

    bottle_info_ready = property(get_bottle_info_ready, set_bottle_info_ready)

    def add_status_override(self, status):
        self.status_overrides.append(status)
        self.status=status
        if self.change_delegate:
            self.change_delegate.bottleChanged(self)

    addStatusOverride_ = add_status_override

    def remove_all_status_overrides(self, status):
        while status in self.status_overrides:
            self.remove_status_override(status)

    def remove_status_override(self, status):
        if status in self.status_overrides:
            self.status_overrides.remove(status)
        if len(self.status_overrides) > 0:
            self.status = self.status_overrides[len(self.status_overrides)-1]
        elif not self.bottle_info_ready or not self.control_panel_ready or not self.installed_packages_ready:
            self.status = _("Initializing")
        else:
            self.status =  _("Ready")
        if self.change_delegate:
            self.change_delegate.bottleChanged(self)

    removeStatusOverride_ = remove_status_override

    def is_ready(self):
        return self.bottle_info_ready and not self.status_overrides

    def get_change_delegate(self):
        return self._change_delegate

    def set_change_delegate(self, inDelegate):
        self._change_delegate = inDelegate

    change_delegate = property(get_change_delegate, set_change_delegate)

    def get_can_run_commands(self):
        return self.up_to_date and not self.marked_for_death and _("Shutting down") not in self.status_overrides

    can_run_commands = property(get_can_run_commands)

    def get_can_edit(self):
        return self.up_to_date and not self.marked_for_death

    can_edit = property(get_can_edit)

    def get_is_active(self):
        return self.is_ready and self.up_to_date and _("Shutting down") not in self.status_overrides

    is_active = property(get_is_active)

    # For ObjC:
    def setMarkedForDeath_(self, inBool):
        if (inBool):
            self.add_status_override("Deleting")
        self.marked_for_death = inBool

    def is_running(self):
        return bottlemanagement.is_running(self.name)

    def quit(self):
        self.add_status_override(_("Shut down"))
        self.add_status_override(_("Shutting down"))
        success =  bottlemanagement.quit_bottle(self.name)
        self.remove_status_override(_("Shutting down"))
        if not success:
            self.remove_status_override(_("Shut down"))
        return success

    def refresh_up_to_date(self):
        if self.is_managed:
            self.up_to_date = bottlemanagement.get_up_to_date(self.name, "managed")
        else:
            self.up_to_date = bottlemanagement.get_up_to_date(self.name)

        if self.is_managed:
            if not self.up_to_date:
                self.add_status_override(_("Needs upgrade"))
            else:
                self.remove_all_status_overrides(_("Needs upgrade"))


    refreshUpToDate = refresh_up_to_date

    def force_quit(self):
        self.add_status_override(_("Shut down"))
        self.add_status_override(_("Forcing shutdown"))
        success =  bottlemanagement.kill_bottle(self.name)
        self.remove_status_override(_("Forcing shutdown"))
        if not success:
            self.remove_status_override(_("Shut down"))
        return success

    forceQuit = force_quit

    # Cancel shutdown is called to restore the status of a bottle
    #  to a non-deleting, non-dying, non-shutting-down status.
    def cancelShutdown(self):
        self.remove_status_override(_("Shut down"))
        self.remove_status_override(_("Forcing shutdown"))
        self.remove_status_override(_("Shutting down"))
        self.remove_status_override(_("Deleting"))

    def get_display_name(self):
        if self.is_default:
            return "%s %s" % (_("(default)"), self.name)
        else:
            return self.name

    display_name = property(get_display_name)

    # A couple of crutches for reverse-compatibility with the old BottleWrapper.m
    def bottleName(self):
        return self.name
   
    displayName = get_display_name


if distversion.IS_MACOSX:
    BottleWrapper.prepare_kvo_notification()
