# (c) Copyright 2009-2010. CodeWeavers, Inc.

"""Defines classes to represent and manipulate the C4 profile information."""

import os
import re

import cxlog
import cxutils
import decorators
import distversion

# for localization
from cxutils import cxgettext as _

# A mapping of the language code to the plain English name for use by the GUI.
LANGUAGES = {
    '':      _('Default'),
    'ar':    _('Arabic'),
    'bg':    _('Bulgarian'),
    'ca':    _('Catalan'),
    'cs':    _('Czech'),
    'da':    _('Danish'),
    'de':    _('German'),
    'el':    _('Greek'),
    'en':    _('English'),
    'es':    _('Spanish'),
    'et':    _('Estonian'),
    'eu':    _('Basque'),
    'fi':    _('Finnish'),
    'fr':    _('French'),
    'he':    _('Hebrew'),
    'hi':    _('Hindi'),
    'hr':    _('Croatian'),
    'hu':    _('Hungarian'),
    'id':    _('Indonesian'),
    'it':    _('Italian'),
    'ja':    _('Japanese'),
    'kk':    _('Kazakh'),
    'ko':    _('Korean'),
    'lt':    _('Lithuanian'),
    'lv':    _('Latvian'),
    'ms':    _('Malay (Malaysia)'),
    'nb':    _('Norwegian (Bokmal)'),      # no
    'nl':    _('Dutch'),
    'pl':    _('Polish'),
    'pt':    _('Portuguese (Portugal)'),   # pt-pt
    'pt-br': _('Portuguese (Brazil)'),
    'ro':    _('Romanian'),
    'ru':    _('Russian'),
    'sk':    _('Slovak'),
    'sl':    _('Slovenian'),
    'sr':    _('Serbian (Latin)'),
    'sv':    _('Swedish'),
    'th':    _('Thai'),
    'tr':    _('Turkish'),
    'uk':    _('Ukrainian'),
    'ur':    _('Urdu'),
    'vi':    _('Vietnamese'),
    'zh-cn': _('Chinese (Simplified)'),
    'zh-hk': _('Chinese (Hong Kong SAR)'),
    'zh-tw': _('Chinese (Traditional)'),
    }


# This is the set of environment variables which are allowed in expandable
# profile strings. The values are given for documentation only.
ENVIRONMENT_VARIABLES = (
    'accessories',     # 'Programs\Accessories'
    'allusersprofile', # 'c:\users\Public'
    'appdata',         # 'c:\users\crossover\Application Data'
    'comspec',         # 'c:\windows\system32\cmd.exe'
    'desktop',         # 'c:\users\crossover\Desktop'
    'programfiles',    # 'c:\Program Files'
    'programs',        # 'Programs'
    'startmenu',       # 'c:\users\crossover\Start Menu'
    'systemdrive',     # 'c:'
    'systemroot',      # 'c:\windows'
    'temp',            # 'c:\windows\temp'
    'userprofile',     # 'c:\users\crossover'
    'windir',          # 'c\windows'
    'winsysdir',       # 'c:\windows\system32'
    )



#####
#
# Helper functions for validating restricted file globs
#
#####

_RESTRICTED_GLOB_RE = re.compile('(?:^|/)\.[.?*]+(?:/|$)')

def _validate_file_globs(globs, location):
    """Check that the file globs don't try to escape the directory they are
    applied to.

    We do that by simply forbidding any path component that could match '..'.
    """
    # FIXME: If the string is expandable, then we could get stuff like
    #      '/.%Empty%./' that would be expanded to '/../'. It's expected that
    # the allowed environment variables won't start or end with a '.'. They
    # could end with a '/' or '\\' which could be exploited.
    for glob in globs:
        if _RESTRICTED_GLOB_RE.search(glob):
            raise AttributeError("invalid file glob '%s' in %s" % (glob, location))

def _validate_file_paths(paths, location):
    """Check that the file paths don't try to escape the directory they are
    applied to.
    """
    # FIXME: If the string is expandable, then we could get stuff like
    #      '/.%Empty%./' that would be expanded to '/../'
    for path in paths:
        # For now we simply apply the same rules as for globs since
        # '?' and '*' are not valid Windows file characters anyway.
        if _RESTRICTED_GLOB_RE.search(path):
            raise AttributeError("invalid file path '%s' in %s" % (path, location))



#####
#
# Helper functions for dumping the profiles
#
#####

def _dump_value(out, indent, value):
    if hasattr(value, 'dump'):
        out.write(" %s\n" % type(value))
        value.dump(out, indent + "| ")
    else:
        out.write(" %s\n" % repr(value))

def _dump_fields(out, indent, obj, fields):
    for field in fields:
        if field not in obj.__dict__:
            continue
        value = obj.__dict__[field]
        if isinstance(value, dict):
            keys = value.keys()
            keys.sort()
            for key in keys:
                out.write("%s%s{%s} =" % (indent, field, key))
                _dump_value(out, indent, value[key])
        elif isinstance(value, list):
            for i in xrange(len(value)):
                out.write("%s%s[%d] =" % (indent, field, i))
                _dump_value(out, indent, value[i])
        else:
            out.write("%s%s =" % (indent, field))
            _dump_value(out, indent, value)



#####
#
# Helper functions for copying/updating C4 profile objects
#
#####

def _copy_value(value):
    try:
        copy_func = value.copy
    except AttributeError:
        if type(value) is list:
            return list(value)
        else:
            return value
    else:
        return copy_func()

def _copy_values(attributes):
    for key, value in attributes.iteritems():
        if key != 'parent':
            attributes[key] = _copy_value(value)

def _update_value(this, oth):
    try:
        update_func = this.update
    except AttributeError:
        if type(this) is list:
            this.extend(oth)
            return this
        else:
            return oth
    else:
        result = update_func(oth)
        if result is None:
            return this
        else:
            return result

def _update_values(attributes, others, exclude=()):
    for key, value in others.iteritems():
        if key in exclude:
            continue
        if key in attributes:
            attributes[key] = _update_value(attributes[key], value)
        else:
            attributes[key] = _copy_value(value)


#####
#
# C4Condition
#
#####

class C4Condition(object):
    """Represents a boolean expression."""

    @decorators.abstractmethod
    def is_true(self, _properties):
        return False


class C4ConditionUnary(C4Condition):
    """A condition node with a single sub-node."""

    # A node to use to compute this condition. This field is mandatory and
    # thus defaults to None.
    child = None

    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'child' not in self.__dict__:
            raise AttributeError("C4ConditionUnary.child is not set for application %s" % app_name)
        self.child.validate(app_name)

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('child',))


    ### Condition evaluation

    def is_true(self, properties):
        return self.child.is_true(properties)


class C4ConditionNot(C4ConditionUnary):
    """Inverts the value of the child condition node."""

    ### Condition evaluation

    def is_true(self, properties):
        return not self.child.is_true(properties)


class C4ConditionNary(C4Condition):
    """A condition node that merges the result of multiple condition subnodes.
    """

    # A list of nodes node to use to compute this condition. This field is
    # mandatory and thus defaults to [].
    children = []

    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if len(self.children) == 0:
            raise AttributeError("C4ConditionNary.children is empty for application %s" % app_name)
        for child in self.children:
            child.validate(app_name)

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('children',))


class C4ConditionAnd(C4ConditionNary):
    """Returns True if all the children nodes evaluate to True."""

    def is_true(self, properties):
        for child in self.children:
            if not child.is_true(properties):
                return False
        return True


class C4ConditionOr(C4ConditionNary):
    """Returns False if all the children nodes evaluate to False."""

    def is_true(self, properties):
        for child in self.children:
            if child.is_true(properties):
                return True
        return False


class C4ConditionEqual(C4Condition):
    """Returns True if the specified property has the expected value."""

    # The property name. This field is mandatory and thus defaults to None.
    name = None

    # The property's expected value.
    # This field is mandatory and thus defaults to None.
    value = None

    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'name' not in self.__dict__:
            raise AttributeError("C4ConditionEqual.name is not set for application %s" % app_name)
        if 'value' not in self.__dict__:
            raise AttributeError("C4ConditionEqual.value is not set for application %s" % app_name)

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('name', 'value'))


    ### Condition evaluation

    def is_true(self, properties):
        return self.value == properties.get(self.name, "")


class C4ConditionMatch(C4Condition):
    """Returns True if the specified property matches the given regular expression."""

    # The property name. This field is mandatory and thus defaults to None.
    name = None

    # The regular expression that the property should match.
    # This field is mandatory and thus defaults to None.
    regexp = None

    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'name' not in self.__dict__:
            raise AttributeError("C4ConditionEqual.name is not set for application %s" % app_name)
        if 'regexp' not in self.__dict__:
            raise AttributeError("C4ConditionEqual.regexp is not set for application %s" % app_name)

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('name', 'regexp'))


    ### Condition evaluation

    def is_true(self, properties):
        try:
            return re.match(self.regexp, properties.get(self.name, ""))
        except re.error:
            cxlog.warn("useif match pattern %s for name %s is an invalid regular expression" % (cxlog.debug_str(self.regexp), cxlog.debug_str(self.name)))
            return None



#####
#
# C4CDProfile class
#
#####

class C4CDGlob(object):
    """Contains a specific glob for detecting an application CD."""

    ### Property defaults

    # A string containing a glob to identify the files to analyze. This field
    # is mandatory and thus defaults to none.
    glob = None

    # A list of regular expressions that the selected file must all match.
    # If the list is empty, then it's enough for the file to just exist.
    patterns = tuple()


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'glob' not in self.__dict__:
            raise AttributeError("C4CDGlob.glob is not set for application %s" % app_name)
        _validate_file_globs((self.glob,), "the C4CDGlob.glob field of %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('glob', 'patterns'))


class C4CDProfile(object):
    """Contains the C4 CD Profile information."""

    ### Property defaults

    # Points to the umbrella C4Profile object that ties a C4 entry's
    # specialized profiles together. This field is mandatory and thus defaults
    # to none.
    parent = None

    # A list of C4CDGlobs to be used to recognize a given CD. This field is
    # mandatory and thus defaults to None.
    globs = None


    def copy(self):
        result = C4CDProfile()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__, exclude=('parent',))


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if not self.parent:
            raise AttributeError("C4CDProfile.parent is not set for application %s" % app_name)
        if not self.globs:
            raise AttributeError("C4CDProfile.globs is not set for application %s" % app_name)
        for glob in self.globs:
            glob.validate(app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('globs',))



#####
#
# C4ApplicationProfile class
#
#####


class C4RegistryGlob(object):
    """Contains a specific registry glob for detecting an installed
    application.
    """

    ### Property defaults

    # A string containing a glob to identify the registry keys to analyze.
    # This field is mandatory and thus has no default.
    # key_glob = ""

    # A string containing a regular expression to identify an optional key
    # value to analyze. If left empty, then the values are ignored. Note that
    # a key's default value is called '@'.
    value_pattern = ""

    # A string containing a regular expression to check if the value contains
    # the expected data. If left empty, then the presence of the value is
    # enough.
    data_pattern = ""


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'key_glob' not in self.__dict__:
            raise AttributeError("C4RegistryGlob.key_glob is not set for application %s" % app_name)
        if self.data_pattern != "" and self.value_pattern == "":
            raise AttributeError("C4RegistryGlob.data_pattern is set but not the corresponding value_pattern for application %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self,
                     ('key_glob', 'value_pattern', 'data_pattern'))


class C4ApplicationProfile(object):
    """Contains the C4 Application Profile information."""

    ### Enumerated types

    # A set of all known medals
    ALL_MEDALS = set(('gold', 'silver', 'bronze',
                      'ungold', 'unsilver', 'unbronze',
                      'knownnottowork', 'untested'))

    # A set of the medals for supported applications
    SUPPORTED_MEDALS = set(('gold', 'silver', 'bronze'))

    # A set of medals not known to work
    BRONZE_PLUS_MEDALS = set(('gold', 'silver', 'bronze',
                              'ungold', 'unsilver', 'unbronze'))

    # Maps trusted medals to their untrusted equivalent
    UNTRUSTED_MEDAL_MAP = {'gold':'ungold',
                           'silver':'unsilver',
                           'bronze':'unbronze'}

    ### Property defaults

    # Points to the umbrella C4Profile object that ties a C4 entry's
    # specialized profiles together. This field is mandatory and thus defaults
    # to None.
    parent = None

    # A dictionary mapping the language to the application description. The
    # special '' mapping corresponds to the default application description.
    _description = {'': ''}

    # The C4 application category string. This field is mandatory and thus
    # defaults to None.
    category = None

    # A dictionary mapping the platform to the application medal string for
    # that platform. Each medal must be in the all_medals set.
    _medals = {}

    # A set of flag strings.
    flags = set()

    # If we know how to install this application, then this is the id of the
    # C4 entry that has the installation information.
    installer_id = None

    # If this regular expression matches an Uninstall key, then it means this
    # application is installed.
    installed_key_pattern = None

    # If this regular expression matches an Uninstall key's display name,
    # then it means this application is installed.
    installed_display_pattern = None

    # A list of C4RegistryGlobs to look for to check if the application is
    # installed.
    installed_registry_globs = tuple()

    # A list of strings containing the file globs to look for to check if the
    # application is installed.
    installed_file_globs = tuple()

    # The list of bottle types the application can be installed in.
    # The first entry is the default one.
    bottle_types = {'use': ("winxp",), 'install': ()}

    # Used to group related applications in the same bottle.
    application_group = ""

    # Maps a language to the URL where one can get the corresponding
    # installer.
    download_urls = {}

    # Maps a language to the URL of a web page from which the installer can
    # be manually downloaded.
    download_page_urls = {}

    # A set of application ids that this is an upgrade or add-on for. Note
    # that this is not a dependency but a recommendation system. This can be
    # used for the following scenarios:
    # - A component which is an upgrade for another component.
    # - A recommended 'upgrade' which is a full install and thus does not
    #   depend on the item it upgrades.
    # - A codec / plugin / add-on which we recommend installing, but which
    #   technically is not an upgrade.
    extra_fors = set()


    def copy(self):
        result = C4ApplicationProfile()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__, exclude=('parent',))


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.
        """
        if not self.parent:
            raise AttributeError("C4ApplicationProfile.parent is not set for application %s" % app_name)
        if '' not in self._description:
            raise AttributeError("C4ApplicationProfile has no default description for application %s" % app_name)
        if not self.category:
            raise AttributeError("C4ApplicationProfile.category is not set for application %s" % app_name)
        check_template = False
        for platform in self._medals:
            if self._medals[platform] not in C4ApplicationProfile.ALL_MEDALS:
                raise AttributeError("invalid C4ApplicationProfile.medal '%s' for platform '%s' for application %s" % (self._medals[platform], platform, app_name))
            if not self.parent.trusted and self._medals[platform] in C4ApplicationProfile.SUPPORTED_MEDALS:
                cxlog.log_("c4profiles","untrusted source cannot specify C4ApplicationProfile.medal '%s' for platform '%s' for application %s" % (cxlog.debug_str(self._medals[platform]), cxlog.debug_str(platform), cxlog.debug_str(app_name)))
                self._medals[platform] = C4ApplicationProfile.UNTRUSTED_MEDAL_MAP[self._medals[platform]]
            if self._medals[platform] in self.BRONZE_PLUS_MEDALS:
                check_template = True
        if check_template and ('bottle_types' not in self.__dict__ or \
                                   not self.bottle_types['use']):
            cxlog.warn('%s has a bronze or higher medal so it should specify the bottle template(s) to use' % cxlog.to_str(app_name))
        for registryglob in self.installed_registry_globs:
            registryglob.validate(app_name)
        _validate_file_globs(self.installed_file_globs, "the installed_file_globs field of %s" % cxlog.to_str(app_name))
        if 'component' in self.flags:
            if self.download_urls and '' not in self.download_urls:
                raise AttributeError("C4ApplicationProfile has no default download url for the %s component" % app_name)
            if self.download_page_urls and '' not in self.download_page_urls:
                raise AttributeError("C4ApplicationProfile has no default download page url for the %s component" % app_name)

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self,
                     ('_description', 'category', '_medals', 'flags',
                      'installer_id', 'installed_key_pattern',
                      'installed_display_pattern', 'installed_registry_globs',
                      'installed_file_globs', 'bottle_types',
                      'application_group', 'download_urls',
                      'download_page_urls', 'extra_fors'
                      ))


    ### Accessors for special properties

    def _getdescription(self):
        """Returns the application description in the current language.

        If no description is available for the current language, then the
        default description is returned.
        """
        _lang, description = cxutils.get_language_value(self._description)
        return description

    description = property(_getdescription)


    def _getmedal(self):
        """Returns the application medal for the current platform."""
        if distversion.PLATFORM in self._medals:
            return self._medals[distversion.PLATFORM]
        return 'untested'

    medal = property(_getmedal)

    def _getmedal_icon(self):
        """Returns the icon name for this application's medal."""
        return os.path.join(cxutils.CX_ROOT,
                            'share/images/medal_' + self._getmedal() + '.png')

    medal_icon = property(_getmedal_icon)


    def _getis_supported(self):
        """Returns true if the application is supported."""
        try:
            return self.parent.trusted and self._medals[distversion.PLATFORM] in C4ApplicationProfile.SUPPORTED_MEDALS
        except KeyError:
            return False

    is_supported = property(_getis_supported)

    def _getis_virtual(self):
        return "virtual" in self.flags

    is_virtual = property(_getis_virtual)


    def _getpreferred_bottle_type(self):
        """Returns the profile's preferred bottle type."""
        return self.bottle_types['use'][0]

    preferred_bottle_type = property(_getpreferred_bottle_type)

    def _getis_trusted_source(self):
        return self.parent.trusted

    is_trusted_source = property(_getis_trusted_source)

    def _get_origin(self):
        return self.parent.origin

    origin = property(_get_origin)

#####
#
# C4InstallerProfile class
#
#####

class C4EnvVar(object):
    """Describes an environment variable to set."""

    ### Property defaults

    # The name of the environment variable to set.
    # This field is mandatory and thus has no default.
    # name = ""

    # The value of the environment variable to set.
    # This field is mandatory and thus has no default.
    # value = ""


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'name' not in self.__dict__:
            raise AttributeError("C4Environment.name is not set for application %s" % app_name)
        if 'value' not in self.__dict__:
            raise AttributeError("C4Environment.value is not set for application %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('name', 'value'))


class C4FilesToCopy(object):
    """Specifies a set of files to copy."""

    ### Property defaults

    # A glob identifying the files to copy.
    # This field is mandatory and thus defaults to None.
    glob = None

    # A restricted path specifying where to copy the above files.
    # This field is mandatory and thus defaults to None.
    destination = None


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'glob' not in self.__dict__:
            raise AttributeError("C4FilesToCopy.glob is not set for application %s" % app_name)
        _validate_file_globs((self.glob,), "the C4FilesToCopy.glob field of %s" % app_name)
        if 'destination' not in self.__dict__:
            raise AttributeError("C4FilesToCopy.destination is not set for application %s" % app_name)
        _validate_file_paths((self.destination,), "the C4FilesToCopy.destination field of %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('glob', 'destination'))


class C4RegistryValue(object):
    """Specifies a registry value to set."""

    ### Property defaults

    # The name of the registry value to set.
    # This field is mandatory and thus has no default.
    # name = ""

    # The content of the registry value to set.
    # This field is mandatory and thus has no default.
    # data = ""


    ### Instance validation and dumping

    def validate(self, app_name, key):
        """Checks that all the mandatory fields have been set."""
        if 'name' not in self.__dict__:
            raise AttributeError("C4RegistryValue.name is not set for %s in application %s" % (key, app_name))
        if 'data' not in self.__dict__:
            raise AttributeError("C4RegistryValue.data is not set for %s in application %s" % (key, app_name))
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('name', 'data'))


class C4Registry(object):
    """Specifies a set of registry values to set."""

    ### Property defaults

    # The name of the registry key.
    # This field is mandatory and thus defaults to None.
    key = None

    # The name of the registry value to set if any.
    values = tuple()


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if not self.key:
            raise AttributeError("C4Registry.key is not set for application %s" % app_name)
        if self.values:
            for value in self.values:
                value.validate(app_name, self.key)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('key', 'values'))


class C4LnkFile(object):
    """Specifies a wnidows shortcut to create."""

    ### Property defaults

    # The path of the shortcut file to create.
    # This field is mandatory and thus defaults to None.
    shortcut = None

    # The path of the file the shortcut should point to.
    # This field is mandatory and thus defaults to None.
    target = None

    # The work directory to store in the shortcut if any.
    workdir = ""


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'shortcut' not in self.__dict__:
            raise AttributeError("C4LnkFile.shortcut is not set for application %s" % app_name)
        if 'target' not in self.__dict__:
            raise AttributeError("C4LnkFile.target is not set for %s in application %s" % (self.shortcut, app_name))
        _validate_file_paths((self.shortcut, self.target), "the C4LnkFile shortcut or target fields of %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('shortcut', 'target', 'workdir'))


class C4InstallerProfile(object):
    """Contains the C4 Installer Profile information."""

    ### 'Pre-Install' Property defaults

    # Points to the umbrella C4Profile object that ties a C4 entry's
    # specialized profiles together. This field is mandatory and thus defaults
    # to None.
    parent = None

    # A string storing the application id of the profile these settings apply
    # to. Note that it may not match the application id of the parent
    # C4Profile. It is optional and defaults to None.
    appid = None

    # Specifies when this installer profile should be used.
    use_if = None

    # Used to warn users about issues they may encounter during the
    # installation.
    _installation_notes = {}

    # A file that should be present on the DVD. If it is missing, then the UDF
    # side has been mounted and it needs remounting.
    udf_remount_check = ""

    # List of names under which the installer may be found on local media.
    local_installer_file_globs = tuple()

    # The list of applications that must be installed before we can install
    # this one.
    pre_dependencies = set()

    # The important cxdiag warnings, or those to ignore.
    cxdiag_checks = tuple()

    ### 'Install' Property defaults

    # The file globs to use to find the installer when faced with a directory.
    installer_file_globs = tuple()

    # A full basename to treat the installer as despite what its real name
    # may be
    installer_treatas = ""

    # A list of options to pass to the application installer.
    installer_options = tuple()

    # A list of options to pass to the application installer when doing a
    # silent installation.
    installer_silent_options = tuple()

    # The windows version to use to run the installer.
    installer_winver = ""

    # A Wine DllOverrides string to use to run the installer.
    installer_dlloverrides = ""

    # A list of C4EnvVar objects describing the environment variables to set
    # before running the installer.
    installer_environment = tuple()

    # A size in KB above which the installer is considered to be a
    # self-extracting archive.
    selfextract_threshold = None

    # A list of options to pass to the self-extracting archive to get it to
    # extract in a known location. These options can make use of the special
    # %ExtractDir% environment variable.
    selfextract_options = tuple()

    # A list of options to pass to the self-extracting archive when doing a
    # silent installation.
    selfextract_silent_options = tuple()

    # A list of C4FilesToCopy objects describing how to copy the files for
    # applications that don't have a proper installer.
    files_to_copy = tuple()

    # If True, force a reboot after the installer completes.
    post_install_reboot = False

    # A list of C4Registry objects describing the registry keys and values to
    # create before running the installer.
    pre_install_registry = tuple()

    # A list of fake dlls to remove before running the installer.
    pre_rm_fake_dlls = tuple()

    # A list of Netscape plugin dlls to export before running the installer.
    pre_install_nsplugin_dlls = tuple()

    # A list of Netscape plugins to ignore.
    ignore_nsplugin_dlls = tuple()

    # A list of C4LnkFile objects describing the Windows shortcuts to be
    # created.
    lnk_files = tuple()

    ### 'Post-Install' Property defaults

    # Set to True if the application is known to not create any menu.
    skip_menu_creation = False

    # Set to True if the application is known to not create any association.
    skip_assoc_creation = False

    # The list of EAssocs that should be made the default.
    default_eassocs = tuple()

    # The list of EAssocs that should be exported as alternatives.
    alt_eassocs = tuple()

    # A list of dll file paths to register after install
    post_registerdll = tuple()

    # A web page to open after the application has installed.
    post_install_url = ""


    def copy(self):
        result = C4InstallerProfile()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__, exclude=('parent',))


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.
        """
        if not self.parent:
            raise AttributeError("C4InstallerProfile.parent is not set for application %s" % app_name)
        if self._installation_notes and '' not in self._installation_notes:
            raise AttributeError("C4InstallerProfile has no default installation notes for application %s" % app_name)
        _validate_file_globs(self.local_installer_file_globs, "the local_installer_file_globs field of %s" % app_name)
        for dependency in self.pre_dependencies:
            if dependency.startswith('.'):
                raise AttributeError("C4InstallerProfile.pre_dependencies '%s' must not start with a dot for application %s" % (dependency, app_name))
        _validate_file_globs(self.installer_file_globs, "the installer_file_globs field of %s" % app_name)
        for envvar in self.installer_environment:
            envvar.validate(app_name)
        if self.selfextract_threshold is None:
            if self.selfextract_options:
                raise AttributeError("C4InstallerProfile.selfextract_options is set but no threshold has been specified for application %s" % app_name)
            if self.selfextract_silent_options:
                raise AttributeError("C4InstallerProfile.selfextract_silent_options is set but no threshold has been specified for application %s" % app_name)
        for file_copy in self.files_to_copy:
            file_copy.validate(app_name)
        for regentry in self.pre_install_registry:
            regentry.validate(app_name)
        if self.lnk_files:
            for lnk_file in self.lnk_files:
                lnk_file.validate(app_name)
            if self.skip_menu_creation:
                raise AttributeError("C4InstallerProfile.lnk_files must be empty when skip_menu_creation is True for application %s" % app_name)
        if self.skip_assoc_creation and (self.default_eassocs or self.alt_eassocs):
            raise AttributeError("C4InstallerProfile.default_eassocs and alt_eassocs must be empty when skip_assoc_creation is True for application %s" % app_name)
        return True


    def dump(self, out, indent=""):
        _dump_fields(out, indent, self,
                     ('use_if', 'installation_notes', 'udf_remount_check',
                      'local_installer_file_globs', 'pre_dependencies',
                      'cxdiag_checks', 'installer_file_globs',
                      'installer_treatas', 'installer_options',
                      'installer_silent_options', 'installer_winver',
                      'installer_dlloverrides', 'installer_environment',
                      'selfextract_threshold', 'selfextract_options',
                      'selfextract_silent_options', 'files_to_copy',
                      'post_install_reboot', 'pre_install_registry',
                      'pre_rm_fake_dlls', 'pre_install_nsplugin_dlls',
                      'ignore_nsplugin_dlls', 'lnk_files',
                      'skip_menu_creation', 'skip_assoc_creation',
                      'default_eassocs', 'alt_eassocs',
                      'post_registerdll', 'post_install_url'))


    ### Accessors for special properties

    def use(self, use_if_props):
        if self.appid is not None and self.appid != use_if_props['appid']:
            return False
        if self.use_if is None:
            return True
        return self.use_if.is_true(use_if_props)


    def _getinstallation_notes(self):
        """Returns the application installation notes in the current language.

        If no installation notes are available for the current language, then
        the default installation notes are returned.
        """
        _lang, installation_notes = cxutils.get_language_value(self._installation_notes)
        return installation_notes

    installation_notes = property(_getinstallation_notes)


    def get_default_eassocs_str(self):
        """Returns a string concatenating all the EAssocs that are to be made
        defaults.

        The string is suitable for cxassoc.
        """
        return ":".join(self.default_eassocs)


    def get_alt_eassocs_str(self):
        """Returns a string concatenating all the EAssocs that are to be made
        alternatives.

        The string is suitable for cxassoc.
        """
        return ":".join(self.alt_eassocs)


#####
#
# Umbrella C4Profile class
#
#####

class C4CXVersion(object):
    """Identifies a product plus version combination which is compatible with
    the corresponding profile.
    """

    ### Property defaults

    # These class variables provide defaults in case the corresponding
    # instance variable is not set. They should all be imutable objects,

    # A string storing the product name.
    # This field is mandatory and thus defaults to None.
    product = None

    # A string storing the version prefix.
    # This field is mandatory and thus defaults to None.
    cxversion = None


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.

        Note that this only checks internal consistency, not consistency with
        other profiles.
        """
        if not self.product:
            raise AttributeError("C4Profile.product is not set for application %s" % app_name)
        if not self.cxversion:
            raise AttributeError("C4Profile.cxversion is not set for application " % app_name)
        # FIXME: We should check that the cxversion field is valid
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('product', 'cxversion'))

    def copy(self):
        result = C4CXVersion()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__)


### Helper for computing the profile scores

def _version2int(version):
    """Converts a version for the form xx.yy.zz into a tuple of the form
    (xxyyzz, m), xxyyzz is an integer representing the version, and m is
    the number of missing version components.
    """
    components = version.split('.', 3)[0:3]
    value = 0
    missing = 0
    pad = False
    for i in xrange(3):
        value *= 100
        if pad:
            continue
        try:
            component = int(components[i])
            if component < 0 or component > 99:
                cxlog.log_("c4profiles", "the %s version is invalid" % cxlog.debug_str(version))
                missing = 3 - i
                pad = True
            else:
                value += component
        except IndexError:
            missing = 3 - i
            pad = True
        except ValueError:
            cxlog.log_("c4profiles", "the %s version is invalid" % cxlog.debug_str(version))
            missing = 3 - i
            pad = True
    return (value, missing)


class C4Profile(object):
    """This is an umbrella class which ties a C4 entry's CD, Application and
    Installer profiles together.

    It also stores the common parts between these, that is the application id
    and its name.
    """

    ### Property defaults

    # These class variables provide defaults in case the corresponding
    # instance variable is not set. They should all be imutable objects,

    # this field does not appear in the C4P file. It reflects if this profile
    # has been parsed from a trusted source or not. It is set by the
    # c4parser
    trusted = False

    # at present, this field does not appear in the C4P file.
    # A string representing the origin of the c4profile.
    # It is set by the c4parser
    origin = None

    # A string storing the application id, usually a C4 identifier.
    # This field is mandatory and thus defaults to None.
    appid = None

    # The CrossOver products and versions this profile was tested with.
    cxversions = tuple()

    # A dictionary mapping the language to the application name in that
    # language. The special '' mapping corresponds to the default application
    # name. This field is mandatory and thus defaults to None.
    _name = None

    # Points to the CD profile information if any.
    # In pysetup we may never use this field but it would be used by the CD
    # detection tool.
    cd_profile = None

    # Points to the application profile information if any.
    app_profile = None

    # The list of installer profiles. There may be more than one, because some
    # may be modifying properties of other profiles, and some may only be used
    # in specific conditions.
    installer_profiles = tuple()


    ### Instance validation and dumping

    def validate(self):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.

        Note that this only checks internal consistency, not consistency with
        other profiles.
        """
        if not self.appid:
            raise AttributeError("C4Profile.appid is not set")
        # pylint: disable-msg=E1101
        elif self.appid.startswith('.'):
            raise AttributeError("C4Profile.appid '%s' must not start with a dot" % self.appid)
        if not self._name:
            raise AttributeError("C4Profile._name is not set for %s" % self.appid)
        if '' not in self._name or not self._name['']:
            raise AttributeError("C4Profile has no default name for %s" % self.appid)
        for cxversion in self.cxversions:
            cxversion.validate(self._name[''])

        # Pylint does not see that if these are not None, then then will point
        # to instances that do have the validate function()
        # pylint: disable-msg=E1101
        if self.cd_profile:
            self.cd_profile.validate(self._name[''])
        if self.app_profile:
            self.app_profile.validate(self._name[''])
        for inst_profile in self.installer_profiles:
            inst_profile.validate(self._name[''])
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self,
                     ('appid', 'cxversions', '_name', 'app_profile',
                      'installer_profiles', 'cd_profile'))


    def copy(self):
        result = C4Profile()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        if result.cd_profile:
            result.cd_profile.parent = result
        if result.app_profile:
            result.app_profile.parent = result
        for inst_profile in result.installer_profiles:
            inst_profile.parent = result

        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__)
        # The profiles may have been copied, in which case the parent needs to
        # be updated.
        if self.cd_profile:
            self.cd_profile.parent = self
        if self.app_profile:
            self.app_profile.parent = self
        for inst_profile in self.installer_profiles:
            inst_profile.parent = self


    ### Helpers to compute the profile score

    # The product's code and numerical version data
    _product = None
    _product_version = None
    _product_missing = None

    # A cache mapping the (product, cxversion) tuples to the corresponding
    # score
    _scores = {}

    @staticmethod
    def _cxversion_score(cxversion):
        """Computes how close the specified cxversion is to match the current
        product. The lower the score, the better the match.
        """
        cxversion_id = (cxversion.product, cxversion.cxversion)
        if cxversion_id in C4Profile._scores:
            return C4Profile._scores[cxversion_id]

        if C4Profile._product is None:
            C4Profile._product = distversion.BUILTIN_PRODUCT_ID
            # Some hacks for the Mac
            if C4Profile._product == "CrossOver":
                C4Profile._product = "cxoffice"
            elif C4Profile._product == "CrossOverGames":
                C4Profile._product = "cxgames"
            C4Profile._product_version, C4Profile._product_missing = \
                                        _version2int(distversion.CX_VERSION)

        version, missing = _version2int(cxversion.cxversion)
        score = C4Profile._product_version - version
        if score < 0:
            score = 1000000 - score
        if cxversion.product != C4Profile._product:
            score += 2000000
        score = 10 * score + abs(C4Profile._product_missing - missing)

        C4Profile._scores[cxversion_id] = score
        return score


    ### Accessors for special properties

    def _getname(self):
        """Returns the application name in the current language.

        If no name is available for the current language, then the
        default name is returned.
        """
        _lang, name = cxutils.get_language_value(self._name)
        return name

    name = property(_getname)


    _score = None

    def _get_score(self):
        """A score that defines the profile's adequation to the current
        product. The lower the score the better.
        """
        if self._score is None:
            self._score = 40000000
            for cxversion in self.cxversions:
                score = C4Profile._cxversion_score(cxversion)
                if score < self._score:
                    self._score = score
        return self._score

    score = property(_get_score)

    def _is_for_current_product(self):
        """Returns True if one of the cxversion properties match the current
        product (though not necessarily this version), and False otherwise.
        """
        if self.score >= 20000000:
            # A score that big means a product mismatch. We don't allow
            # using such profiles to install applications.
            return False
        return True

    is_for_current_product = property(_is_for_current_product)


    def _getdetails_url(self):
        """Returns the url to a page with more information about this profile,
        or '' if there is no such page."""
        # For now, just treat com.codeweavers.c4.NNN as special.
        # In the future, we may want to add a field in c4p files for this.
        # pylint: disable-msg=E1101
        if self.appid.startswith('com.codeweavers.c4.'):
            c4_id = self.appid[19:]
            if c4_id.isdigit():
                return 'http://www.codeweavers.com/compatibility/browse/name?app_id=%s' % c4_id
        return ''

    details_url = property(_getdetails_url)


    def _is_application(self):
        """Returns True if the application flag is set."""
        # pylint: disable-msg=E1101
        return 'application' in self.app_profile.flags

    is_application = property(_is_application)


    def _is_component(self):
        """Returns True if the component flag is set."""
        # pylint: disable-msg=E1101
        return 'component' in self.app_profile.flags

    is_component = property(_is_component)


    def _is_nsplugin(self):
        """Returns True if the nsplugin flag is set."""
        # pylint: disable-msg=E1101
        return 'nsplugin' in self.app_profile.flags

    is_nsplugin = property(_is_nsplugin)

