# (c) Copyright 2009-2010. CodeWeavers, Inc.

import os.path
import re

import bottlemanagement
import cxlog
import cxobjc
import cxutils
import cxconfig

# for localization
from cxutils import cxgettext as _

import c4profiles
import c4profilesmanager
import bottlequery
import cxaiemedia
import cddetector


#####
#
# Keep track of the installation SourceMedia and of their properties
#
#####

class SourceMedia(object):
    """Keeps track of installation wizard data associated to a potential
    installation media.
    """

    def __init__(self, installtask, path, label):
        """Source is the absolute path of the installation media. This normally
        corresponds to the mount point of a CD or a directory, but it may also
        point to a user-specified file.

        The label is the string that should be used to describe that media.
        """
        self.installtask = installtask
        self.path = path
        self.label = label
        self._profile_ids = None

    def _get_profile_ids(self):
        """Returns the list of profile ids that this media could correspond to.
        """
        if self._profile_ids is None:
            if os.path.isdir(self.path):
                self._profile_ids = cddetector.get_cd_profiles(self.path, self.installtask.profiles)
            else:
                self._profile_ids = []
        return self._profile_ids

    profile_ids = property(_get_profile_ids)



#####
#
# Keep track of the target bottles and of their properties
#
#####

# Bottle categories
CAT_NONE = 0
CAT_RECOMMENDED = 1
CAT_COMPATIBLE = 2
CAT_INCOMPATIBLE = 3

class TargetBottle(object):

    #####
    #
    # Initialization
    #
    #####

    def __init__(self, installtask, bottle):
        self.installtask = installtask
        self._bottle = bottle

        if bottle:
            self.bottlename = bottle.name
        else:
            self.bottlename = None
        self._newtemplate = None

        # The profile-related properties
        # This should match check_cache()
        self._profile = None
        self._locale = None
        self._category = CAT_NONE
        self._installers = None
        self._missing_media = None
        self._missing_profiles = None
        self._warnings = None

    def _gettemplate(self):
        if self._newtemplate is None:
            return self._bottle.bottletemplate
        return self._newtemplate

    template = property(_gettemplate)


    #####
    #
    # Profile dependency analysis
    #
    #####

    def check_cache(self):
        """Checks whether the cached data is still relevant and throws it away
        if not (typically because either the profile or the locale changed).
        """
        # If further optimisation is needed we can check for the presence of
        # useif or even scan its content before reseting the cache for a mere
        # locale change.
        if self._profile != self.installtask.profile or \
                self._locale != self.installtask.installerDownloadLocale:
            self._profile = self.installtask.profile
            self._locale = self.installtask.installerDownloadLocale
            # Wipe everything out
            self._category = CAT_NONE
            self._installers = None
            self._missing_media = None
            self._missing_profiles = None
            self._warnings = None

    def analyzed(self):
        """Returns True if the bottle has been analyzed already and False
        otherwise.
        """
        self.check_cache()
        return self._installers is not None

    def _break_dependency_loops(self, appid, parents):
        """Detects, breaks and reports dependency loops."""
        parents.add(appid)
        to_remove = set()
        for depid in self._installers[appid].pre_dependencies:
            if depid in parents:
                if not self.bottlename:
                    cxlog.warn("The dependency of %s on %s creates a loop in bottle template %s." % (cxlog.to_str(self.installtask.profiles[appid].name), cxlog.to_str(self.installtask.profiles[depid].name), self.template))
                to_remove.add(depid)
            elif depid in self._installers:
                self._break_dependency_loops(depid, parents)
            else:
                # This dependency is either already installed, or it has no
                # profile at all. Either way we should ignore it.
                to_remove.add(depid)
        if to_remove:
            self._installers[appid].pre_dependencies -= to_remove
        parents.remove(appid)

    def analyze(self):
        """Analyzes the bottle so that we know all there is to know for
        installation and about its suitability for the selected profile and
        locale. This means:
        - creating aggregated installer profiles for the selected profile and
          all its dependencies,
        - detecting and reporting any potential issue like dependency loops,
          template incompatibilities, etc.
        """
        if self.analyzed():
            return True

        if self._bottle:
            if not self._bottle.get_installed_packages_ready():
                # This bottle's installed applications list is not ready yet
                # so we cannot perform the analysis. So return False so the
                # caller knows he should try again later.
                cxlog.log("analyze: %s is not ready" % cxlog.to_str(self.bottlename))
                return False
            installed_apps = self._bottle.installed_packages
        else:
            installed_apps = {}

        use_if_props = self.installtask.get_use_if_properties(bottle_type=self.template)
        profiles = self.installtask.profiles

        self._installers = {}
        self._missing_media = set()
        self._missing_profiles = set()
        self._warnings = []

        incompatible_profiles = []
        overrides = {}
        dependencies = {}
        todo = [self._profile.appid]
        while todo:
            appid = todo.pop()
            if appid in self._installers or \
                   (appid != self._profile.appid and appid in installed_apps):
                # We've done that one already. This check also lets us avoid
                # infinite loops.
                continue

            if appid not in profiles or \
                   not profiles[appid].is_for_current_product:
                # This dependency has no profile *at all* (not even a name),
                # or it is for the wrong CrossOver product.
                self._missing_profiles.add(appid)
                continue

            # Aggregate all the installer profile chunks into one installer
            # profile.
            installer = c4profiles.C4InstallerProfile()
            nomatch = True
            for inst_profile in profiles.installer_profiles(appid):
                if inst_profile.appid is None:
                    use_if_props['appid'] = appid
                else:
                    use_if_props['appid'] = inst_profile.appid
                if inst_profile.use(use_if_props):
                    if inst_profile.appid is None or \
                           inst_profile.appid == appid:
                        installer.update(inst_profile)
                        nomatch = False
                    else:
                        # Remember the override information so we can apply it
                        # after we have built the main installer profile.
                        if inst_profile.appid not in overrides:
                            overrides[inst_profile.appid] = []
                        overrides[inst_profile.appid].append(inst_profile)

                        if inst_profile.appid in self._installers:
                            # We may have just discovered new dependencies for
                            # a profile we already handled.
                            todo.extend(inst_profile.pre_dependencies)
                        else:
                            # Remember the extra dependencies for that profile
                            # so we take them into account if we run into it
                            # later on.
                            if inst_profile.appid not in dependencies:
                                dependencies[inst_profile.appid] = set()
                            dependencies[inst_profile.appid].update(inst_profile.pre_dependencies)

            if nomatch:
                # There is no installer profile so use the one for the unknown
                # applications.
                installer = profiles.unknown_installer().copy()
            installer.parent = profiles[appid]
            self._installers[appid] = installer

            # Check that bottle's template is compatible with this installer
            if appid == self._profile.appid:
                purpose = 'use'
            else:
                purpose = 'install'
            if self.template not in profiles[appid].app_profile.bottle_types[purpose]:
                incompatible_profiles.append(profiles[appid].name)

            if appid != self._profile.appid and \
                   'virtual' not in profiles[appid].app_profile.flags and \
                   not cxaiemedia.get_builtin_installer(appid) and \
                   not profiles[appid].app_profile.download_urls:
                # We cannot automatically download this dependency.
                # The installation engine does not care, but still make a note
                # of it for the GUI.
                self._missing_media.add(appid)

            # Schedule the dependencies for analysis
            if appid not in dependencies:
                dependencies[appid] = set()
            dependencies[appid].update(installer.pre_dependencies)
            todo.extend(dependencies[appid])

        if incompatible_profiles:
            apps = ', '.join(sorted(incompatible_profiles))
            if self.template in self._profile.app_profile.bottle_types['use']:
                if not self.bottlename:
                    cxlog.warn("The %s template is compatible with %s but is incompatible with dependencies: %s" % (self.template, cxlog.to_str(self._profile.appid), cxlog.to_str(apps)))
            else:
                if self.bottlename:
                    self._warnings.append(_("You have selected the '%(bottlename)s' bottle, which is incompatible with %(applications)s.") % {
                            'bottlename': self.bottlename,
                            'applications': apps})
                else:
                    self._warnings.append(_("You have chosen to install in a new %(template)s bottle but it is incompatible with %(applications)s.") % {
                            'template': self.template,
                            'applications': apps})

        # Apply all the installer overrides so we don't have to worry about
        # them later.
        for appid, installer_profiles in overrides.iteritems():
            if appid in self._installers:
                for inst_profile in installer_profiles:
                    # Note that this may impact the pre_dependencies list, but
                    # we took precautions in the above loop so it won't
                    # reference new profiles.
                    installer = self._installers[appid]
                    installer.update(inst_profile)

        # Break dependency loops
        self._break_dependency_loops(self._profile.appid, set())
        return True

    def _getinstallers(self):
        """Returns the aggregated installer profiles for the application and
        all its dependencies.

        Note: Installation information can be split up across multiple
        C4InstallerProfile objects which makes it unusable as is. This is
        unlike the installer profiles in this map where all the pieces have
        been 'aggregated' together into a single installer profile object.
        """
        self.analyze()
        return self._installers

    installers = property(_getinstallers)

    def _getmissing_media(self):
        """Returns the dependency application ids for which we cannot
        automatically get an installer. The GUI will have to ask the user for
        an installation source for these.
        """
        self.analyze()
        return self._missing_media

    missing_media = property(_getmissing_media)

    def _getmissing_profiles(self):
        """Returns the dependencies for which there is no C4 profile
        information at all for the current product.
        """
        self.analyze()
        return self._missing_profiles

    missing_profiles = property(_getmissing_profiles)

    def _getwarnings(self):
        """Returns a list of localized strings describing the issues found
        while analyzing this bottle.

        This may be the presence of dependency loops for instance, missing
        profiles, etc.
        """
        self.analyze()
        return self._warnings

    warnings = property(_getwarnings)


    #####
    #
    # Assign a category to the bottle
    #
    #####

    def has_category(self):
        """Returns True if we have determined this bottle's category."""
        self.check_cache()
        return self._category != CAT_NONE

    def _cache_category(self, category):
        """A convenience method that caches the category and returns it."""
        self._category = category
        return self._category

    def get_category(self):
        """Returns the bottle's category which is a mesure of its suitability
        for the profile selected for installation.

        CAT_INCOMPATIBLE means the bottle is incompatible with it.
        CAT_RECOMMENDED means we recommend installing in this bottle, which
          happens if it contains an application the profile is an extra for,
          or that belongs in the same application bottle group as the profile.
        CAT_COMPATIBLE means the bottle is compatible.
        """
        self.check_cache()
        if self._category != CAT_NONE or not self._profile:
            return self._category

        if self.template not in self._profile.app_profile.bottle_types['use']:
            # The bottle's template is not in the profile's set of compatible
            # templates. So we need look no further.
            return self._cache_category(CAT_INCOMPATIBLE)

        app_profile = self._profile.app_profile
        if self._bottle and not app_profile.extra_fors and \
                not app_profile.application_group:
            # This is a real bottle, and the profile has no property that would
            # let us recommend a specific bottle. So we may be able to base our
            # analysis on the template alone.
            target_template = self.installtask.templates[self.template]
            if target_template.get_category() == CAT_COMPATIBLE:
                # If the profile is compatible with the template, then it
                # should be compatible with any bottle based on it (because
                # the only source of incompatibilities are the dependencies and
                # new bottles are where we have most of them).
                return self._cache_category(CAT_COMPATIBLE)

        # We need to really analyze the dependencies now.
        if not self.analyze():
            # Return CAT_NONE so the caller knows he should try again later.
            return CAT_NONE

        # If the bottle analysis found issues, then the bottle is incompatible.
        if self.warnings:
            return self._cache_category(CAT_INCOMPATIBLE)

        # Prefer bottles containing one of the applications we are an extra
        # for, and those containing applications that belong to our
        # application bottle group.
        if app_profile and self._bottle and \
                (app_profile.extra_fors or app_profile.application_group):
            for (appid, installed) in self._bottle.installed_packages.iteritems():
                if appid in app_profile.extra_fors:
                    return self._cache_category(CAT_RECOMMENDED)
                if app_profile.application_group and \
                        installed.profile and installed.profile.app_profile:
                    app_group = installed.profile.app_profile.application_group
                    if app_profile.application_group == app_group:
                        return self._cache_category(CAT_RECOMMENDED)

        return self._cache_category(CAT_COMPATIBLE)

    # Alias for Objective-C
    getCategory = get_category


class TargetTemplate(TargetBottle):
    def __init__(self, installtask, template):
        TargetBottle.__init__(self, installtask, None)
        self._newtemplate = template


#####
#
# InstallTask delegate
#
#####

class InstallTaskDelegate(object):
    """This class defines the delegate interface that InstallTask uses to
    notify the GUI of changes in the installtask state.

    The GUI (especially the Mac one) does not have to provide an object that
    derives from this class, but it must implement all of the methods defined
    here.
    """

    @cxobjc.delegate
    def profileChanged(self):
        """Notifies the GUI that the profile selected for installation has
        changed.
        """
        pass

    @cxobjc.delegate
    def sourceChanged(self):
        """Notifies the GUI that the media selected for the installation has
        changed.
        """
        pass

    @cxobjc.delegate
    def categorizedBottle_(self, target_bottle):
        """Notifies the GUI that the specified bottle's category is now
        available, where target_bottle is either a TargetBottle or a
        TargetTemplate object.

        This may be used to implement a GUI that shows the bottles right away
        and then updates a field as their categories become known.
        """
        pass

    @cxobjc.delegate
    def categorizedAllBottles(self):
        """Notifies the GUI that all the bottles have been assigned a
        category.

        It also means that the bottle picking, if any, is done.
        This can be used to implement a GUI which only shows list of bottles
        one can install into once all their categories are known.
        """
        pass

    @cxobjc.delegate
    def analyzedBottle_(self, target_bottle):
        """Notifies the GUI that the specified bottle has been fully analyzed.

        This means we have determined everything there is to know in order to
        perform the installation into this bottle. So once one has received
        this notification one can query the list of dependencies that will
        need to be installed, and the list issues (such as dependency loops,
        etc).
        Note that the order of the analyzedBottle_() and categorizedBottle_()
        notifications is undefined. However, once a bottle has been analyzed
        its category can be computed and it is ready for installation (if
        compatible).
        """
        pass

    @cxobjc.delegate
    def invalidNewBottleName(self):
        """Notifies the GUI that the newBottleName field has been set to an
        invalid value.

        This may either be because it contains bad characters or because a
        bottle, file or folder by that name already exists.
        However such notifications are only sent if InstallTask is instructed
        to install in a new bottle.
        """
        pass

    @cxobjc.delegate
    def bottleCreateChanged(self):
        """Notifies the GUI that the flag specifying whether to install in a
        new bottle or in an existing one has changed.
        """
        pass

    @cxobjc.delegate
    def bottleNewnameChanged(self):
        """Notifies the GUI that the new bottle name has been modified.
        FIXME: Specify if one can receive such notifications when not
        installing in a new bottle.
        """
        pass

    @cxobjc.delegate
    def bottleTemplateChanged(self):
        """Notifies the GUI that the new bottle template has been modified.
        FIXME: Specify if one can receive such notifications when not
        installing in a new bottle.
        """
        pass

    @cxobjc.delegate
    def bottleNameChanged(self):
        """Notifies the GUI that the name of the bottle to install into has
        been modified.
        FIXME: Specify if one can receive such notifications when
        installing in a new bottle.
        """
        pass

_NULL_DELEGATE = InstallTaskDelegate()


#####
#
# InstallTask
#
#####

class InstallTask(cxobjc.Proxy):
    """The InstallTask class will keep track of the state of a user-selected
    c4 profile as we go through the process of installing it. It will also
    provide an interface to installation information that must be returned at
    run-time, e.g. bottle compatibility.
    """

    def __init__(self, delegate=None):
        cxobjc.Proxy.__init__(self)
        self.profiles = c4profilesmanager.C4ProfilesSet.all_profiles()

        # An object with methods that will be called when attributes change.
        self._delegate = _NULL_DELEGATE
        self.set_delegate(delegate)

        # The C4 profile of the application to install
        self._profile = None
        self._profileRequested = False

        # The corresponding installer profile. Note that it may belong to
        # another C4 profile, such as the profile for unknown applications.
        self.installer = None

        # A mapping of absolute paths to the corresponding SourceMedia
        # objects. Only the GUI knows about and keeps track of mount points,
        # etc. So it is responsible for filling in this map and keeping it up
        # to date.
        self.source_medias = {}

        # The source media.
        self.installerSourceRequested = False # True if the user requested this source
        self.installerDownloadSource = None
        self.installerDownloadLocale = None
        self.installerSource = None

        # The target bottle
        self.bottles = {}
        self.templates = {}
        for template in bottlemanagement.template_list():
            self.templates[template] = TargetTemplate(self, template)
        self.createNewBottle = True
        self.createNewBottleRequested = False
        self.newBottleName = ""
        self.newBottleNameRequested = False
        self.newBottleTemplate = "winxp"
        self.newBottleTemplateRequested = False
        self.existingBottleName = ""
        self.existingBottleNameRequested = False

        self.cxdiag_messages = None

    # Special initializer for objc. This must /always/ be called explicitly
    # on the Mac.
    def initWithDelegate_(self, delegate):
        self = cxobjc.Proxy.nsobject_init(self)
        if self is not None:
            self.__init__(delegate)
        return self

    def init(self):
        return self.initWithDelegate_(None)

    def set_delegate(self, delegate):
        if delegate:
            self._delegate = delegate
        else:
            self._delegate = _NULL_DELEGATE

    setDelegate_ = set_delegate

    def AutoFillSettings(self):
        # After all intitial settings have been set, and this object and the
        # change delegate are initialized, try to guess at anything remaining.

        if not self._profileRequested and not self.installerSourceRequested:
            # If we can find exactly one profile to select based on media, select it.
            package_to_select = None
            for source_path in self.source_medias:
                profiles = cddetector.get_cd_profiles(source_path, self.profiles)
                if len(profiles) == 1:
                    if package_to_select:
                        # too many profiles
                        break
                    else:
                        package_to_select = profiles.values()[0]
                elif profiles:
                    # too many profiles
                    break
            else:
                # 0 or 1 profiles found
                if package_to_select:
                    self._setprofile(package_to_select, requested=False)


    #####
    #
    # Profile selection
    #
    #####

    def _getprofile(self):
        return self._profile

    def _setprofile(self, profile, requested=True):
        if self._profile == profile:
            return

        if profile is None:
            requested = False

        # Profile
        self._profile = profile
        self._profileRequested = requested

        # Source Media

        # If the user had requested a download that's not available from the new
        # profile, we must discard the user's request.
        if profile and self.installerSourceRequested and self.installerDownloadSource is not None and \
                (not profile.app_profile.download_urls or \
                self.installerDownloadLocale not in profile.app_profile.download_urls or \
                profile.app_profile.download_urls[self.installerDownloadLocale] != self.installerDownloadSource):
            self.installerSourceRequested = False

        if profile and not self.installerSourceRequested:
            installerSource = None
            installerDownloadSource = None

            builtin = cxaiemedia.get_builtin_installer(self._profile.appid)
            if builtin is not None:
                installerSource = builtin
            elif self._profile.app_profile.download_urls:
                preferred_lang, url = cxutils.get_language_value(self._profile.app_profile.download_urls)
                installerDownloadLocale = preferred_lang
                installerDownloadSource = url
            else:
                for source in self.source_medias.itervalues():
                    if self._profile.appid in source.profile_ids:
                        if installerSource is None:
                            installerSource = source.path
                        else:
                            # If there are multiple matches, then let the user
                            # pick one
                            installerSource = None
                            break

            if installerSource:
                self.SetInstallerSource(installerSource, requested=False)
            elif installerDownloadSource:
                self.SetInstallerDownload(installerDownloadSource, installerDownloadLocale, requested=False)
            elif not self.installerSourceRequested:
                self.ClearInstallerSource()
        self._delegate.profileChanged()

        self._trigger_bottle_analysis()


    profile = property(_getprofile, _setprofile)

    # Wrappers for Objective-C
    def setProfile_(self, profile):
        self.profile = profile
    def setProfile_forUser_(self, profile, requested):
        self._setprofile(profile, requested)


    def update_profiles(self):
        self.profiles = c4profilesmanager.C4ProfilesSet.all_profiles()
        if self._profile:
            if self._profile.appid in self.profiles:
                self._setprofile(self.profiles[self._profile.appid], requested=self._profileRequested)
            else:
                self.profile = None

    updateProfiles = update_profiles


    def use_autorun_file(self, c4pfile):
        if self._profileRequested:
            return
        autorun = c4pfile.autorun
        if not autorun:
            profiles = c4pfile.profiles
            # If all profiles have the same id, use that id
            if profiles:
                profile_id = profiles[0].appid
                for profile in profiles[2:]:
                    if profile.appid != profile_id:
                        break
                else:
                    autorun = profile_id
        if autorun and autorun in self.profiles:
            self._setprofile(self.profiles[autorun], False)

    useAutorunFile_ = use_autorun_file


    def _is_virtual(self):
        return self._profile and 'virtual' in self._profile.app_profile.flags

    virtual = property(_is_virtual)


    def _get_display_name(self):
        if self._profile:
            return self._profile.name
        else:
            return None

    displayName = property(_get_display_name)


    #####
    #
    # Source media selection
    #
    #####

    def SetInstallerSource(self, inSource, requested=True):
        if self.installerSource == inSource:
            return

        self.installerSource = inSource

        if inSource:
            self.installerDownloadSource = None
            self.installerDownloadLocale = None
            self.installerSourceRequested = requested
            if not self._profileRequested:
                # Try to select a profile based on the source media.
                profiles = cddetector.get_cd_profiles(inSource, self.profiles)
                if len(profiles) == 1:
                    self._setprofile(profiles.values()[0], requested=False)

        self._delegate.sourceChanged()

    # Wrapper for Objective-C
    def setInstallerSource_(self, inSource):
        self.SetInstallerSource(inSource)

    def GetInstallerSource(self):
        return self.installerSource

    def SetInstallerDownload(self, url, locale, requested=True):
        locale_changed = (self.installerDownloadLocale != locale)
        if self.installerDownloadSource == url and not locale_changed:
            return

        self.installerSource = None
        self.installerDownloadSource = url
        self.installerDownloadLocale = locale
        self.installerSourceRequested = requested

        self._delegate.sourceChanged()
        if locale_changed and self._profile:
            # A change of locale may impact the bottle selection
            self._trigger_bottle_analysis()

    # Wrapper for Objective-C
    def setInstallerDownloadURL_locale_(self, url, locale):
        self.SetInstallerDownload(url, locale)

    def GetInstallerDownload(self):
        return self.installerDownloadSource, self.installerDownloadLocale

    def ClearInstallerSource(self):
        if self.installerSource is None and self.installerDownloadSource is None:
            return

        self.installerSource = None
        self.installerDownloadSource = None
        old_locale = self.installerDownloadLocale
        self.installerDownloadLocale = None
        self.installerSourceRequested = False

        self._delegate.sourceChanged()

        if old_locale and self._profile:
            # A change of locale may impact the bottle selection
            self._trigger_bottle_analysis()


    #####
    #
    # Automatic bottle selection
    #
    #####

    def _take_bottle_snapshot(self):
        self._bottle_snapshot = {
            'createNewBottle': self.createNewBottle,
            'newBottleName': self.newBottleName,
            'newBottleTemplate': self.newBottleTemplate,
            'existingBottleName': self.existingBottleName
            }

    def _send_bottle_change_notifications(self):
        """An internal helper for sending the relevant notification(s) when
        the selected bottle gets changed.
        """
        # FIXME: Do we really need such a complex change notification mechanism?
        # Switch to a single bottleChanged() function?
        if not self._bottle_snapshot:
            cxlog.err('No bottle snapshot was taken!')
            return
        if self.createNewBottle != self._bottle_snapshot['createNewBottle']:
            self._delegate.bottleCreateChanged()
        if self.newBottleName != self._bottle_snapshot['newBottleName']:
            self._delegate.bottleNewnameChanged()
        if self.newBottleTemplate != self._bottle_snapshot['newBottleTemplate']:
            self._delegate.bottleTemplateChanged()
        if self.existingBottleName != self._bottle_snapshot['existingBottleName']:
            self._delegate.bottleNameChanged()

    def _unpick_bottle(self):
        """This gets called when the target profile is unset and resets the
        relevant fields.
        """
        self._take_bottle_snapshot()
        if not self.existingBottleNameRequested: # FIXME: May be wrong
            self.createNewBottle = False
            self.existingBottleName = None
        if not self.newBottleNameRequested:
            self.newBottleName = None
        if not self.newBottleTemplateRequested:
            self.newBottleTemplate = None
        self._send_bottle_change_notifications()

    def _pick_new_bottle_name(self):
        """Picks a suitable name for the new bottle if we are free to do so.

        The caller is responsible for sending the bottle change notification(s).
        """
        if not self.newBottleNameRequested:
            # Note that here we know that the new bottle name is necessarily
            # valid.
            self.newBottleName = bottlequery.unique_bottle_name(cxutils.bottlename_sanitize(self._profile.name))

    def _pick_new_bottle_template(self):
        """Picks a suitable template for the new bottle if we are free to do so.

        The caller is responsible for sending the bottle change notification(s).
        """
        if self.newBottleTemplateRequested:
            return
        # Scan the profile's preferred template list and pick the first one
        # which is actually compatible
        fallback = None
        for template in self._profile.app_profile.bottle_types['use']:
            if template not in self.templates:
                # Not supported in the CrossOver version!
                pass
            elif self.templates[template].get_category() == CAT_COMPATIBLE:
                self.newBottleTemplate = template
                return
            elif not fallback:
                fallback = template

        if fallback:
            # No template is compatible with both the application and all its
            # dependencies. So ignore the dependencies.
            self.newBottleTemplate = fallback
            return

        # None of this application's templates is supported by this CrossOver
        # version! So pick the same one as for the unknown profile.
        self.newBottleTemplate = self.profiles.unknown_installer().app_profile.preferred_bottle_type

    def _pick_bottle(self):
        """This method performs the automatic bottle selection when
        appropriate.

        It is a helper for _categorize_bottles() and must only be called when
        all bottles have been assigned a category.
        """
        self._take_bottle_snapshot()
        if not self.existingBottleNameRequested: # FIXME: May be wrong
            self.createNewBottle = True
            self.existingBottleName = None
            for target_bottle in self.bottles.itervalues():
                if target_bottle.get_category() != CAT_RECOMMENDED:
                    pass
                elif self.existingBottleName:
                    # We have more than one recommended bottle so we need
                    # the user to pick one
                    self.existingBottleName = None
                    break
                else:
                    self.createNewBottle = False
                    self.existingBottleName = target_bottle.bottlename
        self._pick_new_bottle_name()
        self._pick_new_bottle_template()
        self._send_bottle_change_notifications()


    #####
    #
    # Bottle analysis and categorization
    #
    #####

    def _trigger_bottle_analysis(self):
        """Triggers a bottle analysis to:
        - determine which ones are compatible or incompatible with the
          selected profile and installation source,
        - determine which ones to recommend installation into (if any),
        - pick a bottle if possible,
        - and acquire all the data we need to start the installation.
        """
        # Should this prove too slow this will remain the entry point for
        # triggering the bottle analysis but part of the work will be done in
        # the background.
        if self._profile:
            self._categorize_bottles()
        else:
            # Note that we don't need to go through all the target bottle
            # objects to reset their categories thanks to their cache mechanism.
            # So all we need is to unset the selected bottle if appropriate.
            self._unpick_bottle()
            # Notify the GUI that the category of each bottle and template
            # has changed.
            for target in self.templates.itervalues():
                self._delegate.categorizedBottle_(target)
            for target in self.bottles.itervalues():
                self._delegate.categorizedBottle_(target)
            # Also notify it that we have recategorized all the bottles
            self._delegate.categorizedAllBottles()

    def _categorize_list(self, target_iter):
        """This is a helper for _categorize_bottles()."""
        categorized_all = True
        for target in target_iter:
            if not target.analyzed() and target.analyze():
                # Notify the GUI that we have freshly analyzed this
                # target bottle
                self._delegate.analyzedBottle_(target)
            if not target.has_category():
                if target.get_category() != CAT_NONE:
                    # Notify the GUI that we have freshly categorized this
                    # target bottle
                    self._delegate.categorizedBottle_(target)
                else:
                    categorized_all = False
        return categorized_all

    def _categorize_bottles(self):
        """This goes through all the bottles and templates trying to determine
        their category. If it was successful it then calls _pick_bottle() and
        notifies the GUI.
        """
        if self._categorize_list(self.templates.itervalues()) and \
                self._categorize_list(self.bottles.itervalues()):
            # We now know the category of all the bottles and templates and
            # thus we can now pick one
            self._pick_bottle()
            # The GUI may have been waiting for all the categories to be known
            # to display the bottle list. So notify it we're done.
            self._delegate.categorizedAllBottles()
        # else:
        #     This means we are waiting for one (or more) of the bottles
        #     installed applications list to compute the categories. So this
        #     method will be called again when installed_applications_ready()
        #     gets called for the relevant bottles.

    def get_compatible_templates(self):
        """Returns the list of bottle templates compatible with the currently
        selected application, in preference order.
        """
        compatible_templates = []
        for template_name in self._profile.app_profile.bottle_types['use']:
            if self.templates[template_name].get_category() == CAT_COMPATIBLE:
                compatible_templates.append(template_name)
        return compatible_templates


    #####
    #
    # Bottle addition and removal
    #
    #####

    def add_bottle(self, bottle):
        """Notifies InstallTask that a new bottle was detected.

        The caller MUST then schedule the detection of that bottle's installed
        applications and, when that's done, it MUST notify InstallTask through
        installed_applications_ready(). The one exception is if the bottle
        disappears before then and in that case it MUST notify InstallTask
        through remove_bottle().

        This MUST be called in the main thread.
        """
        self.bottles[bottle.name] = TargetBottle(self, bottle)
        return self.bottles[bottle.name]

    # Alias for Objective-C
    addBottle_ = add_bottle

    def installed_applications_ready(self, _bottle):
        """Notifies InstallTask that the specified bottle's installed
        applications list is now available.

        This MUST be called in the main thread.
        """
        # FIXME: Scan the installed applications list and update the extrafors.
        # This will use the _bottle parameter.
        self._trigger_bottle_analysis()

    # Alias for Objective-C
    installedApplicationsReady_ = installed_applications_ready

    def remove_bottle(self, bottle):
        """Notifies InstallTask that the specified bottle has been removed.

        This may be called _instead of_ installed_applications_ready() if the
        removal happened before the installed application detection could
        complete. In this case installed_applications_ready() must not be
        called for that bottle.

        It is also a bug to call remove_bottle() twice or to call it on
        unregistered bottles.

        This MUST be called in the main thread.
        """
        del self.bottles[bottle.name]
        # FIXME: Remove this bottle from the extrafors
        if self.existingBottleName == bottle.name:
            self._take_bottle_snapshot()
            self.existingBottleName = None
            # createNewBottle was necessarily False
            # FIXME: or maybe not with the current ultra-complex scheme
            self.existingBottleNameRequested = False # FIXME: May be wrong
            self._send_bottle_change_notifications()
        # The bottle removal does not impact any of the bottle categories but
        # it may impact the automatic bottle selection. So review things again.
        self._trigger_bottle_analysis()

    # Alias for Objective-C
    removeBottle_ = remove_bottle


    #####
    #
    # Target bottle selection
    #
    #####

    def is_new_bottle_name_valid(self):
        """Returns True if the new bottle name is valid and False otherwise.

        In particular this checks for that it does not match an existing
        bottle name, file or directory.
        This MUST be called in the main thread.
        """
        is_valid = True
        if not self.newBottleName:
            is_valid = False
        elif self.newBottleName != bottlequery.unique_bottle_name(cxutils.bottlename_sanitize(self.newBottleName)):
            is_valid = False
        return is_valid

    isNewBottleNameValid = is_new_bottle_name_valid

    def SetCreateNewBottle(self, create=None, newbottlename=None, template=None, requested=True):
        if create is None:
            create = self.createNewBottle
        if newbottlename is None:
            newbottlename = self.newBottleName
        if template is None:
            template = self.newBottleTemplate

        create_changed = self.createNewBottle != create
        name_changed = self.newBottleName != newbottlename
        template_changed = self.newBottleTemplate != template

        if create and newbottlename == '':
            self.createNewBottleRequested = False
            self.newBottleNameRequested = False
            self.newBottleTemplateRequested = False
        else:
            if create_changed:
                self.createNewBottleRequested = requested
            if name_changed:
                self.newBottleNameRequested = requested
            if template_changed:
                self.newBottleTemplateRequested = requested

        self.createNewBottle = create
        self.newBottleName = newbottlename
        if create and not self.is_new_bottle_name_valid():
            self._delegate.invalidNewBottleName()
        self.newBottleTemplate = template
        if create:
            self.existingBottleName = ''
            self.existingBottleNameRequested = False

        if create_changed:
            self._delegate.bottleCreateChanged()
        if name_changed:
            self._delegate.bottleNewnameChanged()
        if template_changed:
            self._delegate.bottleTemplateChanged()

    # Wrappers for Objective-C
    def createNewBottle_template_(self, newbottlename, template):
        self.SetCreateNewBottle(True, newbottlename, template)
    def createNewBottle_template_forUser_(self, newbottlename, template, requested):
        self.SetCreateNewBottle(True, newbottlename, template, requested)

    def GetCreateNewBottle(self):
        return self.createNewBottle

    def GetNewBottleTemplate(self):
        return self.newBottleTemplate

    def _get_bottle_name(self):
        if self.createNewBottle:
            return self.newBottleName
        else:
            return self.existingBottleName

    def _set_bottle_name(self, inName, requested=True):
        create_changed = self.createNewBottle != False
        name_changed = self.existingBottleName != inName

        if create_changed:
            self.createNewBottleRequested = requested
        if name_changed:
            self.existingBottleNameRequested = requested

        self.existingBottleName = inName
        self.createNewBottle = False

        if create_changed:
            self._delegate.bottleCreateChanged()
        if name_changed:
            self._delegate.bottleNameChanged()

    bottlename = property(_get_bottle_name, _set_bottle_name)


    def _gettarget_bottle(self):
        """Returns the TargetBottle object for the bottle selected for
        installation. This object can then be used to get all the details about
        that bottle: its category, the dependencies to install, and potential
        issues.

        Note also that this works equally well for existing bottles and new
        bottles.
        """
        if self.existingBottleName:
            return self.bottles[self.existingBottleName]
        elif self.createNewBottle and self.newBottleTemplate:
            return self.templates[self.newBottleTemplate]
        return None

    target_bottle = property(_gettarget_bottle)


    #####
    #
    # Can Install?
    #
    #####

    def can_install(self):
        """Returns True if we have all the data required to perform the
        installation and there is no blocking issue, and False otherwise.
        """
        # Do we have a profile?
        if not self._profile:
            return False
        # Do we have an installation source?
        if not self.installerDownloadSource and not self.installerSource and \
                not self.virtual:
            return False
        # Do we have a target bottle?
        if self.createNewBottle:
            if not self.is_new_bottle_name_valid() or \
                    not self.newBottleTemplate:
                return False
        elif not self.existingBottleName:
            return False
        # Is there any blocking compatibility issue?
        if not self.target_bottle.analyzed() or \
                self.target_bottle.missing_media:
            return False
        # All good
        return True

    # Wrapper for Objective-C
    canInstall = can_install


    #####
    #
    # Getting the summary data
    #
    #####

    def get_use_if_properties(self, bottle_type=None):
        """Returns a mapping containing the use_if properties for the selected
        profile, locale and bottle type combination."""
        if bottle_type is None:
            if self.createNewBottle:
                bottle_type = self.newBottleTemplate
            elif self.existingBottleName:
                properties = bottlequery.get_bottle_properties(self.existingBottleName)
                bottle_type = properties['template']
            else:
                bottle_type = self._profile.app_profile.preferred_bottle_type
        return {'appid': self._profile.appid,
                'locale': self.installerDownloadLocale,
                'bottletemplate': bottle_type}

    def get_installer_profile(self, bottle_type=None):
        """Computes the appropriate installer profile for the selected
        profile, locale and bottle type combination."""
        # FIXME: This method should go away and the code that uses it should
        # use or be migrated to target_bottle instead.
        use_if_props = self.get_use_if_properties(bottle_type=bottle_type)
        nomatch = True
        result = c4profiles.C4InstallerProfile()
        for inst_profile in self.profiles.installer_profiles(self._profile.appid):
            if inst_profile.use(use_if_props):
                result.update(inst_profile)
                nomatch = False
        if nomatch:
            result = self.profiles.unknown_installer().copy()
        return result

    def get_dependencies(self):
        """Returns a human-readable list of dependencies."""
        to_install = []
        installers = self.target_bottle.installers
        missing_media = self.target_bottle.missing_media
        missing_profiles = self.target_bottle.missing_profiles
        for depid in installers:
            if depid == self._profile.appid:
                continue

            if depid in missing_profiles:
                cxlog.warn( "The dependency " + cxlog.to_str(depid) + " is not available in our profile list.")
                to_install.append(_("%s (unavailable)") % depid)
            elif depid in missing_media:
                to_install.append(_("%s (needs cd)") % installers[depid].parent.name)
            else:
                to_install.append(installers[depid].parent.name)
        return to_install

    getDependencies = get_dependencies


    def _get_installation_notes(self):
        return self.get_installer_profile().installation_notes

    installationNotes = property(_get_installation_notes)


    def _find_warning_in_mapping(self, warning, mapping):
        # Search for the given key in a cxdiag mapping
        if '*' in warning or '?' in warning:
            # A wildcard string.
            regex = ['^']
            for char in warning:
                if char == '*':
                    regex.append('.*')
                elif char == '?':
                    regex.append('.')
                else:
                    regex.append(re.escape(char))
            regex.append('$')
            regex = re.compile(''.join(regex), re.IGNORECASE)

            for key in mapping:
                if regex.match(key):
                    yield key
        elif warning in mapping:
            yield warning

    def get_cxdiag_messages(self):
        """Returns a dictionary of cxdiag messages that apply to the current profile.

        The values are tuples of the form (level, title, description), already localized."""
        if self.cxdiag_messages is None:
            retcode, out, _err = cxutils.run(os.path.join(cxutils.CX_ROOT, "bin", "cxdiag"), stdout=cxutils.GRAB)
            if retcode:
                cxlog.warn("cxdiag failed to run")
                return {}
            config = cxconfig.CXMergeConfig()
            config.read_string(out)
            cxdiag_messages = {}
            for section in config.itervalues():
                cxdiag_messages[section.name.lower()] = (section['Level'], section['Title'], section['Description'])
            self.cxdiag_messages = cxdiag_messages

        result = self.cxdiag_messages.copy()

        if self._profile is not None:
            app_overrides = {}
            for check in self.get_installer_profile().cxdiag_checks:
                check = check.lower()
                if check.startswith('apprequire:') or check.startswith('apprecommend:'):
                    warning_name = check.split(':', 1)[1]
                    if check.startswith('apprequire:'):
                        new_level = 'Require'
                    elif check.startswith('apprecommend:'):
                        new_level = 'Recommend'
                    for key in self._find_warning_in_mapping(warning_name, result):
                        _level, title, description = result[key]
                        result[key] = app_overrides[key] = (new_level, title, description)
                elif check.startswith('ignore:'):
                    warning_name = check.split(':', 1)[1]
                    warnings_to_ignore = list(self._find_warning_in_mapping(warning_name, result))
                    for key in warnings_to_ignore:
                        del result[key]
                        if key in app_overrides:
                            del app_overrides[key]
                elif check == 'closed':
                    result = app_overrides.copy()
                else:
                    cxlog.warn("unrecognized cxdiagcheck string: %s" % cxlog.to_str(check))

        return result

